// vincia01.cc is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to run Z decays at LEP I
// Uses VINCIA for showering and matching and PYTHIA 8 for all else
// Copyright (c) 2016 Peter Skands

// PYTHIA and VINCIA include files and namespaces
#include "Vincia/Vincia.h"
using namespace Pythia8;
using namespace Vincia;

// ROOT: include files for linking to ROOT, uncomment as necessary
// #include "TROOT.h"
// #include "TDirectory.h"
// #include "TFile.h"
// #include "TH1.h"
// #include "TH2.h"

// Main Program
int main() {
  
  //************************************************************************
  // Define Pythia 8 generator with plug-in
  
  Pythia pythia;
  VinciaPlugin vincia(&pythia,"vincia01.cmnd");
  
  // Everything from here on is normal Pythia8
  // (with some slight extra functionality, such as post-run statistics from
  // Vincia and checking whether all events have unit weights)
  
  //************************************************************************
  
  // Shorthands
  Event& event = pythia.event;
  Settings& settings = pythia.settings;
  
  // Extract settings to be used in the main program.
  int    nEvent     = settings.mode("Main:numberOfEvents");
  bool   vinciaOn   = settings.flag("Vincia");
  int    verbose    = settings.mode("Vincia:verbose");
  bool   helicityOn = settings.flag("Vincia:helicityShower");
  int    iEventPri  = -1;
  double eCM        = settings.parm("Beams:eCM");
  
  //************************************************************************
  
  // Initialize
  vincia.init();
  
  //************************************************************************
  
  // Define a few PYTHIA utilities
  Thrust Thr(1);
  Sphericity SphLin(1,2);
  
  //************************************************************************
  
  // Define PYTHIA histograms
  // Rootlink: to use ROOT instead, HIST -> TH1F ("tag",...)
  
  // Number of quarks, partons, and charged particles
  double nQsum      = 0.0;
  double nPsum      = 0.0;
  double nCsum      = 0.0;
  double nStrangeSum = 0.0;
  double nCharmSum  = 0.0;
  double nBottomSum = 0.0;
  double nBaryonSum = 0.0;
  Hist HnQuarks("nQuarks", 50, -0.5, 49.5);
  Hist HnPartons("nPartons", 100, -0.5, 99.5);
  Hist HnCharged("nCharged", 200, -1.0, 199.0);
  Hist HxStrange("LnXstrange", 10, -5.0, 0.0);
  Hist HxCharm("LnXcharm", 10, -5.0, 0.0);
  Hist HxBottom("LnXbottom", 10, -5.0, 0.0);
  
  // Thrust
  int sBins=100;
  int sBinsCoarse=25;
  Hist Hthr("1-T",sBins, 0.0, 0.5);
  Hist HthrTail("1-T", sBinsCoarse, 0.0, 0.5);
  double wthr=sBins/0.5;
  double wthrCoarse=sBinsCoarse/0.5;
  // Event Shapes
  double wShape=sBins/1.0;
  Hist Hmaj("Major",sBins,0.0,1.0);
  Hist Hmin("Minor",sBins,0.0,1.0);
  Hist Hobl("Oblateness",sBins,0.0,1.0);
  Hist HistC("C",sBins,0.0,1.0);
  Hist HistD("D",sBins,0.0,1.0);
  
  //************************************************************************
  
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  
  // (Optionally) re-initialize random number engine before event loop
  const int initseed = settings.mode("Random:seed");
  vincia.rndmPtr->init(initseed);
  
  // Counter for negative-weight events
  double evwt=1.0;
  
  // Begin event loop
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {
    
    if (!pythia.next()) continue;
    
    // Check for weights
    if (vinciaOn) evwt = vincia.weight();

    // Print event with helicities
    if (iEvent == 0 || iEvent == iEventPri)
      if (helicityOn) event.list(helicityOn);
    
    // Count FS charged hadrons, partons, and quarks
    int nc=0;
    int np=0;
    int nq=0;
    int nStrange=0;
    int nCharm  =0;
    int nBottom =0;
    for (int i=1;i<event.size();i++) {
      // Count up final-state charged hadrons
      if (event[i].isCharged() && event[i].status() > 80) nc++;
      // Find last parton-level partons
      int iDau1 = event[i].daughter1();
      if (iDau1 == 0 || abs(event[iDau1].status()) > 80) {
        // Count up partons and quarks
        if (event[i].isQuark() || event[i].isGluon() ) np++;
        if (event[i].isQuark()) nq++;
        if (abs(event[i].id()) == 3) {
          nStrange++;
          HxStrange.fill(log(2*event[i].pAbs()/eCM));
        }
        if (abs(event[i].id()) == 4) {
          nCharm++;
          HxCharm.fill(log(2*event[i].pAbs()/eCM));
        }
        if (abs(event[i].id()) == 5) {
          nBottom++;
          HxBottom.fill(log(2*event[i].pAbs()/eCM));
        }
      }
      if (event[i].isHadron() && event[i].isFinal()) {
        int idAbs = abs(event[i].id());
        if (idAbs > 1000 && idAbs < 10000) nBaryonSum++;
      }
    }
    HnQuarks.fill(0.5*nq,evwt);
    HnPartons.fill(1.0*np,evwt);
    HnCharged.fill(1.0*nc,evwt);
    nQsum += 0.5*nq;
    nPsum += np;
    nCsum += nc;
    nStrangeSum += nStrange;
    nCharmSum += nCharm;
    nBottomSum += nBottom;
    
    // Histogram thrust
    Thr.analyze( event );
    Hthr.fill(1.0-Thr.thrust(),wthr*evwt);
    HthrTail.fill(1.0-Thr.thrust(),wthrCoarse*evwt);
    Hmaj.fill(Thr.tMajor(),wShape*evwt);
    Hmin.fill(Thr.tMinor(),wShape*evwt);
    Hobl.fill(Thr.oblateness(),wShape*evwt);
    
    // Histogram Linear Sphericity values
    if (np >= 2.0) {
      SphLin.analyze( event );
      double evC=3*(SphLin.eigenValue(1)*SphLin.eigenValue(2)
                    + SphLin.eigenValue(2)*SphLin.eigenValue(3)
                    + SphLin.eigenValue(3)*SphLin.eigenValue(1));
      double evD=27*SphLin.eigenValue(1)*SphLin.eigenValue(2)
      *SphLin.eigenValue(3);
      HistC.fill(evC,wShape*evwt);
      HistD.fill(evD,wShape*evwt);
    }
    
  }
  
  //************************************************************************
  
  // POST-RUN FINALIZATION
  // Normalization, Statistics, Output
  
  //Normalize histograms to effective number of positive-weight events.
  double normFac=1.0/nEvent;
  Hthr      *= normFac;
  HthrTail  *= normFac;
  HistC     *= normFac;
  HistD     *= normFac;
  Hmaj      *= normFac;
  Hmin      *= normFac;
  Hobl      *= normFac;
  HnQuarks    *= normFac;
  HnPartons   *= normFac;
  HnCharged   *= normFac;
  HxStrange   *= normFac;
  HxCharm     *= normFac;
  HxBottom    *= normFac;
  
  // Save a few tables to files
  /*
  HxStrange.table("strangeVJ.dat");
  HxCharm.table("charmVJ.dat");
  HxBottom.table("bottomVJ.dat");
  */

  // Print out (and possibly save) Vincia internal histograms
  if (vinciaOn && verbose >= 2) {
    vincia.printHistos();
    if (verbose >= 3) vincia.writeHistos("vincia01");
  }
  
  // Print a few histograms
  cout<<HnPartons<<endl;
  cout<<Hthr<<endl;
  cout<<HistC<<endl;
  cout<<HistD<<endl;

  /*
  cout<<HxStrange<<endl;
  cout<<HxCharm<<endl;
  cout<<HxBottom<<endl;
  */
  
  // Print out Vincia end-of-run information
  pythia.stat();
  if (vinciaOn) vincia.printInfo();
  
  cout<<endl;
  cout<<fixed;
  cout<< " <nGluonSplit> = "<<num2str(nQsum * normFac-1.0)<<endl;
  cout<< " <nPartons>    = "<<num2str(nPsum * normFac)<<endl;
  cout<< " <nCharged>    = "<<num2str(nCsum * normFac)<<endl;  
  cout<< " <nBaryons>    = "<<num2str(nBaryonSum * normFac)<<endl;  
  cout<<endl;
  
  // Done.
  return 0;
}
