// AntennaFunctions.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2016 P. Skands
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the AntennaFSR and 
// AntennaISR classes 
  
#ifndef VINCIAANTENNAFUNCTIONS_H
#define VINCIAANTENNAFUNCTIONS_H

#include "Pythia8/Basics.h"
#include "Pythia8/Event.h"
#include "Pythia8/PythiaStdlib.h"
#include "Vincia/VinciaCommon.h"      

#include "Vincia/SubtractionTerms.h"   // needed for DGLAP tests

namespace Vincia {
  
  // *****************************************************************
  // class AntennaFSR:
  // Base class implementation for generic Antenna objects

  class AntennaFSR {
    
  public:
    
    // Constructor
    AntennaFSR() {
      // Init flags
      isInitPtr = false;
      isInit    = false;      
      // Default at construction time is global antenna (reset by init)
      isSectorSav          = false;
      // Default values (reset by init)
      verbose              = 1;
      useCollinearTermsSav = true;
      useFiniteTermsSav    = true;
      // Initialize LH and RH maps (DO NOT CHANGE)
      RH[9]  = true;
      RH[1]  = true;
      RH[-1] = false;
      LH[9]  = true;
      LH[1]  = false;
      LH[-1] = true;
      // At construction time, everything is zero (reset by init)
      hasFiniteTerms  = false;
      Cs00   = 0.0;
      Cs10   = 0.0;
      Cs01   = 0.0;
      Cf00   = 0.0;
      Cf10   = 0.0;
      Cf01   = 0.0;
      Cp00   = 0.0;
      Cp10   = 0.0;
      Cp01   = 0.0;
      chargeFactorSav = 0.0;
      kineMapTypeSav  = 0;
    };
    
    // Destructor 
    virtual ~AntennaFSR() {};
    
    //=================================================================
    // List of VIRTUAL functions to be redefined by inheriting classes

    // Names of this antenna, for VINCIA, and for humans
    virtual string vinciaName()   const {return "Vincia:Antenna";};

    // Parton types (AB -> IJK): needed by soft- and collinear-limit checks 
    virtual int idA()    const {return 0;}
    virtual int idB()    const {return 0;}
    virtual int idJ()    const {return 0;}

    // The (dimensionless) antenna function.
    // Most generic form including masses and helicities.
    // Default parameter assignments => massless and unpolarized.
    virtual double antennaFunction(double yij, double yjk, 
                     double mui=0.0, double muj=0.0, double muk=0.0,
                     int hA=9, int hB=9, int hi=9, int hj=9, int hk=9);
    
    // Optional implementation of the AP kernels corresponding
    // to the two sides of the antenna, for collinear-limit checks
    // Defined as PI/yij + PK/yjk, i.e., equivalent to antennae.
    virtual double AltarelliParisi(double yij, double yjk,
                                   double mui=0., double muj=0., double muk=0.,
                                   int hA=9, int hB=9, 
                                   int hI=9, int hJ=9, int hK=9);

    //=================================================================
    // List of VIRTUAL functions that MAY be redefined (if needed)

    // Default initialization (minMaxVar = 0 -> user settings)
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);
      
    // Construct baseName from idA, idB, and idJ
    virtual string baseName()    const {
      string name = id2str(idJ()) + "/" + id2str(idA()) + id2str(idB());
      return name;
    }

    // Wrapper that appends (sct) to name if antenna is sector type
    // Is made virtual so can in principle be redefined, but should
    // be fine for most implementations
    virtual string humanName()   const {
      string antName = baseName();
      if (isSectorSav) antName +=" (sct)";
      return antName;
    }

    // Function to check singularities, positivity, etc.
    virtual bool check();

    //=================================================================
    // List of NON-VIRTUAL functions (i.e., that may not be redefined)

    // Initialize pointers:
    //    ParticleData : required for masses, color charges, etc
    //    Rndm : required to generate random test points to check antenna
    //    Settings : required for all coefficients, switches, etc.
    void initPtr(Functions* functionsPtrIn, ParticleData* particleDataPtrIn, 
                 Rndm* rndmPtrIn, Settings* settingsPtrIn) {
      functionsPtr    = functionsPtrIn;
      particleDataPtr = particleDataPtrIn; 
      rndmPtr         = rndmPtrIn;
      settingsPtr     = settingsPtrIn; 
      isInitPtr       = true;
      isInit          = false;
    };

    // -------
    // Functions to GET

    // Shorthand to get the massless antenna function
    double antennaMassless(double yij, double yjk, int hA=9, int hB=9, 
                           int hi=9, int hj=9, int hk=9) {
      return antennaFunction(yij,yjk,0.0,0.0,0.0,hA,hB,hi,hj,hk);
    }

    // Get parameters
    bool isOn() { return isOnSav; }
    bool isSector() { return isSectorSav; };    
    bool useFiniteTerms() { return useFiniteTermsSav; }
    bool useCollinearTerms() { return useCollinearTermsSav; }
    bool useSectorTerms() {
      if (isSectorSav) return useSectorTermsSav;
      else return false;
    }
    double getChargeFactor() { return chargeFactorSav; }
    int    getKineMapType() { return kineMapTypeSav; }
    double alpha() { return alphaSav; }

    // Shorthand for temporarily switching off finite terms to check
    // singular part only
    double getSingularTerms(double yij, double yjk, 
                            double mui=0.0, double muj=0.0, double muk=0.0,
                            int hA=9, int hB=9, 
                            int hi=9, int hj=9, int hk=9) {
      // Temporarily switch off finite terms 
      bool useFiniteTermsOrg = useFiniteTermsSav;
      useFiniteTermsSav = false;
      double a = antennaFunction(yij,yjk,mui,muj,muk,hA,hB,hi,hj,hk);
      // Restore finite terms and return
      useFiniteTermsSav = useFiniteTermsOrg;
      return a;
    }

    // Functions to get Altarelli-Parisi energy fractions from scaled invariants
    double zi(double yij,double yjk) {return (1.-yjk)/(1.+yij);}
    double zk(double yij,double yjk) {return (1.-yij)/(1.+yjk);}

    // *****************************************************************
    // id2str()
    // Auxiliary function to translate an ID code to a string, 
    // for constructing antenna names
    
    string id2str(int id) const {
      // Vector bosons
      if (id == 21) return "g";
      else if (id == 22) return "gamma";
      else if (id == 23) return "Z";
      else if (abs(id) == 24) return "W";
      // Fermions
      else if (id >= 1 && id <= 10) return "q";
      else if (-id >= 1 && -id <= 10) return "qbar";
      else if (id >= 11 && id <= 20 && id%2 == 1) return "l-"; 
      else if (id >= 11 && id <= 20 && id%2 == 0) return "nu"; 
      else if (-id >= 11 && -id <= 20 && id%2 == 1) return "l+"; 
      else if (-id >= 11 && -id <= 20 && id%2 == 0) return "nubar"; 
      // Octet fermion: use gluino
      else if (id == 1000021) return "~g";
      // Charged scalar: use H+
      else if (id == 37) return "H+";
      else if (id == -37) return "H-";
      // Colored scalars: use squarks
      else if (id >= 1000000 && id <= 1000010) return "~q";
      else if (-id >= 1000000 && -id <= 1000010) return "~q*";
      // Unknown particle
      else return "X";
    }
  
    //=================================================================
    // List of PROTECTED data members
    // (Not private, so can be accessed by inheriting classes)

  protected:
    
    // Is initialized
    bool isInitPtr, isInit;
    
    // Type of antenna
    bool isSectorSav;    

    // Antenna and terms on/off
    bool isOnSav;
    bool useFiniteTermsSav;
    bool useCollinearTermsSav;
    bool useSectorTermsSav;
    
    // Charge factor and kinematics map
    double chargeFactorSav;
    int kineMapTypeSav;

    // The alpha collinear-partitioning parameter
    double alphaSav;
    
    // Finite terms: separate for s-wave (MHV), s-wave (NMHV), and p-wave
    bool hasFiniteTerms;
    double Cs00, Cs01, Cs10;
    double Cf00, Cf01, Cf10;
    double Cp00, Cp01, Cp10;
    // So far not allowing finite terms for sector antennae with external
    // helicity flips, but these could in principle also be added here.
    
    // Map to tell whether a given helicity value maps to L- and/or R-handed
    // Defined by constructor and not to be changed dynamically.
    map<int,bool> LH;
    map<int,bool> RH;

    // Verbose level
    int verbose;

    // Pointers to Vincia classes
    Functions* functionsPtr;

    // Pointers to Pythia-8 classes
    ParticleData* particleDataPtr;
    Rndm*         rndmPtr;
    Settings*     settingsPtr;  
    
  };

  //**************************************************************************
  //
  // Inherited Antennae

  //--------------------------------------------------------------------------
  // class QQEmit 

  class QQEmit : public AntennaFSR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const {return "Vincia:QQemit";};

    // Functions needed by soft- and collinear-limit checks (AB -> IJK)
    virtual int idA()    const {return 1;}
    virtual int idB()    const {return -1;}
    virtual int idJ()    const {return 21;}

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The (dimensionless) antenna function
    virtual double antennaFunction(double yij, double yjk, 
                     double mui=0.0, double muj=0.0, double muk=0.0,
                     int hA=9, int hB=9, int hi=9, int hj=9, int hk=9);

    // Auxiliary: unpolarized antenna function for massive particles
    virtual double antennaMassive(double yij, double yjk, 
                                  double mui, double muk);
    
    // Function to give Altarelli-Parisi limits of this antenna
    // Defined as PI/yij + PK/yjk, i.e., equivalent to antennae.
    virtual double AltarelliParisi(double yij, double yjk, 
                                   double, double, double, int hA=9, int hB=9, 
                                   int hI=9, int hJ=9, int hK=9){
      if (hA != hI || hB != hK) return -1.;
      else return functionsPtr->Pq2qg(zi(yij,yjk),hA,hI,hJ)/yij 
             + functionsPtr->Pq2qg(zk(yij,yjk),hB,hK,hJ)/yjk;
    }

    // Additional data members specific to this class
  private:

  };


  //--------------------------------------------------------------------------
  // class QGEmit 

  class QGEmit : public AntennaFSR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const {return "Vincia:QGemit";};

    // Parton types (AB -> IJK): needed by soft- and collinear-limit checks 
    virtual int idA()    const {return 1;}
    virtual int idB()    const {return 21;}
    virtual int idJ()    const {return 21;}

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The (dimensionless) antenna function
    virtual double antennaFunction(double yij, double yjk, 
                     double mui=0.0, double muj=0.0, double muk=0.0,
                     int hA=9, int hB=9, int hi=9, int hj=9, int hk=9);

    // Auxiliary: unpolarized antenna function for massive particles
    virtual double antennaMassive(double yij, double yjk, double mui);
    
    // Function to give Altarelli-Parisi limits of this antenna
    virtual double AltarelliParisi(double yij, double yjk,
                                   double, double, double, int hA=9, int hB=9, 
                                   int hI=9, int hJ=9, int hK=9) {
      double sum(0.);
      if (hI != hA) return -1;
      if (hK == hB) sum += functionsPtr->Pq2qg(zi(yij,yjk),hA,hI,hJ)/yij;
      sum += functionsPtr->Pg2gg(zk(yij,yjk),hB,hK,hJ)/yjk;
      return sum;
    }

    // Additional data members specific to this class
  private:
    
  };

  //--------------------------------------------------------------------------
  // class GGEmit 

  class GGEmit : public AntennaFSR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const {return "Vincia:GGemit";};

    // Parton types (AB -> IJK): needed by soft- and collinear-limit checks 
    virtual int idA()    const {return 21;}
    virtual int idB()    const {return 21;}
    virtual int idJ()    const {return 21;}

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The (dimensionless) antenna function
    virtual double antennaFunction(double yij, double yjk, 
                     double mui=0.0, double muj=0.0, double muk=0.0,
                     int hA=9, int hB=9, int hi=9, int hj=9, int hk=9);

    // Function to give Altarelli-Parisi limits of this antenna
    virtual double AltarelliParisi(double yij, double yjk,
                                   double, double, double, int hA=9, int hB=9, 
                                   int hI=9, int hJ=9, int hK=9) {
      double sum(0.);
      if ( hB == hK ) sum += functionsPtr->Pg2gg(zi(yij,yjk),hA,hI,hJ)/yij;
      if ( hA == hI ) sum += functionsPtr->Pg2gg(zk(yij,yjk),hB,hK,hJ)/yjk;
      return sum;
    }

    // Additional data members specific to this class
  private:

  };

  //--------------------------------------------------------------------------
  // class QGSplit

  class QGSplit : public AntennaFSR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const {return "Vincia:QGsplit";};

    // Parton types (AB -> IJK): needed by soft- and collinear-limit checks 
    virtual int idA()    const {return 1;}
    virtual int idB()    const {return 21;}
    virtual int idJ()    const {return -2;}

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The (dimensionless) antenna function
    virtual double antennaFunction(double yij, double yjk, 
                     double mui=0.0, double muj=0.0, double muk=0.0,
                     int hA=9, int hB=9, int hi=9, int hj=9, int hk=9);

    // Auxiliary: unpolarized antenna function for massive particles
    virtual double antennaMassive(double yij, double yjk, 
                                  double mui, double muj);

    // Function to give Altarelli-Parisi limits of this antenna
    virtual double AltarelliParisi(double yij, double yjk,
                                   double, double, double, int hA=9, int hB=9, 
                                   int hI=9, int hJ=9, int hK=9) {
      if (hA != hI) return 0.0; 
      else return functionsPtr->Pg2qq(zk(yij,yjk),hB,hK,hJ)/yjk;
    }

    // Additional data members specific to this class
  private:

  };

  //--------------------------------------------------------------------------
  // class GGSplit 
  // Note: function coded is for g g -> g qbar q
  // Vincia automatically includes the other permutation

  class GGSplit : public AntennaFSR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const {return "Vincia:GGsplit";};

    // Parton types (AB -> IJK): needed by soft- and collinear-limit checks 
    virtual int idA()    const {return 21;}
    virtual int idB()    const {return 21;}
    virtual int idJ()    const {return -2;}

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The (dimensionless) antenna function
    virtual double antennaFunction(double yij, double yjk, 
                     double mui=0.0, double muj=0.0, double muk=0.0,
                     int hA=9, int hB=9, int hi=9, int hj=9, int hk=9);

    // Auxiliary: unpolarized antenna function for massive particles
    virtual double antennaMassive(double yij, double yjk, double muj);    

    // Function to give Altarelli-Parisi limits of this antenna
    virtual double AltarelliParisi(double yij, double yjk,
                                   double, double, double, int hA=9, int hB=9, 
                                   int hI=9, int hJ=9, int hK=9) {
      if (hA != hI) return 0.;
      else return functionsPtr->Pg2qq(zk(yij,yjk),hB,hK,hJ)/yjk;
    }

    // Additional data members specific to this class
  private:

  };
  
  
  // *****************************************************************
  // Separate classes for ISR, added by NF as the antennae will be 
  // dimensionful.
  
  // *****************************************************************
  // class AntennaISR:
  // Antenna Functions for the Space Shower, similar to the Antenna class
  // for the Time Shower.

  class AntennaISR {
    
  public:
    
    // Constructor
    AntennaISR() {
      // Init flags
      isInitPtr           = false;
      isInit              = false;
      // Default values (reset by init)
      verbose             = 1;
      isOnSav             = true;
      chargeFactorSav     = 0.0;
      preFacFiniteTermSav = 0.0;
      antMinSav           = 0.0;
      // Initialize LH and RH maps (DO NOT CHANGE)
      RH[9]  = true;
      RH[1]  = true;
      RH[-1] = false;
      LH[9]  = true;
      LH[1]  = false;
      LH[-1] = true;
    };
    
    // Destructor 
    virtual ~AntennaISR() {};
    
    //=================================================================
    // List of VIRTUAL functions to be redefined by inheriting classes

    // Names of this antenna, for VINCIA, and for humans
    virtual string vinciaName()   const {return "Vincia:AntennaISR";};

    // Parton types AB -> ajb for II
    //              AK -> ajk for IF
    virtual int idA()    const { return 0; }
    virtual int idB()    const { return 0; }
    virtual int ida()    const { return 0; }
    virtual int idb()    const { return 0; }
    virtual int idj()    const { return 0; }

    // The antenna function
    virtual double antennaFunction(double s1j, double sj2, double S12, 
                                   int h1Old=9, int h2Old=9, 
                                   int h1=9, int h2=9, int h3=9);
    
    // Optional implementation of the AP kernels corresponding
    // to the two sides of the antenna, for collinear-limit checks.
    // NOTE AP kernels are implemented such that they are easily
    // comparable to the antennae, ie alpha/2pi*C vs alpha/4pi*C results
    // in a factor 1 for CF vs 2CF and TR/2 vs TR, but AP get multiplied
    // by 2 for CA vs CA. Colour factors: CA = 3, CF = 4/3, TR = 1.
    // Also additional factors of 1/2 are possible (taking into
    // account shared singularities)
    virtual double AltarelliParisi(double s1j, double sj2, double S12, 
                                   int h1Old=9, int h2Old=9, 
                                   int h1=9, int h2=9, int h3=9);

    //=================================================================
    // List of VIRTUAL functions that MAY be redefined (if needed)

    // Default initialization
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);
      
    // Construct baseName from idA, idB, and idJ
    virtual string baseName() const {
      string name = id2str(idj()) + "/" + id2str(idA()) + id2str(idB());
      return name;
    }

    // Functions to get AP energy fractions
    virtual double z1(double, double, double) { return -1.0; }
    virtual double z2(double, double, double) { return -1.0; }
    
    // Function to tell if this is an II antenna
    virtual bool isIIant() { return false; }

    // Function to check singularities, positivity, etc.
    virtual bool check();

    //=================================================================
    // List of NON-VIRTUAL functions (i.e., that may not be redefined)

    // Initialize pointers:
    //    ParticleData : required for masses, color charges, etc
    //    Rndm : required to generate random test points to check antenna
    //    Settings : required for all coefficients, switches, etc.
    void initPtr(Functions* functionsPtrIn, ParticleData* particleDataPtrIn, 
                 Rndm* rndmPtrIn, Settings* settingsPtrIn) {
      functionsPtr    = functionsPtrIn;
      particleDataPtr = particleDataPtrIn; 
      rndmPtr         = rndmPtrIn;
      settingsPtr     = settingsPtrIn; 
      isInitPtr       = true;
      isInit          = false;
    };

    // -------
    // Functions

    // Get parameters
    bool isOn() { return isOnSav; }
    double getChargeFactor() { return chargeFactorSav; }

    // *****************************************************************
    // id2str()
    // Auxiliary function to translate an ID code to a string, 
    // for constructing antenna names
    
    string id2str(int id) const {
      // Vector bosons
      if (id == 21) return "g";
      else if (id == 22) return "gamma";
      else if (id == 23) return "Z";
      else if (abs(id) == 24) return "W";
      // Fermions
      else if (id >= 1 && id <= 10) return "q";
      else if (-id >= 1 && -id <= 10) return "qbar";
      else if (id >= 11 && id <= 20 && id%2 == 1) return "l-"; 
      else if (id >= 11 && id <= 20 && id%2 == 0) return "nu"; 
      else if (-id >= 11 && -id <= 20 && id%2 == 1) return "l+"; 
      else if (-id >= 11 && -id <= 20 && id%2 == 0) return "nubar"; 
      // Octet fermion: use gluino
      else if (id == 1000021) return "~g";
      // Charged scalar: use H+
      else if (id == 37) return "H+";
      else if (id == -37) return "H-";
      // Colored scalars: use squarks
      else if (id >= 1000000 && id <= 1000010) return "~q";
      else if (-id >= 1000000 && -id <= 1000010) return "~q*";
      // Unknown particle
      else return "X";
    }
  
    //=================================================================
    // List of PROTECTED data members
    // (Not private, so can be accessed by inheriting classes)

  protected:
    
    // Is initialized
    bool isInitPtr, isInit;  

    // Antenna on/off
    bool isOnSav;
    
    // Charge factor
    double chargeFactorSav;
    
    // Prefactor of finite term and ant minimum
    double preFacFiniteTermSav;
    bool   isMinVar;
    double antMinSav;
    
    // Map to tell whether a given helicity value maps to L- and/or R-handed
    // Defined by constructor and not to be changed dynamically.
    map<int,bool> LH;
    map<int,bool> RH;

    // Verbose level
    int verbose;

    // Pointers to Vincia classes
    Functions* functionsPtr;

    // Pointers to Pythia-8 classes
    ParticleData* particleDataPtr;
    Rndm*         rndmPtr;
    Settings*     settingsPtr;  
    
  };

  //**************************************************************************
  //
  // Inherited Antennae for Initital - Initial

  //--------------------------------------------------------------------------
  // class QQEmitII

  class QQEmitII : public AntennaISR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const { return "Vincia:QQemitII"; };

    // Parton types AB -> ajb with A,B,a,b initial and j final
    virtual int idA()    const { return 1; }
    virtual int idB()    const { return -1; }
    virtual int ida()    const { return 1; }
    virtual int idb()    const { return -1; }
    virtual int idj()    const { return 21; }

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The antenna function
    virtual double antennaFunction(double saj, double sjb, double sAB, 
                                   int hA=9, int hB=9, 
                                   int ha=9, int hj=9, int hb=9);
    
    // AP splitting kernel
    virtual double AltarelliParisi(double saj, double sjb, double sAB, 
                                   int hA=9, int hB=9, 
                                   int ha=9, int hj=9, int hb=9);

    // Functions to get Altarelli-Parisi energy fractions
    virtual double z1(double    , double sjb, double sAB) { return sAB/(sAB+sjb); }
    virtual double z2(double saj, double    , double sAB) { return sAB/(sAB+saj); }
    
    // Function to tell if this is an II antenna
    virtual bool isIIant() { return true; }

    // Additional data members specific to this class
    private:

  };

  //--------------------------------------------------------------------------
  // class GQEmitII

  class GQEmitII : public AntennaISR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const { return "Vincia:GQemitII"; };

    // Parton types AB -> ajb with A,B,a,b initial and j final
    virtual int idA()    const { return 21; }
    virtual int idB()    const { return 1; }
    virtual int ida()    const { return 21; }
    virtual int idb()    const { return 1; }
    virtual int idj()    const { return 21; }

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The antenna function
    virtual double antennaFunction(double saj, double sjb, double sAB, 
                                   int hA=9, int hB=9, 
                                   int ha=9, int hj=9, int hb=9);
    
    // AP splitting kernel
    virtual double AltarelliParisi(double saj, double sjb, double sAB, 
                                   int hA=9, int hB=9, 
                                   int ha=9, int hj=9, int hb=9);

    // Functions to get Altarelli-Parisi energy fractions
    virtual double z1(double    , double sjb, double sAB) { return sAB/(sAB+sjb); }
    virtual double z2(double saj, double    , double sAB) { return sAB/(sAB+saj); }
    
    // Function to tell if this is an II antenna
    virtual bool isIIant() { return true; }

    // Additional data members specific to this class
    private:
    
  };

  //--------------------------------------------------------------------------
  // class GGEmitII 

  class GGEmitII : public AntennaISR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const { return "Vincia:GGemitII"; };

    // Parton types AB -> ajb with A,B,a,b initial and j final
    virtual int idA()    const { return 21; }
    virtual int idB()    const { return 21; }
    virtual int ida()    const { return 21; }
    virtual int idb()    const { return 21; }
    virtual int idj()    const { return 21; }

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The antenna function
    virtual double antennaFunction(double saj, double sjb, double sAB, 
                                   int hA=9, int hB=9, 
                                   int ha=9, int hj=9, int hb=9);
    
    // AP splitting kernel
    virtual double AltarelliParisi(double saj, double sjb, double sAB, 
                                   int hA=9, int hB=9, 
                                   int ha=9, int hj=9, int hb=9);

    // Functions to get Altarelli-Parisi energy fractions
    virtual double z1(double    , double sjb, double sAB) { return sAB/(sAB+sjb); }
    virtual double z2(double saj, double    , double sAB) { return sAB/(sAB+saj); }
    
    // Function to tell if this is an II antenna
    virtual bool isIIant() { return true; }

    // Additional data members specific to this class
    private:

  };

  //--------------------------------------------------------------------------
  // class QXSplitII
  // NOTE Splitting in the forwards sense aka quark backwards evolving
  //       to a gluon and emitting a antiquark in the final state

  class QXSplitII : public AntennaISR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const { return "Vincia:QXsplitII"; };

    // Parton types AB -> ajb with A,B,a,b initial and j final
    virtual int idA()    const { return 1; }
    virtual int idB()    const { return 0; }
    virtual int ida()    const { return 21; }
    virtual int idb()    const { return 0; }
    virtual int idj()    const { return -1; }

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The antenna function
    virtual double antennaFunction(double saj, double sjb, double sAB, 
                                   int hA=9, int hB=9, 
                                   int ha=9, int hj=9, int hb=9);
    
    // AP splitting kernel
    virtual double AltarelliParisi(double saj, double sjb, double sAB, 
                                   int hA=9, int hB=9, 
                                   int ha=9, int hj=9, int hb=9);

    // Functions to get Altarelli-Parisi energy fractions
    virtual double z1(double    , double sjb, double sAB) { return sAB/(sAB+sjb); }
    virtual double z2(double    , double    , double    ) { return -1.0; }
    
    // Function to tell if this is an II antenna
    virtual bool isIIant() { return true; }

    // Additional data members specific to this class
    private:

  };

  //--------------------------------------------------------------------------
  // class GXConvII
  // NOTE Gluon evolves backwards into a quark and emits a quark
  //       in the final state

  class GXConvII : public AntennaISR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const { return "Vincia:GXconvII"; };

    // Parton types AB -> ajb with A,B,a,b initial and j final
    virtual int idA()    const { return 21; }
    virtual int idB()    const { return 0; }
    virtual int ida()    const { return 2; }
    virtual int idb()    const { return 0; }
    virtual int idj()    const { return 2; }

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The antenna function
    virtual double antennaFunction(double saj, double sjb, double sAB, 
                                   int hA=9, int hB=9, 
                                   int ha=9, int hj=9, int hb=9);
    
    // AP splitting kernel
    virtual double AltarelliParisi(double saj, double sjb, double sAB, 
                                   int hA=9, int hB=9, 
                                   int ha=9, int hj=9, int hb=9);

    // Functions to get Altarelli-Parisi energy fractions
    virtual double z1(double    , double sjb, double sAB) { return sAB/(sAB+sjb); }
    virtual double z2(double    , double    , double    ) { return -1.0; }
    
    // Function to tell if this is an II antenna
    virtual bool isIIant() { return true; }

    // Additional data members specific to this class
    private:

  };

  //**************************************************************************
  //
  // Inherited Antennae for Initital - Final

  //--------------------------------------------------------------------------
  // class QQEmitIF

  class QQEmitIF : public AntennaISR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const { return "Vincia:QQemitIF"; };

    // Parton types AB -> ajb with A,a initial and B,b,j final
    virtual int idA()    const { return 1; }
    virtual int idB()    const { return -1; }
    virtual int ida()    const { return 1; }
    virtual int idb()    const { return -1; }
    virtual int idj()    const { return 21; }

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The antenna function
    virtual double antennaFunction(double saj, double sjk, double sAK, 
                                   int hA=9, int hK=9, 
                                   int ha=9, int hj=9, int hk=9);
    
    // AP splitting kernel
    virtual double AltarelliParisi(double saj, double sjk, double sAK, 
                                   int hA=9, int hK=9, 
                                   int ha=9, int hj=9, int hk=9);

    // Functions to get Altarelli-Parisi energy fractions
    virtual double z1(double    , double sjk, double sAK) { return sAK/(sAK+sjk); }
    virtual double z2(double saj, double    , double sAK) { return (sAK-saj)/sAK; }
    
    // Function to tell if this is an II antenna
    virtual bool isIIant() { return false; }

    // Additional data members specific to this class
    private:

  };

  //--------------------------------------------------------------------------
  // class QGEmitIF

  class QGEmitIF : public AntennaISR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const { return "Vincia:QGemitIF"; };

    // Parton types AB -> ajb with A,a initial and B,b,j final
    virtual int idA()    const { return 1; }
    virtual int idB()    const { return 21; }
    virtual int ida()    const { return 1; }
    virtual int idb()    const { return 21; }
    virtual int idj()    const { return 21; }

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The antenna function
    virtual double antennaFunction(double saj, double sjk, double sAK, 
                                   int hA=9, int hK=9, 
                                   int ha=9, int hj=9, int hk=9);
    
    // AP splitting kernel
    virtual double AltarelliParisi(double saj, double sjk, double sAK, 
                                   int hA=9, int hK=9, 
                                   int ha=9, int hj=9, int hk=9);

    // Functions to get Altarelli-Parisi energy fractions
    virtual double z1(double    , double sjk, double sAK) { return sAK/(sAK+sjk); }
    virtual double z2(double saj, double    , double sAK) { return (sAK-saj)/sAK; }
    
    // Function to tell if this is an II antenna
    virtual bool isIIant() { return false; }

    // Additional data members specific to this class
    private:
    
  };

  //--------------------------------------------------------------------------
  // class GQEmitIF

  class GQEmitIF : public AntennaISR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const { return "Vincia:GQemitIF"; };

    // Parton types AB -> ajb with A,a initial and B,b,j final
    virtual int idA()    const { return 21; }
    virtual int idB()    const { return 1; }
    virtual int ida()    const { return 21; }
    virtual int idb()    const { return 1; }
    virtual int idj()    const { return 21; }

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The antenna function
    virtual double antennaFunction(double saj, double sjk, double sAK, 
                                   int hA=9, int hK=9, 
                                   int ha=9, int hj=9, int hk=9);
    
    // AP splitting kernel
    virtual double AltarelliParisi(double saj, double sjk, double sAK, 
                                   int hA=9, int hK=9, 
                                   int ha=9, int hj=9, int hk=9);

    // Functions to get Altarelli-Parisi energy fractions
    virtual double z1(double    , double sjk, double sAK) { return sAK/(sAK+sjk); }
    virtual double z2(double saj, double    , double sAK) { return (sAK-saj)/sAK; }
    
    // Function to tell if this is an II antenna
    virtual bool isIIant() { return false; }

    // Additional data members specific to this class
    private:
    
  };

  //--------------------------------------------------------------------------
  // class GGEmitIF 

  class GGEmitIF : public AntennaISR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const { return "Vincia:GGemitIF"; };

    // Parton types AB -> ajb with A,a initial and B,b,j final
    virtual int idA()    const { return 21; }
    virtual int idB()    const { return 21; }
    virtual int ida()    const { return 21; }
    virtual int idb()    const { return 21; }
    virtual int idj()    const { return 21; }

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The antenna function
    virtual double antennaFunction(double saj, double sjk, double sAK, 
                                   int hA=9, int hK=9, 
                                   int ha=9, int hj=9, int hk=9);
    
    // AP splitting kernel
    virtual double AltarelliParisi(double saj, double sjk, double sAK, 
                                   int hA=9, int hK=9, 
                                   int ha=9, int hj=9, int hk=9);

    // Functions to get Altarelli-Parisi energy fractions
    virtual double z1(double    , double sjk, double sAK) { return sAK/(sAK+sjk); }
    virtual double z2(double saj, double    , double sAK) { return (sAK-saj)/sAK; }
    
    // Function to tell if this is an II antenna
    virtual bool isIIant() { return false; }

    // Additional data members specific to this class
    private:

  };

  //--------------------------------------------------------------------------
  // class QXSplitIF
  // NOTE Splitting in the forwards sense aka quark backwards evolving
  //       to a gluon and emitting a antiquark in the final state

  class QXSplitIF : public AntennaISR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const { return "Vincia:QXsplitIF"; };

    // Parton types AB -> ajb with A,a initial and B,b,j final
    virtual int idA()    const { return 1; }
    virtual int idB()    const { return 0; }
    virtual int ida()    const { return 21; }
    virtual int idb()    const { return 0; }
    virtual int idj()    const { return -1; }

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The antenna function
    virtual double antennaFunction(double saj, double sjk, double sAK, 
                                   int hA=9, int hK=9, 
                                   int ha=9, int hj=9, int hk=9);
    
    // AP splitting kernel
    virtual double AltarelliParisi(double saj, double sjk, double sAK, 
                                   int hA=9, int hK=9, 
                                   int ha=9, int hj=9, int hk=9);

    // Functions to get Altarelli-Parisi energy fractions
    virtual double z1(double, double sjk, double sAK) { return sAK/(sAK+sjk); }
    virtual double z2(double, double    , double    ) { return -1.0; }
    
    // Function to tell if this is an II antenna
    virtual bool isIIant() { return false; }

    // Additional data members specific to this class
    private:

  };

  //--------------------------------------------------------------------------
  // class GXConvIF
  // NOTE Gluon evolves backwards into a quark and emits a quark
  //       in the final state

  class GXConvIF : public AntennaISR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const { return "Vincia:GXconvIF"; };

    // Parton types AB -> ajb with A,a initial and B,b,j final
    virtual int idA()    const { return 21; }
    virtual int idB()    const { return 0; }
    virtual int ida()    const { return 2; }
    virtual int idb()    const { return 0; }
    virtual int idj()    const { return 2; }

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The antenna function
    virtual double antennaFunction(double saj, double sjk, double sAK, 
                                   int hA=9, int hK=9, 
                                   int ha=9, int hj=9, int hk=9);
    
    // AP splitting kernel
    virtual double AltarelliParisi(double saj, double sjk, double sAK, 
                                   int hA=9, int hK=9, 
                                   int ha=9, int hj=9, int hk=9);

    // Functions to get Altarelli-Parisi energy fractions
    virtual double z1(double, double sjk, double sAK) { return sAK/(sAK+sjk); }
    virtual double z2(double, double    , double    ) { return -1.0; }
    
    // Function to tell if this is an II antenna
    virtual bool isIIant() { return false; }

    // Additional data members specific to this class
    private:

  };

  //--------------------------------------------------------------------------
  // class XGSplitIF
  // NOTE Gluon splitting in the final state

  class XGSplitIF : public AntennaISR {

    // Names (remember to redefine both for each inherited class)
    virtual string vinciaName()    const { return "Vincia:XGsplitIF"; };

    // Parton types AB -> ajb with A,a initial and B,b,j final
    virtual int idA()    const { return 0; }
    virtual int idB()    const { return 21; }
    virtual int ida()    const { return 0; }
    virtual int idb()    const { return 1; }
    virtual int idj()    const { return -1; }

    // Init
    virtual bool init(bool isSectorIn = false, int minMaxVar = 0);

    // The antenna function
    virtual double antennaFunction(double saj, double sjk, double sAK, 
                                   int hA=9, int hK=9, 
                                   int ha=9, int hj=9, int hk=9);
    
    // AP splitting kernel
    virtual double AltarelliParisi(double saj, double sjk, double sAK, 
                                   int hA=9, int hK=9, 
                                   int ha=9, int hj=9, int hk=9);

    // Functions to get Altarelli-Parisi energy fractions
    virtual double z1(double    , double, double    ) { return -1.0; }
    virtual double z2(double saj, double, double sAK) { return (sAK-saj)/sAK; }
    
    // Function to tell if this is an II antenna
    virtual bool isIIant() { return false; }

    // Additional data members specific to this class
    private:

  };
  
} // End Vincia namespace

#endif

