// AntennaSets.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2016 P. Skands
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the AntennaSet class 
  
#ifndef VINCIAANTENNASETS_H
#define VINCIAANTENNASETS_H

// Global VINCIA functionality
#include "Vincia/VinciaCommon.h"
// Specific VINCIA headers
#include "Vincia/AntennaFunctions.h"

namespace Vincia {
  
  //******************************************************************
  // The AntennaSet class:
  // Contains a list of Antennae (e.g., the "DEF" set, the "GGG" set, etc)
  
  class AntennaSetFSR {
    
  public:
    
    // Destructor: delete the antennae
    ~AntennaSetFSR(){
      for (vector<AntennaFSR*>::iterator antPtrIt = antPtrs.begin(); 
           antPtrIt != antPtrs.end(); ++antPtrIt) delete *antPtrIt;
    }
    // Initialize pointers
    void initPtr(Functions* functionsPtrIn, ParticleData* particleDataPtrIn, 
                 Rndm* rndmPtrIn, Settings* settingsPtrIn) {
      functionsPtr    = functionsPtrIn;
      particleDataPtr = particleDataPtrIn; 
      rndmPtr         = rndmPtrIn;
      settingsPtr     = settingsPtrIn; 
      isInitPtr       = true;
    }
 
    // Initialize antenna set (optionally with min or max variation)
    void init(bool isSectorIn = false, int variation = 0);

    // Function to chek if an antenna with the given index exists
    bool exists(int iAntIn) {
      return ( iAntIn<0 || iAntIn >= int(antPtrs.size()) ) ? false : true;
    }
    
    // Gets an antenna iAntIn from the AntennaSet
    AntennaFSR* getAnt(vector<AntennaFSR*>::size_type iAntIn) { 
      if (iAntIn <= antPtrs.size()) {
        return antPtrs[iAntIn];
      }
      else return NULL;
    }

    // Get Vincia name (e.g., "Vincia:QQ:emit")
    string vinciaName(vector<AntennaFSR*>::size_type iAntIn) {
      if (iAntIn <= antPtrs.size()) {
        return antPtrs[iAntIn]->vinciaName();
      }
      else return "noVinciaName";        
    }

    // Get human name (e.g., "g/qq")
    string humanName(vector<AntennaFSR*>::size_type iAntIn) {
      if (iAntIn <= antPtrs.size()) {
        return antPtrs[iAntIn]->humanName();
      }
      else return "noHumanName";        
    }

  private:

    // use a vector of AntennaFSR pointers, create them with new on 
    // initialization
    vector<AntennaFSR*> antPtrs;

    // Verbose level
    int verbose;

    // Pointers to Vincia classes
    Functions*    functionsPtr;

    // Pointers to Pythia-8 classes, needed to initialise antennae
    bool isInitPtr;
    ParticleData* particleDataPtr;
    Rndm*         rndmPtr;
    Settings*     settingsPtr;  

  }; // End class AntennaSetFSR
  

  //******************************************************************
  // The AntennaSetISR class:
  // Added by NF to keep antenna sets for ISR shower separate from
  // from the FSR ones.
  
  class AntennaSetISR {
    
  public:
    
    // Destructor: delete the antennae
    ~AntennaSetISR(){
      for (vector<AntennaISR*>::iterator antPtrIt = antPtrs.begin(); 
           antPtrIt != antPtrs.end(); ++antPtrIt) delete *antPtrIt;
    }
    // Initialize pointers
    void initPtr(Functions* functionsPtrIn, ParticleData* particleDataPtrIn, 
                 Rndm* rndmPtrIn, Settings* settingsPtrIn) {
      functionsPtr    = functionsPtrIn;
      particleDataPtr = particleDataPtrIn; 
      rndmPtr         = rndmPtrIn;
      settingsPtr     = settingsPtrIn; 
      isInitPtr       = true;
    }
 
    // Initialize antenna set (optionally with min or max variation)
    void init(bool isSectorIn = false, int variation = 0);

    // Function to chek if an antenna with the given index exists
    bool exists(int iAntIn) {
      return ( iAntIn<0 || iAntIn >= int(antPtrs.size()) ) ? false : true;
    }
    
    // Gets an antenna iAntIn from the AntennaSetISR
    AntennaISR* getAnt(vector<AntennaISR*>::size_type iAntIn) { 
      if (iAntIn <= antPtrs.size()) return antPtrs[iAntIn];
      else return NULL;
    }

    // Get Vincia name (e.g., "Vincia:QQemitII")
    string vinciaName(vector<AntennaISR*>::size_type iAntIn) {
      if (iAntIn <= antPtrs.size()) return antPtrs[iAntIn]->vinciaName();
      else return "noVinciaName";        
    }

  private:

    // use a vector of AntennaISR pointers, create them with new on 
    // initialization
    vector<AntennaISR*> antPtrs;

    // Verbose level
    int verbose;

    // Pointers to Vincia classes
    Functions*    functionsPtr;

    // Pointers to Pythia-8 classes, needed to initialise antennae
    bool isInitPtr;
    ParticleData* particleDataPtr;
    Rndm*         rndmPtr;
    Settings*     settingsPtr;  

  }; // End class AntennaSetISR
  
} // End namespace Vincia

#endif

  
