// VinciaWeights.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2016 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the VinciaWeights class.
// Added by NF, so both showers can access and change the weights.
// Initial weight (Pythia's weight * PDF reweighting in case of two 
// event samples) and weight from reweighting due to MC overestimate
// violations.
  
#ifndef VINCIAVinciaWeights_H
#define VINCIAVinciaWeights_H

// Include Pythia 8 headers 
#include "Pythia8/Pythia.h"

#include "Vincia/VinciaCommon.h"

namespace Vincia {
  
  class VinciaWeights {
    
  public:
    
    // Constructor.
    VinciaWeights() {
      weightMC.resize(1);
      weightMC[0] = 1.0;
      weightInit  = 1.0;
    };
  
    // Destructor.
    ~VinciaWeights() { };
    
    //---------------------------------------
    // Initilize Pointers.
    bool initPtr(Settings* settingsPtrIn) {
      settingsPtr = settingsPtrIn; 
      isInitPtr   = true;
      return true;
    }
    
    //---------------------------------------
    // Initilization.
    void init() {
      // Check initPtr
      if (!isInitPtr) {
        printOut("VinciaWeights::init","Error! pointers not initialized"); 
        return;
      }
      
      verbose     = settingsPtr->mode("Vincia:verbose");
      
      nWeights    = (settingsPtr->flag("Vincia:uncertaintyBands") ? 11 : 1);
      weightLimit = settingsPtr->parm("Vincia:uncertaintyLimit");
      
      allowReweighting = settingsPtr->flag("Vincia:allowReweighting");
      
      // Resize weight arrays to the relevant number of elements
      weightMC.resize(nWeights);
      weightOld.resize(nWeights);
      weightSum.resize(nWeights);
      weightSum2.resize(nWeights);
      weightMaxInitMC.resize(nWeights);
      
      // Initialize counters to 0
      // and negative-weight and non-unity warning flags
      nTotWeights         = 0;
      
      nUnityInitialWeight        = 0;
      nNonunityInitialWeight     = 0;
      nNegativeInitialWeight     = 0;
      initialReweightingOccurred = false;
      
      nUnityInitialWeightNow     = 0;
      nNonunityInitialWeightNow  = 0;
      nNegativeInitialWeightNow  = 0;
      
      nUnityShowerWeight        = 0;
      nNonunityShowerWeight     = 0;
      nNegativeShowerWeight     = 0;
      showerReweightingOccurred = false;
      
      // Initialize to 0 (or 1 respectively)
      for (int iVar=0; iVar < nWeights; iVar++) {
        weightMC[iVar]        = 1.0;
        weightOld[iVar]       = 0.0;
        weightSum[iVar]       = 0.0;
        weightSum2[iVar]      = 0.0;
        weightMaxInitMC[iVar] = 0.0;
      }
      weightInit    = 1.0;
      weightMaxInit = 0.0;
      
      firstCall = false;
    }
    
    //---------------------------------------
    // Check if particular weight exists.
    bool existsWeight(int iVarIn=0) {
      if (iVarIn >= 0 && iVarIn < nWeights) return true;
      else return false;
    }

    //---------------------------------------
    // Get weight of current event weight
    // Normally 1 for set 0 (user settings) 
    // Use iVarIn > 0 to access uncertainty weights
    double getWeight(int iVarIn=0) const {
      if (iVarIn >= 0 && iVarIn < nWeights)
        return weightInit * min(weightLimit*weightMC[0],
                                weightMC[iVarIn]);
      else return 0.0;
    }
    double getWeightInit() const {
      return weightInit;
    }
    
    //---------------------------------------
    // Functions to access by VINCIA plugin
    bool weightingOccurred() {
      return initialReweightingOccurred;
    }
    bool reweightingOccurred() {
      return showerReweightingOccurred;
    }
    int nVariations() {
      return nWeights;
    }
    
    //---------------------------------------
    // Set all event weights;
    // NOTE to be called at the beginning of each event by
    // FSR and ISR for system 0 with info.weight()
    void resetWeights(double weightIn, int nAccepted) {
      for (int iVar=0; iVar < nWeights; iVar++) {
        weightMC[iVar]    = 1.0;
        weightOld[iVar]   = 0.0;
      }
      weightInit = weightIn;
      // Next doWeighting call will be the first one
      firstCall = true;
      // Check if Pythia/Vincia vetoed event, restore counters
      if (nAccepted < nTotWeights) {
        nTotWeights--;
        nUnityInitialWeight    -= nUnityInitialWeightNow;
        nNonunityInitialWeight -= nNonunityInitialWeightNow;
        nNegativeInitialWeight -= nNegativeInitialWeightNow;
      }
      nUnityInitialWeightNow    = 0;
      nNonunityInitialWeightNow = 0;
      nNegativeInitialWeightNow = 0;
    }
    
    //---------------------------------------
    // Scale all initial event weights;
    // NOTE to be called by ISR with PDF reweighting factor
    void scaleWeightInit(double scaleFacIn) {
      weightInit *= scaleFacIn;
    }
    
    //---------------------------------------
    // Scale all event weights
    void scaleWeightAll(double scaleFacIn) {
      for (int iVar=0; iVar < nWeights; iVar++)
        weightMC[iVar] *= scaleFacIn;
    }
    
    //---------------------------------------
    // Scale event weight
    void scaleWeight(double scaleFacIn, int iVarIn=0) {
      if (existsWeight(iVarIn))
        weightMC[iVarIn] *= scaleFacIn;
    }
    
    //---------------------------------------
    // Set the weight sum and so on;
    // NOTE to be called after FSR/ISR are done in pTnext
    // might be called several times if we have systems in series, not in parallel
    void doWeighting() {
      // Add weight correction to cumulative sum (!= w0 systems in series)
      double w0       = getWeight(); // = wMC * wInit
      double wOld     = weightOld[0]*weightInit;
      weightSum[0]   += (w0 - wOld);
      weightSum2[0]  += (pow2(w0) - pow2(wOld));
      
      // Check for new max weight (initial weight * MC weight)
      if (w0 > weightMaxInitMC[0]) weightMaxInitMC[0] = w0;
      
      // For first time called,
      if (/*wOld == 0.0*/firstCall) {
        firstCall = false;
        // add one to number of events,
        nTotWeights++;
        // check for initial non-unity weights,
        if (abs(abs(weightInit)-1.0) > TINY) {
          initialReweightingOccurred = true;
          if (nNonunityInitialWeight < 5 && verbose >= 3)
            printOut("VinciaWeights","Nonunity initial weight occurred, w = "+num2str(weightInit)+
                     ((nNonunityInitialWeight == 4) ?  ": further output suppressed" : " "));
          nNonunityInitialWeight++;
          nNonunityInitialWeightNow++;
        } else {
          nUnityInitialWeight++;
          nUnityInitialWeightNow++;
        }
        // check for initial negative weights,
        if (weightInit < 0.0) {
          initialReweightingOccurred = true;
          if (nNegativeInitialWeight < 5 && verbose >= 3)
            printOut("VinciaWeights","Negative initial weight occurred, w = "+num2str(weightInit)+
                     ((nNegativeInitialWeight == 4) ?  ": further output suppressed" : " "));
          nNegativeInitialWeight++;
          nNegativeInitialWeightNow++;
        }
        // check for new max weight (initial weight only)
        if (weightInit > weightMaxInit) weightMaxInit = weightInit;
      }
      
      w0   = weightMC[0];
      wOld = weightOld[0];
      
      // Check for nonunity weights due to MC violations
      if (abs(abs(w0)-1.0) > TINY) {
        if ((abs(abs(wOld)-1.0) < TINY || wOld == 0.0)) {
          showerReweightingOccurred = true;
          if (nNonunityShowerWeight < 5 && ((verbose >= 1 && !allowReweighting) || 
                                            (verbose >= 5 &&  allowReweighting)))
            printOut("VinciaWeights","Nonunity MC weight occurred, w = "+num2str(w0)+
                     ((nNonunityShowerWeight == 4) ?  ": further output suppressed" : " "));
          // Count if old MC weight = 0 (first call) or = 1 (changed to weighted)
          nNonunityShowerWeight++;
          // If we changed to weighted, reduce unity counter
          if (wOld != 0.0) nUnityShowerWeight--;
        }
      } else if (wOld == 0.0) {
        // Count if old MC weight = 0 (first call)
        nUnityShowerWeight++;
      }
      
      // Check for negative weights due to MC violations
      if (w0 < 0.0 && wOld >= 0.0) {
        showerReweightingOccurred = true;
        if (nNegativeShowerWeight < 5 && ((verbose >= 1 && !allowReweighting) || 
                                          (verbose >= 5 &&  allowReweighting)))
          printOut("VinciaWeights","Negative MC weight occurred, w = "+num2str(w0)+
                   ((nNegativeShowerWeight == 4) ?  ": further output suppressed" : " "));
        // NOTE this counter can be > nTot in case of multiple sign changes
        nNegativeShowerWeight++;
      }
      
      // Update old weight
      weightOld[0] = w0;
      
      // Save uncertainty band weights
      if (nWeights>1) {
        for (int iVar=1; iVar<nWeights; iVar++) {
          double wVar    = getWeight(iVar); // = wMC * wInit
          double wVarOld = weightOld[iVar]*weightInit;
          // Add weight correction to cumulative sum
          weightSum[iVar]  += (wVar - wVarOld);
          weightSum2[iVar] += (pow2(wVar) - pow2(wVarOld));
          // Check for new max weight (initial weight * MC weight)
          if (weightInit*weightMC[iVar] > weightMaxInitMC[iVar]) 
            weightMaxInitMC[iVar] = weightInit*weightMC[iVar];
          // Update old weight
          weightOld[iVar] = wVar/weightInit;
        }
      }
      // Done
    }

    int verbose;

    // Pointers
    Settings* settingsPtr; 
    
    // Internal flag
    bool isInitPtr;
    
    // Common parameters
    double nWeights, weightLimit;
    bool allowReweighting;
    
    // Pure weights
    vector<double> weightMC;   // Weights due to MC violations
           double  weightInit; // Initial weight (info.weight()*PDFweight)
    vector<double> weightOld;  // Weights due to MC violations from last step
    // WeightSums
    vector<double> weightSum, weightSum2;
    // Max occured weight (ex- and inclusive MC violations)
    vector<double> weightMaxInitMC;
           double  weightMaxInit;
    
    // Counters for initial weights
    int  nUnityInitialWeight, nNonunityInitialWeight;
    int  nNegativeInitialWeight;
    bool initialReweightingOccurred;
    
    // Counters for current event
    int  nUnityInitialWeightNow, nNonunityInitialWeightNow;
    int  nNegativeInitialWeightNow;
    
    // Counters for weights due to MC violations
    int  nUnityShowerWeight, nNonunityShowerWeight;
    int  nNegativeShowerWeight;
    bool showerReweightingOccurred;
    
    // Counters total number of weights
    int  nTotWeights;
    
    // doWeighting called for the first time?
    bool firstCall;
    
  };

  
}

#endif
