// VRdisplay.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2015 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the VRdisplay class 

// Include the VinciaRootDisplay header
#include "VinciaRoot/VRdisplay.h"

namespace Vincia {
  
  // Init pointers and set settings
  void VRdisplay::init(VRsettings* settingsPtrIn) {
    settingsPtr    = settingsPtrIn;
    logY           = settingsPtr->logY;
    logX           = settingsPtr->logX;
    normalize      = settingsPtr->normalize;
    normFac        = settingsPtr->normFac;
    optStat        = settingsPtr->optStat; 
    verbose        = settingsPtr->verbose;
    sigmaStat      = settingsPtr->sigmaStat;
    hasLeftLabel  = settingsPtr->hasLeftLabel;
    rebinLikeData  = settingsPtr->rebinLikeData;
    zeroSkipping   = settingsPtr->zeroSkipping;
    if (hasLeftLabel) setLeftLabel(settingsPtr->leftLabel);
    hasRightLabel = settingsPtr->hasRightLabel;
    if (hasRightLabel) setRightLabel(settingsPtr->rightLabel);
    has2d          = false;
    nPad           = 1;      
    setShowRatioPad(settingsPtr->showRatioPad);
    setShowRelUnc(settingsPtr->showRelUncPads);
    setMCnames(settingsPtr->MCnames);
    setLegend(settingsPtr->legend);
    setShowLegendChi2(settingsPtr->showLegendChi2);
    setSignature(settingsPtr->signature);
    setLeftLegend(false);
  }
  
  // Add pad (for ratio to data or decomposition of vector of histos)
  void VRdisplay::addPad() {
    nPad++;
  }

  // Goto pad
  void VRdisplay::gotoPad(int iPadIn) {      
    displayCanvasPtr->cd();      
    
    // If pads already defined, just cd to the desired one
    if (isInitPad) {
      gPad->cd(min(iPadIn,nPad));
      return;
    }

    // If pads not yet defined, create them 
    gPad->Divide(1,nPad);
    double nDiv = 4. + nPad;
    double topMargin    = 0.7*gPad->GetTopMargin();
    double bottomMargin = gPad->GetBottomMargin();
    if (nPad >= 2) bottomMargin *= 1.75 + 0.5*nPad;
    double rightMargin  = 0.5*gPad->GetRightMargin();      
    double leftMargin    = (1.15+nPad/9.)*gPad->GetLeftMargin();
    // Left margin at least 35 pixels
    double leftMarginPix = max(35.,leftMargin * gPad->GetWw());
    leftMargin = leftMarginPix/gPad->GetWw();    
    // Left margin is at least 1.5 times right one (for symmetric look)
    leftMargin = max(leftMargin,1.5*rightMargin);
    for (int iPad=1; iPad<=nPad; ++iPad) {
      displayCanvasPtr->cd();        
      gPad->cd(iPad);
      double yHi = 1.0;
      if (iPad >= 2) yHi = 1.0 - double(2+iPad)/nDiv;
      double yLo = 1.0 - double( (iPad+3)/nDiv );
      if (iPad == nPad) yLo = 0.0;
      // Set pad coordinates
      gPad->SetPad(0.0,yLo,1.0,yHi);
      // Set inner margins : top margin
      if (iPad == 1) gPad->SetTopMargin(topMargin);
      else gPad->SetTopMargin(0.0);
      // Set inner margins : bottom margin
      if (iPad == 1 && nPad >= 2) gPad->SetBottomMargin(0.03);
      else if (iPad == nPad) gPad->SetBottomMargin(bottomMargin);
      else gPad->SetBottomMargin(0.0);
      gPad->SetLeftMargin(leftMargin);
      gPad->SetRightMargin(rightMargin);
      // Optionally set log axes
      if (logX) gPad->SetLogx();

      if (!has2d) {
        // Modify labels
        // When adding many pads, increase label font sizes
        if (iPad >= 2) {
          double labSizeX = 1.13 * axesPtr->GetXaxis()->GetLabelSize();
          double labSizeY = 1.13 * axesPtr->GetYaxis()->GetLabelSize();
          axesPtr->GetXaxis()->SetLabelSize(labSizeX);
          axesPtr->GetYaxis()->SetLabelSize(labSizeY);      
          double titSizeX = 1.16 * axesPtr->GetXaxis()->GetTitleSize();
          double titSizeY = 1.16 * axesPtr->GetYaxis()->GetTitleSize();
          axesPtr->GetXaxis()->SetTitleSize(titSizeX);
          axesPtr->GetYaxis()->SetTitleSize(titSizeY);      
          double titOffsetY = 0.9*axesPtr->GetYaxis()->GetTitleOffset();
          axesPtr->GetYaxis()->SetTitleOffset(titOffsetY);      
          // Ratio pad
          double labNormSizeX = 1.09 * axesNormPtr->GetXaxis()->GetLabelSize();
          double labNormSizeY = 1.09 * axesNormPtr->GetYaxis()->GetLabelSize();
          axesNormPtr->GetXaxis()->SetLabelSize(labNormSizeX);
          axesNormPtr->GetYaxis()->SetLabelSize(labNormSizeY);      
          double titNormSizeX = 1.2*axesNormPtr->GetXaxis()->GetTitleSize();
          double titNormSizeY = 1.2*axesNormPtr->GetYaxis()->GetTitleSize();
          axesNormPtr->GetXaxis()->SetTitleSize(titNormSizeX);
          axesNormPtr->GetYaxis()->SetTitleSize(titNormSizeY);      
          double titNormOffsetY = 0.85*axesNormPtr->GetYaxis()->GetTitleOffset();
          axesNormPtr->GetYaxis()->SetTitleOffset(titNormOffsetY);      
        }
        // if there is more than one pad, don't display labels and titles on
        // x axis of first pad
        if ( (iPad >= 2) && (axesPtr) ) {
          axesPtr->GetXaxis()->SetLabelOffset(1.);
          axesPtr->GetXaxis()->SetTitleOffset(1.);
        }
        // RelUnc pad(s)
        if ( (nPad >= 3) && (axesVarPtr) ) {
          double labVarSizeX = 1.75 * axesVarPtr->GetXaxis()->GetLabelSize();
          double labVarSizeY = 1.3 * axesVarPtr->GetYaxis()->GetLabelSize();
          axesVarPtr->GetXaxis()->SetLabelSize(labVarSizeX);
          axesVarPtr->GetYaxis()->SetLabelSize(labVarSizeY);      
          double titVarSizeX = max(1.75 * axesVarPtr->GetXaxis()->GetTitleSize(),
                                   double(axesVarPtr->GetXaxis()->GetLabelSize()));
          double titVarSizeY = 1.3 * axesVarPtr->GetYaxis()->GetTitleSize();
          double titVarOffsetY = 0.7 * axesVarPtr->GetYaxis()->GetTitleOffset();
          axesVarPtr->GetXaxis()->SetTitleSize(titVarSizeX);
          axesVarPtr->GetYaxis()->SetTitleSize(titVarSizeY);      
          axesVarPtr->GetYaxis()->SetTitleOffset(titVarOffsetY);
          isInitPad = false;
        }
      }
      
      // Draw pad
      if (!isInitPad) {
        gPad->Draw();
      } else {
        gPad->Modified();
      }
    }
    isInitPad = true;
    gPad->cd(min(iPadIn,nPad));
  }
  
  // Check if axes and/or data has been defined
  // Modifications to single-pane displays should be done here
  // Modifications specific to multi-pane displays should be done above
  void VRdisplay::setAxes(TH1D* initHis) {
    if (hasAxes()) {
      delete axesPtr; 
      delete xCtrPtr;
      delete xUppPtr;
      delete axesNormPtr;
      delete axesVarPtr;
    }
    displayCanvasPtr->cd();        
    gPad->cd(1);
    axesPtr = new TH1D(*initHis);
    axesPtr->Reset("ICES");
    hasAxesSav = true;
    // Axis histogram is pure dummy, has no stats, so switch off stats display
    axesPtr->SetStats(false);
    // Set fonts and axis tick divisions 
    axesPtr->GetXaxis()->SetLabelFont(42);
    axesPtr->GetYaxis()->SetLabelFont(42);
    axesPtr->GetXaxis()->SetNdivisions(505);
    axesPtr->GetYaxis()->SetNdivisions(606);
    axesPtr->GetXaxis()->SetTitleFont(42);
    double labSizeX = 1.2 * axesPtr->GetXaxis()->GetLabelSize();
    double labSizeY = 1.2 * axesPtr->GetYaxis()->GetLabelSize();
    axesPtr->GetXaxis()->SetLabelSize(labSizeX);
    axesPtr->GetYaxis()->SetLabelSize(labSizeY);
    axesPtr->GetYaxis()->SetTitleSize(1.1*axesPtr->GetYaxis()->GetTitleSize());
    double titOffset = 1.2*axesPtr->GetYaxis()->GetTitleOffset();
    // More space for extreme axis ranges (more digits)
    double axisMin = axesPtr->GetMinimum();
    double axisMax = axesPtr->GetMaximum();
    // Check if asked for log x axes and whether low = 0
    if (gPad->GetLogx() && axisMin == 0.0) {
      axisMin += 0.5*axesPtr->GetBinLowEdge(2);
      axesPtr->SetAxisRange(axisMin, axisMax);
    }
    if ( (min(abs(axisMin),abs(axisMax)) < 1e-10 && min(abs(axisMin),abs(axisMax)) != 0.0) ||
         max(abs(axisMin),abs(axisMax)) > 1e10) titOffset *= 1.2;
    else if ( (min(abs(axisMin),abs(axisMax)) < 1e-2 && min(abs(axisMin),abs(axisMax)) != 0.0) ||
              max(abs(axisMin),abs(axisMax)) > 1e2) titOffset *= 1.1;
    // Reduce/increase offset for extreme aspect ratios
    titOffset *= sqrt(gPad->GetWh()/(1.0*gPad->GetWw()));
    titOffset = min(1.6,titOffset);
    axesPtr->GetYaxis()->SetTitleOffset(titOffset);
    axesPtr->GetYaxis()->SetTitleFont(42);
    // Switch off lame ROOT title
    axesPtr->SetTitle("");
    
    // Set axis histo for ratio pad
    displayCanvasPtr->cd();        
    gPad->cd(nPad);
    axesNormPtr = new TH1D(*axesPtr);
    double ratioMin = settingsPtr->ratioMin;
    double ratioMax = 1.+0.999*(settingsPtr->ratioMax-1);
    axesNormPtr->SetMinimum(ratioMin);
    axesNormPtr->SetMaximum(ratioMax);
    // X axis
    axesNormPtr->GetXaxis()->SetTickLength(0.1);
    axesNormPtr->GetXaxis()->SetLabelSize(1.9*axesPtr->GetXaxis()->GetLabelSize());
    axesNormPtr->GetXaxis()->SetTitleFont(42);
    axesNormPtr->GetXaxis()->SetTitleSize(1.85*axesPtr->GetXaxis()->GetTitleSize());
    axesNormPtr->GetXaxis()->SetTitleOffset(1.*axesPtr->GetXaxis()->GetTitleOffset());
    // Y axis
    axesNormPtr->GetYaxis()->SetLabelFont(42);
    axesNormPtr->GetYaxis()->SetLabelSize(1.8*axesPtr->GetYaxis()->GetLabelSize());
    axesNormPtr->GetYaxis()->SetLabelOffset(0.01);
    axesNormPtr->GetYaxis()->SetTitleFont(42);
    axesNormPtr->GetYaxis()->SetTitleSize(1.8*axesPtr->GetYaxis()->GetTitleSize());
    axesNormPtr->GetYaxis()->SetTitle("Theory/Data");          
    titOffset = 0.57*axesPtr->GetYaxis()->GetTitleOffset();
    axesNormPtr->GetYaxis()->SetTitleOffset(titOffset);
    // Protect against ridiculous aspect ratios
    
    axesNormPtr->GetYaxis()->CenterTitle();
    axesNormPtr->GetYaxis()->SetTickLength(0.025);
    // Make copy of x axis for overdrawing 
    int nBin = axesNormPtr->GetNbinsX();
    xCtrPtr = new TLine(axesNormPtr->GetBinLowEdge(1),1.0,
                        axesNormPtr->GetBinLowEdge(nBin+1),1.0);
    xUppPtr = new TLine(axesNormPtr->GetBinLowEdge(1),ratioMax,
                        axesNormPtr->GetBinLowEdge(nBin+1),ratioMax);
    
    
    // Set axis histo for relUnc pads (uncertainty variations) 
    displayCanvasPtr->cd();        
    gPad->cd(min(nPad,2));
    axesVarPtr = new TH1D(*axesPtr);
    axesVarPtr->SetMinimum(-1.5);
    axesVarPtr->SetMaximum(1.5);
    // X axis
    axesVarPtr->GetXaxis()->SetTickLength(2.*axesPtr->GetXaxis()->GetTickLength());
    axesVarPtr->GetXaxis()->SetLabelOffset(2.*axesPtr->GetXaxis()->GetLabelOffset());
    axesVarPtr->GetXaxis()->SetLabelSize(2.*axesPtr->GetXaxis()->GetLabelSize());
    axesVarPtr->GetXaxis()->SetTitleFont(42);
    axesVarPtr->GetXaxis()->SetTitleSize(2.5*axesPtr->GetXaxis()->GetTitleSize());
    axesVarPtr->GetXaxis()->SetTitleOffset(
                                           1.*axesPtr->GetXaxis()->GetTitleOffset());
    // Y axis
    axesVarPtr->GetYaxis()->SetNdivisions(203);
    axesVarPtr->GetYaxis()->SetTickLength(axesPtr->GetYaxis()->GetTickLength());
    axesVarPtr->GetYaxis()->SetLabelFont(42);
    axesVarPtr->GetYaxis()->SetLabelSize(1.8*axesPtr->GetYaxis()->GetLabelSize());
    axesVarPtr->GetYaxis()->SetLabelOffset(1.8*axesPtr->GetYaxis()->GetLabelOffset());
    axesVarPtr->GetYaxis()->SetTitleFont(42);
    axesVarPtr->GetYaxis()->SetTitle("Rel. Unc.");          
    axesVarPtr->GetYaxis()->SetTitleSize(2*axesPtr->GetYaxis()->GetTitleSize());
    axesVarPtr->GetYaxis()->SetTitleOffset(0.47*axesPtr->GetYaxis()->GetTitleOffset());
    axesVarPtr->GetYaxis()->CenterTitle();
  }
  
  //========================================================================
  // VRdisplay: FILE I/O
  
  // Set file to be treated as experimental data,
  // with i1 : i12 specifying data file format, columnwise (see VINCIA docs)
  bool VRdisplay::setDataFile(string dataFileName, int i1, int i2, int i3, int i4, int i5, int i6, int i7, int i8, int i9, int i10, int i11, int i12,int i13, int i14, int i15, int i16, int i17, int i18, int i19) {     
    
    // Verbose output
    if (verbose >= 3) cout<<"(VRdisplay): adding dataFile "<<dataFileName
                          <<" to display "<<displayNameSav<<endl;
    
    // Create new dataFilePtr with specified format
    vector<int> format;
    getFormat(format,i1,i2,i3,i4,i5,i6,i7,i8,i9,i10,i11,i12,i13,i14,i15,i16,i17,i18,i19);
    dataFilePtr = new VRfile(format);
    
    // Read file 
    bool ok = dataFilePtr->readFile(dataFileName, verbose);
    if (ok) {
      
      // Protect against memory leaks
      if (hasDataSav) {
        delete dataPtr;
        delete dataNormPtr;
        delete dataTwoSigmaPtr;
      }
        
      // Create new VRgraph to contain data points
      dataPtr = new VRgraphData(dataFilePtr);
      dataNormPtr = new VRgraphNorm(dataPtr,dataPtr);
      dataNormPtr->setColorScheme(0);
      // Copies to display two-sigma uncertainties
      dataTwoSigmaPtr = new VRgraphData(dataFilePtr);
      dataTwoSigmaPtr->setColorScheme(-2);      
      dataTwoSigmaNormPtr = new VRgraphNorm(dataTwoSigmaPtr,dataTwoSigmaPtr);
      dataTwoSigmaNormPtr->setColorScheme(-2);      

      // This display has data
      hasDataSav = true;
      if (showRatioPad && nPad <= 1) addPad();

      // Check if rebinning histograms to same binning as data
      if (rebinLikeData) {
        for (int iGraph=0; iGraph<int(graphs.size()); ++iGraph)
          graphs[iGraph].setRebinAxis(dataPtr,zeroSkipping);
        for (int iGraph=0; iGraph<int(graphsHis.size()); ++iGraph)
          graphsHis[iGraph].setRebinAxis(dataPtr,zeroSkipping);
      }
      
    } else {
      cout<<"VRDisplay::setDataFile() error reading "<<dataFileName<<endl;
    }
    
    // Return
    return ok;
  }
  
  //------------------------------------------------------------------------
  // addHisFile() : 
  // Add file with saved histogram,
  // hasOverFlow: true if first and last data file points represent over- and 
  // underflow bins
  // with i1 : i12 specifying the data format, columnwise (see VINCIA docs)
  bool VRdisplay::addHisFile(string hisFileName, int i1, int i2, int i3, int i4, int i5, int i6, int i7, int i8, int i9, int i10, int i11, int i12, int i13, int i14, int i15, int i16, int i17, int i18, int i19) {     
    
    // Verbose output
    if (verbose >= 3) cout<<"(VRdisplay): adding histogram file "<<hisFileName
                          <<" to display "<<displayNameSav<<endl;
    
    // Add to list of hisFilePtrs.
    int nHisFiles = hisFilePtrs.size();
    hisFilePtrs.resize(nHisFiles+1);
    
    // Create new dataFilePtr with specified format
    vector<int> format;
    getFormat(format,i1,i2,i3,i4,i5,i6,i7,i8,i9,i10,i11,i12,i13,i14,i15,i16,i17,i18,i19);
    hisFilePtrs[nHisFiles] = new VRfile(format);
    
    // Read file and return success/failure
    bool ok = hisFilePtrs[nHisFiles]->readFile(hisFileName, verbose);
    if (ok) {
      
      // Add to list of graphsHis
      int nGraphsHis = graphsHis.size();
      graphsHis.resize(nGraphsHis+1);
      
      // Create new VRgraph to contain histogram points
      graphsHis[nGraphsHis] = VRgraphHis( hisFilePtrs[nHisFiles] );     
      graphsHis[nGraphsHis].setColorScheme(graphs.size()+graphsHis.size());

      // Store name
      hisFileNames.resize(nGraphsHis+1);
      hisFileNames[nGraphsHis] = hisFileName;

      // Check if rebinning main histogram to data 
      if (rebinLikeData && hasDataSav && dataPtr != 0) 
        graphsHis[nGraphsHis].setRebinAxis(dataPtr,zeroSkipping);
      
      // Recreate list of normalized histograms
      // (involves pointers to elements of a vector so must be explicit)
      graphsHisNorm.resize(0);
      for (int i=0; i<int(graphsHis.size()); ++i) {
        VRgraphHis* hisPtr = &graphsHis[i];
        if (hasDataSav) {
          graphsHisNorm.push_back( VRgraphNorm(hisPtr,dataPtr) );
        } else if (graphs.size() >= 1) {
          graphsHisNorm.push_back(VRgraphNorm(hisPtr,&graphs[0]));
        } else {
          graphsHisNorm.push_back(VRgraphNorm(hisPtr,&graphsHis[0]));
        }
        graphsHisNorm[i].setColorScheme(graphs.size()+i+1);
      }
      
      // If there is more than one histogram, we can have a ratio pad
      if (graphsHis.size() + graphs.size() >= 2 
          && showRatioPad && nPad <= 1) addPad();
      
      // Add pad for decomposition of vector band
      if (nPad <= 2 && showRelUncPads) {
        for (int i=0; i<int(graphs.size()); ++i) 
          if (graphs[i].hasOuter()) addPad();
      }
      
    } else {
      cout<<"VRDisplay::addHisFile() error reading "<<hisFileName<<endl;
      delete hisFilePtrs[nHisFiles];
      hisFilePtrs.pop_back();
    }
    
    // Return
    return ok; 
    
  }

  //------------------------------------------------------------------------
  // Force update of histogram data files, useful when doing several
  // runs in parallel.
  bool VRdisplay::refreshHisFiles() {
    bool ok = true;
    for (int i=0; i<(int)hisFileNames.size(); ++i) 
      ok = ok && hisFilePtrs[i]->readFile(hisFileNames[i], verbose);
    return ok;
  }

  //------------------------------------------------------------------------
  // update(): 
  // Update Display 
  void VRdisplay::update() {

    // Refresh any histogram files in case they changed since last update
    refreshHisFiles();
    
    // CD to this display
    displayCanvasPtr->cd();        
    
    // If several pads, first do main plot
    gotoPad(1);
    
    // First check if this is 2D, then plot and exit. Rest is 1D.
    if (has2d) {
      // Draw with colored contours
      th2dPtr->Draw("colz");
      // Then update canvas
      displayCanvasPtr->cd();
      displayCanvasPtr->Update();      
    }

    // Update and (re)draw axis
    if (!hasAxes()) {
      if (graphs.size() >= 1) setAxes(graphs[0].getTH1DPtr());
      else return;
    }
    
    // Update max/min if possible.
    if (graphs.size() >= 1) {
      TH1D* hisPtr = graphs[0].getTH1DPtr();
      if (hisPtr != NULL) {
        // Check if min/max changed by user
        double yMinSav = axesPtr->GetMinimum();
        double yMaxSav = axesPtr->GetMaximum();
        double yMinNow = hisPtr->GetMinimum();
        double yMaxNow = hisPtr->GetMaximum();
        if (yMinSav != yMinNow || yMaxSav != yMaxNow) {
          axesPtr->SetMinimum(yMinNow);
          axesPtr->SetMaximum(yMaxNow);
        }
        
        // Also update log Y axis properties (force linear if yMin <= 0.0)
        if ( yMinNow <= 0.0 && gPad->GetLogy() ) setLogY(false); 
        else if ( yMinNow > 0.0 && logY && !gPad->GetLogy() ) setLogY(true);
        
        // Update title 
        setTitle();
      }
    }
    
    // Draw axes
    if (nPad >= 2) {
      // Switch off X axis title for main histo
      axesPtr->GetXaxis()->SetTitle("");
      // Draw relUnc pads
      if (showRelUncPads) {
        int iPad = 1;
        for (int iGraph = 0; iGraph < int(graphs.size()); ++iGraph) {
          if (graphs[iGraph].hasOuter()) {
            ++iPad;
            displayCanvasPtr->cd(iPad);
            axesVarPtr->Draw();
          }
        }
      }
      // Draw ratio pad
      if (showRatioPad && hasDataSav && dataPtr) {
        displayCanvasPtr->cd(nPad);
        axesNormPtr->Draw();        
      } 
      else if (showRatioPad && nPad >= 2) {
        displayCanvasPtr->cd(nPad);
        axesNormPtr->Draw();        
      }        
    }
    // Draw main display
    displayCanvasPtr->cd(1);
    axesPtr->Draw();      
    
    // Update data 
    // (data is always shown with 1 sigma error bars)
    double ratioMin = axesNormPtr->GetMinimum();
    double ratioMax = axesNormPtr->GetMaximum();
    if (hasDataSav && dataPtr) {
      displayCanvasPtr->cd(1);
      dataTwoSigmaPtr->update(normalize,normFac,2.0);
      dataTwoSigmaPtr->DrawOuter("E0Z");
      dataPtr->update(normalize,normFac,1.0);
      dataPtr->DrawOuter("E10");
      dataPtr->Draw("PE1");
      if (dataNormPtr && nPad >= 2 && showRatioPad) {
        displayCanvasPtr->cd(nPad);
        dataTwoSigmaNormPtr->update(true,1.0,ratioMin,ratioMax);
        dataTwoSigmaNormPtr->DrawOuter("e2");
        dataTwoSigmaNormPtr->Draw("e2");
        dataNormPtr->update(true,1.0,ratioMin,ratioMax);
        dataNormPtr->DrawOuter("e2");
        dataNormPtr->Draw("e2");
        axesNormPtr->Draw("sameaxis");
        //gPad->RedrawAxis();
        // Draw center line at 1
        xCtrPtr->Draw();
      }
    }
    
    // Update graphs made from saved histograms (dat files)
    displayCanvasPtr->cd(1);
    // Determine plot range 
    double xMinSav = axesPtr->GetBinLowEdge(1);
    int nBins = axesPtr->GetNbinsX();
    double xMaxSav = axesPtr->GetBinLowEdge(nBins) 
      + axesPtr->GetBinWidth(nBins);
    for (unsigned int iGraph=0; iGraph<graphsHis.size(); ++iGraph) {
      
      // Set X scatter
      double scatterX = 0.0;
      // 0.1*pow(-1.0,(iGraph+graphs.size())*1.0) 
      //  * (iGraph+graphs.size())/(graphsHis.size()+graphs.size());
      
      // Skip if update fails
      if (!graphsHis[iGraph].update(normalize,normFac,sigmaStat,scatterX)) 
        continue; 

      // Remove points outside plot region (to avoid lines going outside region)
      for (int iPoint = graphsHis[iGraph].GetN()-1; iPoint >= 0; --iPoint) {
        double x, y;
        graphsHis[iGraph].GetPoint(iPoint,x,y);
        if (x < xMinSav || x > xMaxSav) graphsHis[iGraph].RemovePoint(iPoint);
      }

      // Outer band
      if (graphsHis[iGraph].hasOuter()) {
        int fillStyle = 3354;
        if (iGraph == 1) fillStyle = 3345;
        else if (iGraph >= 2) fillStyle = 3307;
        graphsHis[iGraph].getOuter()->SetFillStyle(fillStyle);
      }
      
      // Draw graph
      graphsHis[iGraph].DrawOuter("20");
      graphsHis[iGraph].Draw("LPZ0");
    }
    
    // Update graphs
    displayCanvasPtr->cd(1);
    for (unsigned int iGraph=0; iGraph<graphs.size(); ++iGraph) {

      // Set X scatter
      double scatterX = 0.0;
      // 0.1*pow(-1.0,iGraph*1.0) * iGraph/graphs.size();
      
      // Skip if update fails
      if (!graphs[iGraph].update(normalize,normFac,sigmaStat,scatterX)) {
        continue; 
      }
      
      // Outer band
      if (graphs[iGraph].hasOuter()) {
        int fillStyle = 3354;
        if (iGraph == 1) fillStyle = 3345;            
        else if (iGraph >= 2) fillStyle = 3307;
        graphs[iGraph].getOuter()->SetFillStyle(fillStyle);                    
      }
      
      // Draw graph, first one with systematic errors as a box
      if (iGraph == 0) graphs[iGraph].DrawOuter("20");    
      else graphs[iGraph].DrawOuter("20");    
      graphs[iGraph].Draw("LPZ0");
    }
    
    // Update normalized graphs
    bool hasRatioPad = false;
    if (nPad >=2 && showRatioPad) {           
      hasRatioPad = true;
      displayCanvasPtr->cd(nPad);
      if (graphsNorm.size() < graphs.size()) {
        graphsNorm.resize(0);
        for (unsigned int iGraph=0; iGraph<graphs.size(); ++iGraph) {
          // Normalize to data
          if (dataNormPtr && hasDataSav)
            graphsNorm.push_back(VRgraphNorm(&graphs[iGraph],dataPtr));  
          // Normalize to first histogram
          else {
            graphsNorm.push_back(VRgraphNorm(&graphs[iGraph],&graphs[0]));
            axesNormPtr->GetYaxis()->SetTitle("Ratio");          
          }
          graphsNorm[iGraph].setColorScheme(&graphs[iGraph]);
          if (graphsNorm[iGraph].hasOuter()) {
            int fillStyle = 3354;
            if (iGraph == 1) fillStyle = 3345; 
            else if (iGraph >= 2) fillStyle = 3307;
            graphsNorm[iGraph].SetFillStyle(fillStyle);            
          }
        }
      }
      // Draw saved histo ratio graphs
      for (unsigned int iGraph=0; iGraph<graphsHis.size(); ++iGraph) {
        // Update
        graphsHisNorm[iGraph].update(true,1.,ratioMin,ratioMax);
        // Skip if no points to be plotted
        if (graphsHisNorm[iGraph].GetN() <= 0) continue;
        // Draw graph
        graphsHisNorm[iGraph].DrawOuter("20");
        graphsHisNorm[iGraph].Draw("LPZ0");
      }
      // Draw MC ratio graphs
      for (unsigned int iGraph=0; iGraph<graphs.size(); ++iGraph) {
        // Update
        graphsNorm[iGraph].update(true,1.,ratioMin,ratioMax);
        // Skip if no points to be plotted
        if (graphsNorm[iGraph].GetN() <= 0) continue;
        graphsNorm[iGraph].DrawOuter("20");
        graphsNorm[iGraph].Draw("LPZ0");
      }
      // Redraw upper boundary of ratio plot in case hidden by fill styles
      xUppPtr->Draw();
    }
    
    // Update relative uncertainty variation pads
    int nUncVarPads = nPad - hasRatioPad -1;
    if (nUncVarPads >= 1) {
      int iPad = 1;
      for (int iGraph = 0; iGraph < int(graphs.size()); ++iGraph) {
        // Only do RelUnc pads for graphs that have (>1) outer bands
        if (!graphs[iGraph].hasOuter() || graphs[iGraph].nOuter() <= 1) 
          continue;
        ++iPad;
        // Sanity check: allow for max 3 separate RelUnc pads
        if (iPad > nPad || iPad >= 3) continue;
        displayCanvasPtr->cd(iPad);
        // Create vector if not already done
        if (graphsVar.size() <= 1) {
          graphsVar.resize(0);
          for (int iVar=1; iVar<=graphs[iGraph].nOuter(); ++iVar) {
            graphsVar.push_back(VRgraphVar(&graphs[iGraph],iVar));
            graphsVar[graphsVar.size()-1].setStyle(iVar);
          }
        }
        for (int iVar=0; iVar<int(graphsVar.size()); ++iVar) {
          graphsVar[iVar].update();
          // Skip if no points to be plotted
          if (graphsVar[iVar].GetN() <= 0) continue;
          // Draw area-filled graphs 
          if (graphsVar[iVar].GetFillStyle() != 0) 
            graphsVar[iVar].Draw("e3line");
          // Draw line graphs
          else {
            graphsVar[iVar].Draw("Xline");
          }
        }
      }
    }

    // Draw objects 
    // 1: lines
    displayCanvasPtr->cd(1);
    if (linesPtr.size() >= 1) {
      for (unsigned int iLine=0; iLine < linesPtr.size(); ++iLine)           
        linesPtr[iLine]->Draw();      
    }
    
    // Draw title
    displayCanvasPtr->cd(1);      
    if (hasTitle) titleTexPtr->Draw();
    
    // Draw MC names
    displayCanvasPtr->cd(1);
    if (!hasMCnames) setMCnames(settingsPtr->MCnames);
    if (hasMCnames) MCnamesPtr->Draw();
    
    // Draw data ref
    displayCanvasPtr->cd(1);
    if (hasDataSav) {
      if (!hasDataRef && dataFilePtr->getREF() != "none")
        setDataRef("Data from "+dataFilePtr->getREF());
    }
    if (hasDataRef) dataRefPtr->Draw();
    
    // Draw legend (so it appears on top of curves)
    displayCanvasPtr->cd(1);
    if (!hasLegend) {
      if (forceAlign) setLegend(settingsPtr->legend,alignSav);
      else setLegend(settingsPtr->legend);
    } else if (alignSav == 0) setLegend(legendSav,0);
    else {
      if (forceAlign) setLegend(legendSav,alignSav);
      else setLegend(legendSav);
    }
    if (hasLegend) {
      legendPtr->Draw();  
      // Optionally also draw chi2 legend 
      if (hasLegendChi2 && showLegendChi2) legendChi2Ptr->Draw(); 
    }
    
    // Draw legend on uncertainty band plots
    if (nUncVarPads >= 1) {
      displayCanvasPtr->cd(2);
      if (!hasLegendUncVar) setLegendUncVar(settingsPtr->legendUncVar);
      if (hasLegendUncVar) {
        legendUncVarTopPtr->Draw();
        legendUncVarBotPtr->Draw();
      }
    }
    
    // Draw signature
    displayCanvasPtr->cd(1);
    if (!hasSignature) {
      double xR = 1.-0.9*gPad->GetRightMargin();      
      double yR = gPad->GetBottomMargin();
      if (nPad >= 2) yR = 0.03;
      signaturePtr = new TLatex(xR,yR,signature.c_str());
      signaturePtr->SetNDC(true);
      signaturePtr->SetTextAngle(90);
      signaturePtr->SetTextAlign(13);
      signaturePtr->SetTextColor(12);
      signaturePtr->SetTextFont(42);
      double sigSize = 0.02 + 0.003*nPad;
      signaturePtr->SetTextSize(sigSize);
      hasSignature = true;
    }
    signaturePtr->Draw();
    
    // Draw left and right
    displayCanvasPtr->cd(1);
    if (hasLeftLabel || hasRightLabel) {      
      double xL = 0.+gPad->GetLeftMargin();      
      double xR = 1.-0.95*gPad->GetRightMargin();      
      double yT = 1.0;
      double yB = 1. - gPad->GetTopMargin();
      double yC = 0.5 * (yT + yB);
      if (hasTopBox) delete topBoxPtr;
      hasTopBox = true;
      topBoxPtr = new TPave(xL, yT, xR, yB,0,"NDC");
      topBoxPtr->SetFillColor(kBlack);      
      topBoxPtr->Draw();
      
      // Left label (typically beam)
      if (hasLeftLabel) {
        double xLm = xL + 0.02 * (xR-xL);
        if (hasLeftLabelPtr) delete leftLabelPtr;
        leftLabelPtr = new TLatex(xLm, yC, leftLabel.c_str());
        leftLabelPtr->SetNDC(kTRUE);
        leftLabelPtr->SetTextAlign(12);
        leftLabelPtr->SetTextColor(kWhite);
        leftLabelPtr->SetTextSize(0.055);
        leftLabelPtr->SetTextFont(62);
        leftLabelPtr->Draw();
      }
      
      // Right label (typically energy)
      if (hasRightLabel) {
        double xRm = xR - 0.02 * (xR-xL);
        if (hasRightLabelPtr) delete rightLabelPtr;
        rightLabelPtr = new TLatex(xRm, yC, rightLabel.c_str());
        rightLabelPtr->SetNDC(kTRUE);
        rightLabelPtr->SetTextAlign(32);
        rightLabelPtr->SetTextColor(kWhite);
        rightLabelPtr->SetTextSize(0.05);
        rightLabelPtr->SetTextFont(62);
        rightLabelPtr->Draw();
      }
    }

    // Then update canvas
    displayCanvasPtr->cd();
    displayCanvasPtr->Update();      

  }

  //------------------------------------------------------------------------
  // Save display image
  void VRdisplay::saveDisplay(string prefix, string fileType) {
    // Default file name is the name of display
    string fileName = displayNameSav;
    // If specific prefix given, append display name to that 
    if (prefix != "none") fileName = prefix+"-"+displayNameSav;          
    // First loop for specific file types
    if (fileType.find("eps") != string::npos) fileName+=".eps";
    else if (fileType.find("ps") != string::npos) fileName+=".ps";
    else if (fileType.find("Portrait") != string::npos) fileName+=".ps";
    else if (fileType.find("Landscape") != string::npos) fileName+=".ps";
    else if (fileType.find("gif") != string::npos) fileName+=".gif";
    else if (fileType.find("jpg") != string::npos) fileName+=".jpg";
    else if (fileType.find("png") != string::npos) fileName+=".png";
    else if (fileType.find("pdf") != string::npos) fileName+=".pdf";
    else if (fileType.find("tex") != string::npos) fileName+=".tex";
    else if (fileType.find("svg") != string::npos) fileName+=".svg";
    else if (fileType.find("xpm") != string::npos) fileName+=".xpm";
    else if (fileType.find("cxx") != string::npos) fileName+=".cxx";
    else if (fileType.find("xml") != string::npos) fileName+=".xml";
    else if (fileType.find("root") != string::npos) fileName+=".root";
    // Then try generic extension. Don't add dot if ext already has one.
    else if (fileType.find(".") != string::npos) fileName+=fileType;
    else fileName+="."+fileType;
    displayCanvasPtr->Print(fileName.c_str(),fileType.c_str());
  }    
  
  //------------------------------------------------------------------------
  // saveHistogram():
  bool VRdisplay::saveHistogram(string prefix, string ext) {
    
    // Default file name is the name of display
    string fileName = displayNameSav;
    // If specific prefix given, append display name to that 
    if (prefix != "none") fileName = prefix+"-"+displayNameSav;          
    fileName += "."+ext;
    FILE * pFile = fopen (fileName.c_str(), "w");
    if (pFile == NULL) {
      perror ("(VRdisplay::saveHistogram): Error opening file");
      return false;
    }
    else {
      // Number of bins to print
      TH1D* hisPtr = getTH1DPtr();
      int nBin = hisPtr->GetNbinsX();
      // Optional Normalization factor
      double normFactor = hisPtr->GetNormFactor();
      if (normFactor == 0.0) normFactor = 1.0;
      for (int iBin=0; iBin<=nBin+1; iBin++){
        double dx, xLo, xHi;
        // Underflow bin (save displaced by one)    
        if (iBin == 0) {
          dx  = hisPtr->GetBinWidth(1);
          xLo = hisPtr->GetBinLowEdge(1)-2*dx;
        } 
        // Overflow bin (save displaced by one)
        else if (iBin == nBin+1) {
          dx  = hisPtr->GetBinWidth(nBin);
          xLo = hisPtr->GetBinLowEdge(nBin)+2*dx;
        }
        // Ordinary bins
        else {
          dx  = hisPtr->GetBinWidth(iBin);
          xLo = hisPtr->GetBinLowEdge(iBin);
        }
        if ( abs(xLo/dx) < 1e-12 ) xLo = 0.;
        xHi = xLo + dx;
        if ( abs(xHi/dx) < 1e-12 ) xHi = 0.;
        double xC  = xLo + (xHi - xLo)/2.;
        double y   = normFactor * hisPtr->GetBinContent(iBin);
        // Output standard 1-sigma uncertainties
        double ey  = normFactor * hisPtr->GetBinError(iBin);
        // Systematic errors ?
        double eyThHi(0.), eyThLo(0.);
        if (graphs[0].hasOuter()) {
          for (int iVar=0; iVar < graphs[0].nOuter(); ++iVar) {
            eyThHi = max(eyThHi,
                         normFactor * getTH1DPtr(0,iVar)->GetBinContent(iBin) - y);
            eyThLo = max(eyThLo,
                         y - normFactor * getTH1DPtr(0,iVar)->GetBinContent(iBin));
          }
        }
        
        // If has systematic uncertainties
        if (graphs[0].hasOuter()) {
          fprintf(pFile,
                  "%14.6g %14.6g %14.6g %14.6g %14.6g %14.6g %14.6g %14.6g \n",
                  xLo,xC,xHi,y,ey,ey,eyThHi,eyThLo); 
        } else {
          // Default: write out stat error only
          fprintf(pFile,"%14.6g %14.6g %14.6g %14.6g %14.6g %14.6g \n",
                  xLo,xC,xHi,y,ey,ey); 
          
        }
      }
      
      // Write format as text comment
      if (graphs[0].hasOuter()) {
        fprintf(pFile,"#          xLo             xC            xHi              y        +dyStat        -dyStat          +dyTh          -dyTh\n");
      } 
      else {
        fprintf(pFile,"#          xLo             xC            xHi              y        +dyStat        -dyStat\n");
      }
      fprintf(pFile,"# Note! The first and last bins represent underflow and overflow, respectively (in-\n");
      fprintf(pFile,"# cluding an artificial offset by one bin width from the main histogram bins).\n");
      
      fprintf(pFile,"# Info:\nTIT = %s\n",hisPtr->GetTitle());
      if (hasLegend && legendSav.size() >= 1) {
        fprintf(pFile,"SRC = %s\n",legendSav[0].c_str());
      }
      if (hasMCnames && MCnamesSav != "") {
        fprintf(pFile,"REF = %s\n",MCnamesSav.c_str());
      }
      fprintf(pFile,"#\n# File printed by VinciaRoot\n");        
      fprintf(pFile,"# http://arxiv.org/abs/arXiv:1102.2126\n");
      fprintf(pFile,"# VinciaRoot is distributed as part of the VINCIA code\n"); 
      fprintf(pFile,"# http://vincia.hepforge.org"); 
      fclose(pFile);
      return true;
    }
    
  }
  
  //------------------------------------------------------------------------
  // Print out chi2 values
  void VRdisplay::printChi2() {
    cout<<" "<<setw(18)<<displayNameSav<<"   <chi2(5%)>  = ";
    for (int iGraph=0; iGraph<int(graphsNorm.size()); ++iGraph) 
      cout<<setw(10)<<graphsNorm[iGraph].getChi2()<<" ";
    for (int iGraph=0; iGraph<int(graphsHisNorm.size()); ++iGraph) 
      cout<<setw(10)<<graphsHisNorm[iGraph].getChi2()<<" ";
    cout<<endl;
  }

  //========================================================================
  // VRdisplay: Plot titles, legends, descriptions
  
  //------------------------------------------------------------------------
  // setTitle():
  // Set title
  void VRdisplay::setTitle() {
    // Can only define title if histogram has been defined
    if (!hasAxes() || graphs.size() == 0) return;
    // Get current histogram title from histogram pointer
    TH1D* hisPtr = graphs[0].getTH1DPtr();
    string hisTitle = hisPtr->GetTitle();
    // Destroy pointer if title already exists (to avoid mem leak)
    if (hasTitle) delete titleTexPtr;
    // Get axis definitions from axis histogram
    int NBins = axesPtr->GetNbinsX();
    double xLo = axesPtr->GetBinLowEdge(1);
    double xHi = axesPtr->GetBinLowEdge(NBins+1);
    double xTit = xLo + (xHi-xLo)/2.;
    if (gPad->GetLogx()) {
      if (xLo == 0.0) xLo += 0.5 * axesPtr->GetBinLowEdge(2);
      double lLo  = log(xLo);
      double lHi  = log(xHi);
      double lTit = lLo + (lHi - lLo)/2.;
      xTit = exp(lTit);
    }
    double yMin = axesPtr->GetMinimum();
    double yMax = axesPtr->GetMaximum();
    double dyTit = 0.07;
    double yTit = yMin + (1.-dyTit)*(yMax-yMin);
    if (gPad->GetLogy()) {
      double LnyC = log(yMin) + (1.-dyTit)*log(yMax/yMin);
      yTit = exp(LnyC);
    }      
    // Create new title tex pointer
    titleTexPtr = new TLatex(xTit,yTit,hisTitle.c_str());
    hasTitle    = true;
    titleTexPtr->SetTextFont(42);
    titleTexPtr->SetTextAlign(22);
    double titSize = (1.0+0.028*(nPad-1))*titleTexPtr->GetTextSize();
    titleTexPtr->SetTextSize(titSize);
    titleTexPtr->SetTextColor(1);
  }

  //------------------------------------------------------------------------
  // Set MC name title
  void VRdisplay::setMCnames(string MCnamesIn) {
    if (!hasAxes() || graphs.size() == 0) return;
    // Destroy pointer if MCnames already exists (to avoid mem leak)
    if (hasMCnames) delete MCnamesPtr;
    // Get axis definitions from axis histogram
    int NBins = axesPtr->GetNbinsX();
    double xLo = axesPtr->GetBinLowEdge(1);
    double xHi = axesPtr->GetBinLowEdge(NBins+1);
    double xTit = xLo + (xHi-xLo)/2.;
    if (gPad->GetLogx()) {
      if (xLo == 0.0) xLo += 0.5 * axesPtr->GetBinLowEdge(2);
      double lLo  = log(xLo);
      double lHi  = log(xHi);
      double lTit = lLo + (lHi - lLo)/2.;
      xTit = exp(lTit);
    }
    double yMin = axesPtr->GetMinimum();
    double yMax = axesPtr->GetMaximum();
    double dyTit = 0.085;
    double yTit = yMin + dyTit*(yMax-yMin);
    if (gPad->GetLogy()) {
      double LnyC = log(yMin) + dyTit*log(yMax/yMin);
      yTit = exp(LnyC);
    }      
    MCnamesPtr = new TLatex(xTit,yTit,MCnamesIn.c_str());
    MCnamesPtr->SetTextFont(42);
    MCnamesPtr->SetTextAlign(23);
    MCnamesPtr->SetTextColor(12);
    double namesSize = MCnamesPtr->GetTextSize();
    double scaleFac  = 1. + (nPad-1)/10.;
    MCnamesPtr->SetTextSize(0.7*scaleFac*namesSize);
    hasMCnames = true;
    MCnamesSav = MCnamesIn;
  }
  
  //------------------------------------------------------------------------
  // Set Data reference title
  void VRdisplay::setDataRef(string dataRefIn) {
    if (!hasAxes() || graphs.size() == 0) return;
    // Destroy pointer if MCnames already exists (to avoid mem leak)
    if (hasDataRef) delete dataRefPtr;
    if (dataRefIn == "none") return;
    // Get axis definitions from axis histogram
    int NBins = axesPtr->GetNbinsX();
    double xLo = axesPtr->GetBinLowEdge(1);
    double xHi = axesPtr->GetBinLowEdge(NBins+1);
    double xTit = xLo + (xHi-xLo)/2.;
    if (gPad->GetLogx()) {
      if (xLo == 0.0) xLo += 0.5 * axesPtr->GetBinLowEdge(2);
      double lLo  = log(xLo);
      double lHi  = log(xHi);
      double lTit = lLo + (lHi - lLo)/2.;
      xTit = exp(lTit);
    }
    double yMin = axesPtr->GetMinimum();
    double yMax = axesPtr->GetMaximum();
    double dyTit = 0.112;
    double yTit = yMin + dyTit*(yMax-yMin);
    if (gPad->GetLogy()) {
      double LnyC = log(yMin) + dyTit*log(yMax/yMin);
      yTit = exp(LnyC);
    }
    dataRefPtr = new TLatex(xTit,yTit,dataRefIn.c_str());
    dataRefPtr->SetTextFont(42);
    dataRefPtr->SetTextAlign(21);
    dataRefPtr->SetTextColor(12);
    double namesSize = dataRefPtr->GetTextSize();
    double scaleFac  = 1. + (nPad-1)/10.;
    dataRefPtr->SetTextSize(0.7*scaleFac*namesSize);
    hasDataRef = true;
  }
  
  //------------------------------------------------------------------------
  // Set Legend (align: -1: left, 0: auto, 1:right)
  void VRdisplay::setLegend(vector<string> legendIn, int alignIn) {

    // Sanity check
    if (!hasAxes() || graphs.size() == 0 || legendIn.size() == 0) return;

    // Determine left/right placement
    alignSav = alignIn;
    if (alignIn == -1 || alignIn == 1) forceAlign = true;
    
    // Decide automatic placement based on where histogram is largest
    if (alignIn == 0) {
      double sumLeft  = 0;
      double sumRight = 0;
      // Find largest bin of data histogram (or first MC one if no data) 
      int axisBins   = axesPtr->GetNbinsX();
      double axisMin = axesPtr->GetBinLowEdge(1);
      double axisMax = axesPtr->GetBinLowEdge(axisBins+1);
      double axisCtr = axisMin + 0.5*(axisMax - axisMin);
      if (gPad->GetLogx()) {
        if (axisMin == 0.0) axisMin = 0.5*axesPtr->GetBinLowEdge(2); 
        double lAxisMin = log(axisMin);
        double lAxisMax = log(axisMax);
        double lAxisCtr = lAxisMin + 0.5*(lAxisMax - lAxisMin);
        axisCtr = exp(lAxisCtr);
      }
      double xMax = axisMin - 1;
      double yMax = -1e9;          
      if (hasDataSav && dataPtr->GetN() >= 1) {
        int nPoints = dataPtr->GetN();
        double xTmp,yTmp;
        for (int iPoint = 0; iPoint < nPoints; iPoint++) {          
          dataPtr->GetPoint(iPoint,xTmp,yTmp);
          // Do not count a point exactly in center
          if (xTmp == axisCtr) continue;
          if (yTmp > yMax) {
            yMax = yTmp;
            xMax = xTmp;
          }
        }
        // Let data count for 2 graphs:        
        if (xMax > axisCtr) sumLeft += 2;
        else if (xMax >= axisMin) sumRight += 2;
      }      
      if (graphs.size() >= 1) {
        for (int iGraph=0; iGraph < int(graphs.size()); ++iGraph) {
          int nPoints = graphs[iGraph].GetN();
          double xTmp,yTmp;
          // Initialize xMax to sit below axis 
          xMax = axisMin - 1;
          for (int iPoint = 0; iPoint < nPoints; iPoint++) {          
            graphs[0].GetPoint(iPoint,xTmp,yTmp);
            // Do not count a point exactly in center
            if (xTmp == axisCtr) continue;
            if (yTmp > yMax) {
              yMax = yTmp;
              xMax = xTmp;
            }
          }
          double w = 1.0;
          // First graph has weight 1.5 to split degeneracies
          if (iGraph == 0) w = 1.5;
          if (xMax > axisCtr) sumLeft += w;
          else if (xMax >= axisMin) sumRight += w;
        }
      }
      // Set legend position Left or Right (default right if all == 0)
      if ( sumLeft > sumRight ) alignSav = -1;
      else alignSav = 1;      
    }
    
    // Set legend placement: right or left
    double xLeg = 0.48;
    if (!hasDataSav || !dataPtr) xLeg = 0.60;
    if (alignSav == -1) xLeg = 0.18;

    // Call setLegend with default settings
    setLegend(legendIn, xLeg);
    
  }

  void VRdisplay::setLegend(vector<string> legendIn, double xLeg, 
                            double xWid, double dy, double yLegTop) {

    // Save legend for replotting
    legendSav  = legendIn;
    xLegSav    = xLeg;
    // Allow wider legend if not showing chi2 values
    if (!showLegendChi2) xWid += 0.13;
    xWidSav    = xWid;
    dySav      = 2*dy;
    yLegTopSav = yLegTop;

    if (!hasAxes() || graphs.size() == 0 || legendIn.size() == 0) return;
    // nGraphs
    int nGraphsLeg = min(graphs.size(),legendIn.size());
    // Destroy pointer if MCnames already exists (to avoid mem leak)
    if (hasLegend) delete legendPtr;
    double xLegEnd = xLeg + xWid; 
    double yLegBot = yLegTop - dy*(nGraphsLeg+hasDataSav+graphsHis.size());
    legendPtr = new TLegend(xLeg,yLegBot,xLegEnd,yLegTop,NULL,"rNDC");
    hasLegend = true;
    legendPtr->SetBorderSize(0);
    legendPtr->SetFillStyle(0);
    // Add data legend
    if (hasDataSav) {
      string datLeg = dataFilePtr->getSRC();
      if (datLeg == "" || datLeg == "none") datLeg = "Data";
      TLegendEntry *entry=legendPtr->AddEntry("dat",datLeg.c_str(),"p");
      entry->SetMarkerStyle(dataPtr->GetMarkerStyle());
      entry->SetMarkerSize(dataPtr->GetMarkerSize());
      entry->SetMarkerColor(dataPtr->GetMarkerColor());
      entry->SetTextAlign(12);
      entry->SetTextFont(42);
      entry->SetTextColor(1);	
    }
    // Add histogram legends
    int nLeg = 0;
    for (int iLeg = 0; iLeg < min(int(graphs.size()), int(legendIn.size()));
         ++iLeg) {
      ++nLeg;
      TLegendEntry *entry=legendPtr->AddEntry("mc",legendIn[iLeg].c_str(),"lP");
      entry->SetLineColor(graphs[iLeg].GetLineColor());
      entry->SetLineStyle(graphs[iLeg].GetLineStyle());
      entry->SetLineWidth(graphs[iLeg].GetLineWidth());
      entry->SetMarkerStyle(graphs[iLeg].GetMarkerStyle());
      entry->SetMarkerSize(graphs[iLeg].GetMarkerSize());
      entry->SetMarkerColor(graphs[iLeg].GetMarkerColor());
      entry->SetTextAlign(12);
      entry->SetTextFont(42);
      entry->SetTextColor(1);
    }
    // Add legends from saved histograms
    int nLegTot = nLeg;
    for (int iLeg = 0; iLeg < int(graphsHis.size()); ++iLeg) {
      // Safety catch
      if (iLeg >= int(hisFilePtrs.size())) continue;
      string hisLeg = hisFilePtrs[iLeg]->getSRC();
      if (hisLeg == "" || hisLeg == "none") hisLeg = "File";
      // Allow user to override auto label
      if (nLeg + iLeg < int(legendIn.size())) hisLeg = legendIn[nLeg+iLeg];
      nLegTot++;
      TLegendEntry *entry=legendPtr->AddEntry("his",hisLeg.c_str(),"lP");
      entry->SetMarkerStyle(graphsHis[iLeg].GetMarkerStyle());
      entry->SetMarkerSize(graphsHis[iLeg].GetMarkerSize());
      entry->SetMarkerColor(graphsHis[iLeg].GetMarkerColor());
      entry->SetLineColor(graphsHis[iLeg].GetLineColor());
      entry->SetLineStyle(graphsHis[iLeg].GetLineStyle());
      entry->SetLineWidth(graphsHis[iLeg].GetLineWidth());
      entry->SetTextAlign(12);
      entry->SetTextFont(42);
      entry->SetTextColor(1);	
    }

    // Optionally add separate legend with chi2 values (only if comparing to data)
    if (!hasDataSav || !dataNormPtr || graphsNorm.size() == 0) return;

    if (hasLegendChi2) delete legendChi2Ptr;
    xLeg    = xLegEnd + 0.005;
    xLegEnd = xLeg + 0.128;
    legendChi2Ptr = new TLegend(xLeg,yLegBot,xLegEnd,yLegTop,NULL,"rNDC");
    legendChi2Ptr->SetBorderSize(0);
    legendChi2Ptr->SetFillStyle(0);

    // Aligned with data legend, write: "chi2/Nbins"
    string chi2Leg = "#chi_{5%}^{2}/N_{bins}";
    TLegendEntry *entry=legendChi2Ptr->AddEntry("chi2",chi2Leg.c_str(),"");
    entry->SetTextAlign(12);
    entry->SetTextFont(42);
    entry->SetTextColor(1);	

    // MC graphs
    int nChi2legTot = 0;
    int nChi2valid  = 0;
    for (int iLeg = 0; iLeg < int(graphsNorm.size()+graphsHis.size()); ++iLeg) {
      double chi2, chi2err;
      if (iLeg < int(graphsNorm.size())) {
        chi2 = graphsNorm[iLeg].getChi2();
        chi2err = graphsNorm[iLeg].getChi2err();
      } else {
        chi2 = graphsHisNorm[iLeg-graphsNorm.size()].getChi2();
        chi2err = graphsHisNorm[iLeg-graphsNorm.size()].getChi2err();
      }        
      chi2Leg = "";      
      if (chi2 >= 0.) {
        std::ostringstream s;
        s << setprecision(1);
        s << fixed << chi2 <<" #pm"<<chi2err;
        chi2Leg = s.str();
        // Count up number of valid chi2 values
        nChi2valid++;
      } 
      entry=legendChi2Ptr->AddEntry("chi2",chi2Leg.c_str(),"p");
      nChi2legTot++;
      entry->SetTextAlign(22);
      entry->SetTextFont(42);
      entry->SetTextColor(1);	      
      entry->SetMarkerStyle(21);
      entry->SetMarkerSize(1.1);
      if (chi2 < 0.) entry->SetMarkerColor(kWhite);
      else if (chi2 < 1.05) entry->SetMarkerColor(80);
      else if (chi2 < 4.05) entry->SetMarkerColor(90);
      else if (chi2 < 9.05) entry->SetMarkerColor(93);
      else entry->SetMarkerColor(100);
    }
    // Only print chi2 values if they are aligned with main legend.
    if (nChi2legTot == nLegTot && nChi2valid >= 1) hasLegendChi2 = true;
  }
    
  //------------------------------------------------------------------------
  // Set LegendUncVar 
  void VRdisplay::setLegendUncVar(vector<string> legendIn) {

    if (!hasAxes() || graphs.size() == 0 || legendIn.size() == 0 
        || !showRelUncPads) return;
    // Destroy pointer if MCnames already exists (to avoid mem leak)
    if (hasLegendUncVar) {
      delete legendUncVarTopPtr;
      delete legendUncVarBotPtr;
    }
    // Get axis range from axesPtr
    int NBins   = axesPtr->GetNbinsX();
    double xLo  = axesPtr->GetBinLowEdge(1);
    if (logX && xLo == 0.0) 
      xLo += 0.5 * axesPtr->GetBinLowEdge(2);
    double xHi  = axesPtr->GetBinLowEdge(NBins+1);
    double yLoTop = 1.01;
    double yHiTop = 1.5;
    double yLoBot = -yHiTop;
    double yHiBot = -yLoTop;
    legendUncVarTopPtr = new TLegend(xLo,yLoTop,xHi,yHiTop,NULL,"r");
    legendUncVarBotPtr = new TLegend(xLo,yLoBot,xHi,yHiBot,NULL,"r");
    hasLegendUncVar = true;
    legendUncVarTopPtr->SetBorderSize(0);
    legendUncVarTopPtr->SetFillStyle(0);
    legendUncVarTopPtr->SetNColumns(7);
    legendUncVarBotPtr->SetNColumns(7);
    legendUncVarBotPtr->SetBorderSize(0);
    legendUncVarBotPtr->SetFillStyle(0);
    // Add uncertainty variation legends
    for (int iLeg = 0; iLeg < int(legendIn.size()); ++iLeg) {
      TLegend* legPtr = (iLeg < 5) ? legendUncVarTopPtr : legendUncVarBotPtr;
      if (iLeg >= int(graphsVar.size())) continue;
      string opt = "l";
      if (graphsVar[iLeg].GetFillStyle() != 0.) opt = "lf";
      TLegendEntry *entry=legPtr->AddEntry("RU",legendIn[iLeg].c_str(),
                                           opt.c_str());
      entry->SetLineColor(graphsVar[iLeg].GetLineColor());
      entry->SetLineStyle(graphsVar[iLeg].GetLineStyle());
      entry->SetLineWidth(graphsVar[iLeg].GetLineWidth());
      entry->SetMarkerStyle(graphsVar[iLeg].GetMarkerStyle());
      entry->SetMarkerSize(graphsVar[iLeg].GetMarkerSize());
      entry->SetMarkerColor(graphsVar[iLeg].GetMarkerColor());
      entry->SetFillStyle(graphsVar[iLeg].GetFillStyle());
      entry->SetFillColor(graphsVar[iLeg].GetFillColor());
      entry->SetTextAlign(12);
      entry->SetTextFont(42);
      entry->SetTextColor(1);
    }
  }
      

} // End namespace Vincia

