// AntennaFunctions.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2016 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Include the AntennaFunctions header file
#include "Vincia/AntennaFunctions.h"

namespace Vincia {
  
  //**********************************************************************

  //-----------
  // AntennaFSR
  // Base class definitions not found in header
  
  bool AntennaFSR::init(bool isSectorIn, int minMaxVar) {
      
    // Check whether pointers are initialized
    if (!isInitPtr) return false;

    // Verbose level
    verbose              = settingsPtr->mode("Vincia:verbose");
    
    // Antenna on/off
    isOnSav              = settingsPtr->flag(vinciaName());
    
    // Antenna type
    isSectorSav          = isSectorIn;

    // Terms on/off
    useFiniteTermsSav    = settingsPtr->flag("Vincia:useFiniteTerms");
    useCollinearTermsSav = settingsPtr->flag("Vincia:useCollinearTerms");

    // Charge factor
    chargeFactorSav      = settingsPtr->parm(vinciaName()+":chargeFactor");

    // Kinematics map type (first look for specific, else default to global)
    if (settingsPtr->isMode(vinciaName()+":kineMapType"))
      kineMapTypeSav     = settingsPtr->mode(vinciaName()+":kineMapType");
    else
      kineMapTypeSav     = settingsPtr->mode("Vincia:kineMapType");

    // Initialize user-defined finite terms (minMaxVar = 0)
    Cs00 = settingsPtr->parm(vinciaName()+":Cs00");
    Cs10 = settingsPtr->parm(vinciaName()+":Cs10");
    Cs01 = settingsPtr->parm(vinciaName()+":Cs01");
    Cf00 = settingsPtr->parm(vinciaName()+":Cf00");
    Cf10 = settingsPtr->parm(vinciaName()+":Cf10");
    Cf01 = settingsPtr->parm(vinciaName()+":Cf01");
    if (settingsPtr->isParm(vinciaName()+":Cp00"))
      Cp00 = settingsPtr->parm(vinciaName()+":Cp00");
    if (settingsPtr->isParm(vinciaName()+":Cp10"))
      Cp10 = settingsPtr->parm(vinciaName()+":Cp10");
    if (settingsPtr->isParm(vinciaName()+":Cp01"))
      Cp01 = settingsPtr->parm(vinciaName()+":Cp01");

    // Base class max variation is all C00 = 5
    // (Note: should be good enough for most inheriting classes)
    if (minMaxVar >= 1) {
      Cs00 = 5;
      Cf00 = 5;
      Cp00 = 5;
      Cs01 = 0;
      Cs10 = 0;
      Cf01 = 0;
      Cf10 = 0;
      Cp01 = 0;
      Cp10 = 0;
    }
    
    // Base class min variation is all Cij = 0 
    // (Note: may give problems with positivity in some cases. May need
    // to be redefined by inheriting classes.)
    else if (minMaxVar <= -1) {
      Cs00 = 0;
      Cf00 = 0;
      Cp00 = 0;
      Cs01 = 0;
      Cs10 = 0;
      Cf01 = 0;
      Cf10 = 0;
      Cp01 = 0;
      Cp10 = 0;
    }

    // Global flag to set if there are non-zero finite terms
    // (can be used to speed up generation with zero finite terms slightly)
    hasFiniteTerms = (Cs00 != 0.0 || Cs10 != 0.0 || Cs01 != 0.0 ||
                      Cf00 != 0.0 || Cf10 != 0.0 || Cf01 != 0.0 ||
                      Cp00 != 0.0 || Cp10 != 0.0 || Cp01 != 0.0);
    
    // Collinear partitioning: different for global and sector
    if (!isSectorSav) {
      // Global antennae: alpha is adjustable parameter.
      alphaSav = settingsPtr->parm("Vincia:octetPartitioning");    
    } 
    else {
      // Sector antennae: force alpha = 1 and check sector terms 
      alphaSav = 1.0;
      useSectorTermsSav  =  settingsPtr->flag("Vincia:useSectorTerms");
    }
    
    // Return OK
    isInit = true;
    return isInit;
    
  }
  
  
  // The (dimensionless) antenna function. Base class returns zero.
  double AntennaFSR::antennaFunction(double, double, 
                                     double, double, double,
                                     int, int, int, int, int) {
    printOut("AntennaFSR::antennaFunction","Error. Dummy function in base "
             "class called.");
    return 0.0;
  }

  // The (dimensionless) AP kernel. Base class returns -1.
  double AntennaFSR::AltarelliParisi(double, double, double, double, double, 
                                     int, int, int, int, int) {
    return -1.0;
  }

  // Generic check of antenna function during initialization
  bool AntennaFSR::check() {

    bool isOK = true;
    
    // Check soft singularity against massless dipole factor
    if (idJ() == 21 || idJ() == 22) {
      // Test one symmetric and one asymmetric soft phase-space point
      for (int iTest = 0; iTest <= 1; ++iTest) {
        // Test invariants
        double yij = 1.0e-3 * pow(10.0,iTest);
        double yjk = 1.0e-3 / pow(10.0,iTest);
        // Compute eikonal
        double eik = 2.*(1.-yij-yjk)/yij/yjk;
        // Compute antenna
        double ant = antennaFunction(yij,yjk);
        // Compare
        double ratio = ant/eik;
        if (abs(ratio-1) >= 0.001) {
          isOK = false;
          if (verbose >= 1) printOut(vinciaName()+":check","WARNING: FAILED "
                                     "(eikonal "+num2str(iTest,1)+")");
        } else if (verbose >= 6) {
          printOut(vinciaName()+":check","OK (eikonal "+
                   num2str(iTest,1)+")");
        }
      }
    }
      
    // Test all helicity configurations for collinearity and posivity
    int hA(9), hB(9), hI(9), hJ(9), hK(9);
    string helString = " 9 9 -> 9 9 9";
    for (int iHel = -1; iHel < 16; ++iHel) {
      if (iHel >= 0) {
        hJ = 2*(iHel%2) - 1;
        hI = 2*((iHel/2)%2) - 1;
        hK = 2*((iHel/4)%2) - 1;
        hA = 2*((iHel/8)%2) - 1;
        hB = 2*((iHel/16)%2) - 1;
        helString = num2str(hA,2) + num2str(hB,2) + " ->"
          + num2str(hI,2) + num2str(hJ,2) + num2str(hK,2);
      }

      // Check collinear singularities against massless AP kernels, if 
      // implemented (base class dummy AP function returns -1)
      if (useCollinearTermsSav) {
        for (int iTest=0; iTest<=2; ++iTest) {
          // Test invariants, for a few points along collinear axis.
          double y1   = 1.0e-5;
          double y2   = 0.2 + iTest*0.3;
          // Compute AP kernels
          // V->FF splitting antennae only have JK collinear singularity
          double AP1  = AltarelliParisi(y1,y2,0.,0.,0.,hA,hB,hI,hJ,hK);
          double AP2  = AltarelliParisi(y2,y1,0.,0.,0.,hA,hB,hI,hJ,hK);
          if ( idJ() != 21 && idJ() != 22 ) AP1 = -1.;
          // There is only an IJ singularity if hB is conserved 
          if ( hB != hK ) AP1 = -1.;
          if ( hA != hI ) AP2 = -1.;
          // Sector antennae
          if ( isSectorSav ) {
            // Gluon emission: allow -X -> +-X but not -X -> ++X
            if ( idJ() == 21 || idJ() == 22 ) {
              int aCheck = hA*hI + hA*hJ;
              int bCheck = hB*hJ + hB*hK;
              if (aCheck <= -2 || bCheck <= -2) {
                AP1 = -1.;
                AP2 = -1.;
              }
            }
          }
          // Global antennae
          else {
            // Gluon emission: do not allow external helicity flips
            if ( idJ() == 21 || idJ() == 22 ) {
              if ( hA != hI || hB != hK ) {
                AP1 = -1.;
                AP2 = -1.;
              }
            }
          }
          // Are there any limits to check?
          if (AP1 < 0. && AP2 < 0.) continue;

          // Compute antennae
          double ant1 = antennaFunction(y1,y2,0.,0.,0.,hA,hB,hI,hJ,hK);
          double ant2 = antennaFunction(y2,y1,0.,0.,0.,hA,hB,hI,hJ,hK);
          // For global antennae, add "swapped" terms to represent neighbor
          if (!isSectorSav) {
            if (idA() == 21) {
              ant1 += antennaFunction(y1,1.-y1-y2,0.,0.,0.,hA,hB,hJ,hI,hK);
              ant2 += antennaFunction(y2,1.-y1-y2,0.,0.,0.,hA,hB,hJ,hI,hK);
            }
            if (idB() == 21) {
              ant1 += antennaFunction(1.-y1-y2,y2,0.,0.,0.,hA,hB,hI,hK,hJ);
              ant2 += antennaFunction(1.-y1-y2,y1,0.,0.,0.,hA,hB,hI,hK,hJ);
            }
          }
          
          // Compare
          if (AP1 > 0.) {
            double ratio = ant1/AP1;
            if (abs(ratio-1) >= 0.01) {
              isOK = false;
              if (verbose >= 1) 
                printOut(vinciaName()+":check","WARNING (collinear ij "
                         +num2str(iTest,1)+" "+helString+" )");
              if (verbose >= 3) {
                setprecision(6);
                cout<<"    ant  = "<<num2str(ant1,9)<<" y1 = "<<num2str(y1,9)
                    <<" y2 = "<<num2str(y2,9)<<" "<<endl;
                cout<<"    P(z) = "<<num2str(AP1,9)<<"  z = "
                    <<num2str(zi(y1,y2),9)<<endl;
              }
            } else if (verbose >= 6) {
              printOut(vinciaName()+":check","OK (collinear ij "
                       +num2str(iTest,1)+" "+helString+" )");
            }
          }
          if (AP2 > 0.) {
            double ratio = ant2/AP2;
            if (abs(ratio-1) >= 0.01) {
              isOK = false;
              if (verbose >= 1) 
                printOut(vinciaName()+":check","WARNING: FAILED (collinear jk "
                         +num2str(iTest,1)+" "+helString+" )");
              if (verbose >= 3) {
                setprecision(6);
                cout<<"    ant  = "<<num2str(ant2,9)<<" y1 = "<<num2str(y2,9)
                    <<" y2 = "<<num2str(y1,9)<<" "<<endl;
                cout<<"    P(z) = "<<num2str(AP2,9)<<"  z = "
                    <<num2str(zk(y1,y2),9)<<endl;
              }
            } else if (verbose >= 6) {
              printOut(vinciaName()+":check","OK (collinear jk "
                       +num2str(iTest,1)+" "+helString+" )");
            }
          }
        }
      }
      
      // Check positivity
      int nPointsCheck = settingsPtr->mode("Vincia:nPointsCheck");
      bool isPositive  = true;
      bool isZero      = true;
      bool hasDeadZone = false;
      double deadZoneAvoidance = settingsPtr->parm("Vincia:deadZoneAvoidance");
      for (int iTest = 0; iTest < nPointsCheck; ++iTest) {

        // Generate a random point in phase-space triangle
        double y1 = rndmPtr->flat();
        double y2 = rndmPtr->flat();
        if (y1 + y2 > 1.0) {
          y1 = 1 - y1;
          y2 = 1 - y2;
        }
        double ant = antennaFunction(y1,y2,0.,0.,0.,hA,hB,hI,hJ,hK);

        // Check positivity (strict)
        if (ant < 0.) {
          isPositive = false;
          if (verbose >= 3) 
            printOut(vinciaName()+":check","ERROR ant("
                     +num2str(y1,9)+","+num2str(y2,9)+" ; "+helString+") < 0");
        } else if (ant > 0.) {
          isZero = false;
        }
        // Check for dead zones away from phase-space boundary
        if (1-y1-y2 > 0.05 && y1 > 0.05 && y2 > 0.05
            && ant < deadZoneAvoidance) hasDeadZone = true;
        
      } // End nPointsCheck sum for positivity checks

      // OK?
      isOK = isOK && isPositive;

      // Verbose output
      if (!isPositive && verbose >= 1) {
        printOut(vinciaName()+":check","ERROR (ant < 0 encountered "
                 +helString+" )");
      } else if (isPositive && !isZero && verbose >= 6) {
        printOut(vinciaName()+":check","OK (is positive "
                 +helString+" )");
      }
      if (!isZero) {
        if (hasDeadZone && verbose >= 1) {
          printOut(vinciaName()+":check","WARNING (dead zone encountered "
                   +helString+" )");
        } else if (!hasDeadZone && verbose >= 6) {
          printOut(vinciaName()+":check","OK (no dead zones "
                   +helString+" )"); 
        }
      }
      
    } // End helicity sum


    return isOK;
    
  }
  
  //======================================================================
  // Begin inherited classes: one for each type of antenna 

  //------------
  // QQEmit::init()

  bool QQEmit::init(bool isSectorIn, int minMaxVar) {

    // Start by doing base class initialization
    isInit = AntennaFSR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna

    // Min Variation
    if (minMaxVar == -1) {
      Cs00 = -4;
      Cs01 = 0;
      Cs10 = 0;
      Cf00 = 2;
      Cf01 = 0;
      Cf10 = 0;
      Cp00 = 0;
      Cp01 = 0;
      Cp10 = 0;
    }
          
    // Check whether finite terms are symmetric
    if (Cs10 != Cs01) {
      printOut(vinciaName()+":init","Forcing Cs10 = Cs01 by averaging.");
      Cs10 = 0.5 * (Cs10 + Cs01);
      Cs01 = Cs10;
    }
    if (Cf10 != Cf01) {
      printOut(vinciaName()+":init","Forcing Cf10 = Cf01 by averaging.");
      Cs10 = 0.5 * (Cf10 + Cf01);
      Cs01 = Cf10;
    }
    if (Cs10 != Cs01) {
      printOut(vinciaName()+":init","Forcing Cp10 = Cp01 by averaging.");
      Cp10 = 0.5 * (Cp10 + Cp01);
      Cp01 = Cp10;
    }

    // Global flag to set if there are non-zero finite terms
    // (can be used to speed up generation with zero finite terms slightly)
    hasFiniteTerms = (Cs00 != 0.0 || Cs10 != 0.0 || Cs01 != 0.0 ||
                      Cf00 != 0.0 || Cf10 != 0.0 || Cf01 != 0.0 ||
                      Cp00 != 0.0 || Cp10 != 0.0 || Cp01 != 0.0);
    
    // Return OK
    return isInit;
  }

  //------------
  // QQEmit::antennaFunction()

  double QQEmit::antennaFunction(double yij, double yjk, 
                                 double mui, double muj, double muk,
                                 int hA, int hB, int hi, int hj, int hk) {
    
    // Sanity check. Require positive invariants.
    if (yij <= 0.0 || yjk <= 0.0) return 0.0;

    // No clue about massive gluons
    if (abs(muj) > TINY) {
      if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
                                 "Ignoring gluon virtuality");
    }

    // Massive antennae: hand off to unpolarized massive helper    
    if (mui > TINY || muk > TINY) 
      return antennaMassive(yij, yjk, mui, muk);

    // Massless antennae: treat unknown polarizations as unpolarized
    if ( hA != 1 && hA != -1 ) hA = 9;
    if ( hB != 1 && hB != -1 ) hB = 9;
    if ( hi != 1 && hi != -1 ) hi = 9;
    if ( hj != 1 && hj != -1 ) hj = 9;
    if ( hk != 1 && hk != -1 ) hk = 9;
    
    // How many configurations are we averaging over?
    int nAvg = 1;
    if (hA == 9) nAvg *= 2;
    if (hB == 9) nAvg *= 2;
    
    // Shorthands
    double eik  = 1./yij/yjk;
    double cij  = 1./yij;
    double cjk  = 1./yjk;
    bool doCollinear = useCollinearTermsSav;
    bool doFinite = useFiniteTermsSav && hasFiniteTerms;

    // Do helicity sum
    double hSum = 0.0;

    // S-wave antennae (++ and --)
    if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
      // ++ > +++ && -- > --- antennae (MHV)
      double term = eik;
      if (doFinite) term += Cs00 + Cs01*yjk + Cs10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && RH[hk] ) hSum += term; 
      if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && LH[hk] ) hSum += term;
      
      // ++ > +-+ && -- > -+- antennae (NMHV)
      term = eik - 2*(cij+cjk);
      if (doCollinear) term += yjk*cij + yij*cjk;
      if (doFinite)    term += Cf00 + Cf01*yjk + Cf10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;
    }
    
    // P-wave antennae (+- and -+)
    if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
      // +- > ++- && -+ > --+ antennae
      double term = eik - 2*cjk;
      if (doCollinear) term += yij*cjk; 
      if (doFinite)    term += Cp00 + Cp01*yjk + Cp10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      
      // +- > +-- && -+ > -++ antennae
      term = eik - 2*cij;
      if (doCollinear) term += yjk*cij;
      if (doFinite)    term += Cp00 + Cp01*yjk + Cp10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && RH[hk] ) hSum += term;
    }
    
    // Return helicity sum, averaged over initial helicities
    return hSum/nAvg;
    
  }

  //------------
  // QQEmit::antennaMassive()

  // Function to compute unpolarized massive antenna
  double QQEmit::antennaMassive(double yij, double yjk, 
                                   double mui, double muk) {

    // Compute third invariant
    double mu2i = pow2(mui);
    double mu2k = pow2(muk);
    double yik  = 1.0 - yij - yjk - mu2i - mu2k;

    // Compute eikonal, with mass correction
    double sum  = 2*yik/yij/yjk - 2*mu2i/pow2(yij) - 2*mu2k/pow2(yjk);
       
    // Denominator for collinear and finite terms
    double denom  = 1.0 - mu2i - mu2k;

    // Non-eikonal collinear terms
    if (useCollinearTermsSav) 
      sum += (yij/yjk + yjk/yij) / denom;
 
    // Non-singular terms
    if (useFiniteTermsSav) {
      
      // Add average of summed helicity-dependent finite terms      
      if (hasFiniteTerms) {
        double C00 = 0.5 * ( Cs00 + Cf00 + 2*Cp00 );
        double C01 = 0.5 * ( Cs01 + Cf01 + 2*Cp01 );
        double C10 = 0.5 * ( Cs10 + Cf10 + 2*Cp10 );
        sum += (C00 + C10*yij + C01*yjk) / denom; 
      }
      
    } // end of if (useFiniteTermsSav)
    
    // Return sum of antenna terms
    return sum;
    
  }

  //======================================================================

  //------------
  // QGEmit::init()

  bool QGEmit::init(bool isSectorIn, int minMaxVar) {

    // Start by doing base class initialization
    isInit = AntennaFSR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna

    // Set MIN variation specifically for this antenna
    // (MAX variation is fine just inheriting from base class)
    if (minMaxVar == -1) {
      Cs00 = -3;
      Cf00 = 3;
      Cp00 = 0;
      Cs10 = 0;
      Cs01 = 0;
      Cf10 = 0;
      Cf01 = 0;
      Cp10 = 0;
      Cp01 = 0;
    }

    // Global flag to set if there are non-zero finite terms
    // (can be used to speed up generation with zero finite terms slightly)
    hasFiniteTerms = (Cs00 != 0.0 || Cs10 != 0.0 || Cs01 != 0.0 ||
                      Cf00 != 0.0 || Cf10 != 0.0 || Cf01 != 0.0 ||
                      Cp00 != 0.0 || Cp10 != 0.0 || Cp01 != 0.0);
    
    // Return OK
    return isInit;
  }

  //------------
  // QGEmit::antennaFunction()

  double QGEmit::antennaFunction(double yij, double yjk, 
                                    double mui, double muj, double muk,
                                    int hA, int hB, int hi, int hj, int hk) {
    
    // Sanity check. Require positive invariants.
    if (yij <= 0.0 || yjk <= 0.0) return 0.0;

    // No clue about massive gluons
    if (muj != 0. || muk != 0.) {
      if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
                                 "Ignoring gluon virtuality");
      muj = 0.0;
      muk = 0.0;
    }

    // Massive antennae: hand off to unpolarized massive helper    
    if (mui > 0.) 
      return antennaMassive(yij, yjk, mui);

    // Massless antennae: treat unknown polarizations as unpolarized
    if ( hA != 1 && hA != -1 ) hA = 9;
    if ( hB != 1 && hB != -1 ) hB = 9;
    if ( hi != 1 && hi != -1 ) hi = 9;
    if ( hj != 1 && hj != -1 ) hj = 9;
    if ( hk != 1 && hk != -1 ) hk = 9;
    
    // Hhow many configurations are we averaging over?
    int nAvg = 1;
    if (hA == 9) nAvg *= 2;
    if (hB == 9) nAvg *= 2;
    
    // Shorthands
    double eik      = 1./yij/yjk;
    double cij      = 1./yij;
    double cjk      = 1./yjk;
    double a        = alphaSav;
    double sectorK  = (isSectorSav) ? cjk/(1-yij) : 0.0;
    bool doCollinear = useCollinearTermsSav;
    bool doFinite   = useFiniteTermsSav && hasFiniteTerms;
    bool doSector   = isSectorSav && useCollinearTermsSav && useSectorTermsSav;

    // Do helicity sum
    double hSum = 0.0;

    // S-wave antennae (++ and --)
    if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
      // ++ > +++ && -- > --- antennae (MHV)
      double term = eik;
      if (doSector) term += sectorK;
      else if (doCollinear && a != 1.0) term += (1.-a)*cjk - 2*(1.-a)*yij*cjk;
      if (doFinite) term += Cs00 + Cs01*yjk + Cs10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && RH[hk] ) hSum += term; 
      if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && LH[hk] ) hSum += term;
      
      // ++ > +-+ && -- > -+- antennae (NMHV)
      term = eik - 2*cij - 3*cjk;
      if (doCollinear) term += yjk*cij + 3*yij*cjk - pow2(yij)*cjk;
      if (doFinite)    term += Cf00 + Cf01*yjk + Cf10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;

      // ++ > ++- && -- > --+ antennae (pure sector) 
      if (doSector) {
        term = sectorK - cjk*(1. + yij + pow2(yij));
        if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
        if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      }

    }
    
    // P-wave antennae (+- and -+)
    if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
      // +- > ++- && -+ > --+ antennae
      double term = eik - 3*cjk;
      if (doCollinear) term += 3*yij*cjk - pow2(yij)*cjk; 
      if (doFinite)    term += Cp00 + Cp01*yjk + Cp10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      
      // +- > +-- && -+ > -++ antennae
      term = eik - 2*cij;
      if (doCollinear) term += yjk*cij;
      if (doSector)    term += sectorK;
      else if (doCollinear && a != 1.) term += (1.-a)*cjk - 2*(1.-a)*yij*cjk;
      if (doFinite)    term += Cp00 + Cp01*yjk + Cp10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && RH[hk] ) hSum += term;

      // +- > +-+ && -+ > -+- antennae (pure sector)
      if (doSector) {
        term = sectorK - cjk*(1. + yij + pow2(yij));
        if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
        if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;
      }
    }
    
    // Return helicity sum, averaged over initial helicities
    return hSum/nAvg;
    
  }

  //------------
  // QGEmit::antennaMassive()

  // Function to compute unpolarized massive antenna
  double QGEmit::antennaMassive(double yij, double yjk, double mui) {

    // Compute third invariant
    double mu2i = pow2(mui);
    double yik  = 1.0 - yij - yjk - mu2i;

    // Compute eikonal, with mass correction
    double sum  = 2*yik/yij/yjk - 2*mu2i/pow2(yij);
       
    // Denominator for collinear and finite terms
    double denom  = 1.0 - mu2i;

    // Non-eikonal collinear terms
    if (useCollinearTermsSav) {
      sum += ( yjk/yij + yij/yjk * (1.-yij/denom) ) / denom; 
      // Massive sector terms (tentatively by replacing yij -> yij/denom)
      if (isSectorSav && useSectorTermsSav) 
        sum += (2./(1-yij/denom) - 2. + (1.-yij/denom)*yij/denom)/yjk;
    }

    // Non-singular terms
    if (useFiniteTermsSav) {
      
      // Add average of summed helicity-dependent finite terms      
      if (hasFiniteTerms) {
        double C00 = 0.5 * ( Cs00 + Cf00 + 2*Cp00 );
        double C01 = 0.5 * ( Cs01 + Cf01 + 2*Cp01 );
        double C10 = 0.5 * ( Cs10 + Cf10 + 2*Cp10 );
        sum += (C00 + C10*yij + C01*yjk) / denom; 
      }

    } // end of if (useFiniteTermsSav)
    
    // Return sum of antenna terms
    return sum;
    
  }

  //======================================================================

  //------------
  // GGEmit::init()

  bool GGEmit::init(bool isSectorIn, int minMaxVar) {

    // Start by doing base class initialization
    isInit = AntennaFSR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna

    // Set MIN variation for this specific antenna
    // (MAX variation is fine inheriting from base class)
    if (minMaxVar == -1) {
      Cs00 = -4;
      Cf00 = 3.7;
      Cp00 = 0.0;
      Cs01 = 0;
      Cs10 = 0;
      Cf01 = 0;
      Cf10 = 0;
      Cp01 = 0;
      Cp10 = 0;
    }

    // Global flag to set if there are non-zero finite terms
    // (can be used to speed up generation with zero finite terms slightly)
    hasFiniteTerms = (Cs00 != 0.0 || Cs10 != 0.0 || Cs01 != 0.0 ||
                      Cf00 != 0.0 || Cf10 != 0.0 || Cf01 != 0.0 ||
                      Cp00 != 0.0 || Cp10 != 0.0 || Cp01 != 0.0);
    
    // Check whether finite terms are symmetric
    if (Cs10 != Cs01) {
      printOut(vinciaName()+":init","Forcing Cs10 = Cs01 by averaging.");
      Cs10 = 0.5 * (Cs10 + Cs01);
      Cs01 = Cs10;
    }
    if (Cf10 != Cf01) {
      printOut(vinciaName()+":init","Forcing Cf10 = Cf01 by averaging.");
      Cs10 = 0.5 * (Cf10 + Cf01);
      Cs01 = Cf10;
    }
    if (Cs10 != Cs01) {
      printOut(vinciaName()+":init","Forcing Cp10 = Cp01 by averaging.");
      Cp10 = 0.5 * (Cp10 + Cp01);
      Cp01 = Cp10;
    }

    // Return OK
    return isInit;
  }

  //------------
  // GGEmit::antennaFunction()
  
  double GGEmit::antennaFunction(double yij, double yjk, 
                                    double mui, double muj, double muk,
                                    int hA, int hB, int hi, int hj, int hk) {
    
    // Sanity check. Require positive invariants.
    if (yij <= 0.0 || yjk <= 0.0) return 0.0;
    
    // No clue about massive gluons
    if (mui != 0. || muj != 0. || muk != 0.) {
      if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
                                 "Ignoring gluon virtuality.");
      mui = 0.0;
      muj = 0.0;
      muk = 0.0;
    }
    
    // Massless antennae: treat unknown polarizations as unpolarized
    if ( hA != 1 && hA != -1 ) hA = 9;
    if ( hB != 1 && hB != -1 ) hB = 9;
    if ( hi != 1 && hi != -1 ) hi = 9;
    if ( hj != 1 && hj != -1 ) hj = 9;
    if ( hk != 1 && hk != -1 ) hk = 9;
    
    // Hhow many configurations are we averaging over?
    int nAvg = 1;
    if (hA == 9) nAvg *= 2;
    if (hB == 9) nAvg *= 2;
    
    // Shorthands
    double eik      = 1./yij/yjk;
    double cij      = 1./yij;
    double cjk      = 1./yjk;
    double a        = alphaSav;
    double sectorI  = (isSectorSav) ? cij/(1-yjk) : 0.0;
    double sectorK  = (isSectorSav) ? cjk/(1-yij) : 0.0;
    bool doCollinear = useCollinearTermsSav;
    bool doFinite   = useFiniteTermsSav && hasFiniteTerms;
    bool doSector   = isSectorSav && useCollinearTermsSav && useSectorTermsSav;

    // Do helicity sum
    double hSum = 0.0;

    // S-wave antennae (++ and --)
    if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
      // ++ > +++ && -- > --- antennae (MHV)
      double term = eik;
      if (doSector) term += sectorI + sectorK;
      else if (doCollinear && a != 1.0) 
        term += (1.-a)*(cjk + cij) - 2*(1.-a)*(yij*cjk + yjk*cij);
      if (doFinite) term += Cs00 + Cs01*yjk + Cs10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && RH[hk] ) hSum += term; 
      if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && LH[hk] ) hSum += term;
      
      // ++ > +-+ && -- > -+- antennae (NMHV)
      term = eik - 3*cij - 3*cjk;
      if (doCollinear) 
        term += 3*yjk*cij + 3*yij*cjk - pow2(yij)*cjk - pow2(yjk)*cij;
      if (doFinite) term += Cf00 + Cf01*yjk + Cf10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;

      // ++ > ++- && -- > --+ antennae (pure sector) 
      if (doSector) {
        term = sectorK - cjk*(1. + yij + pow2(yij));
        if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
        if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      }

      // ++ > -++ && -- > +-- antennae (pure sector) 
      if (doSector) {
        term = sectorI - cij*(1. + yjk + pow2(yjk));
        if ( RH[hA] && RH[hB] && LH[hi] && RH[hj] && RH[hk] ) hSum += term;
        if ( LH[hA] && LH[hB] && RH[hi] && LH[hj] && LH[hk] ) hSum += term;
      }

    }
    
    // P-wave antennae (+- and -+)
    if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
      // +- > ++- && -+ > --+ antennae
      double term = eik - 3*cjk;
      if (doCollinear) term += 3*yij*cjk - pow2(yij)*cjk; 
      if (doSector)    term += sectorI;
      else if (doCollinear && a != 1.) term += (1.-a)*cij - 2*(1.-a)*yjk*cij;
      if (doFinite)    term += Cp00 + Cp01*yjk + Cp10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      
      // +- > +-- && -+ > -++ antennae
      term = eik - 3*cij;
      if (doCollinear) term += 3*yjk*cij - pow2(yjk)*cij;
      if (doSector)    term += sectorK;
      else if (doCollinear && a != 1.) term += (1.-a)*cjk - 2*(1.-a)*yij*cjk;
      if (doFinite)    term += Cp00 + Cp01*yjk + Cp10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && RH[hk] ) hSum += term;

      // +- > +-+ && -+ > -+- antennae (pure sector)
      if (doSector) {
        term = sectorK - cjk*(1. + yij + pow2(yij));
        if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
        if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;
      }

      // +- > -+- && -+ > +-+ antennae (pure sector)
      if (doSector) {
        term = sectorI - cij*(1. + yjk + pow2(yjk));
        if ( RH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;
        if ( LH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      }

    }
    
    // Return helicity sum, averaged over initial helicities 
    return hSum/nAvg;
    
  }

  //======================================================================

  //------------
  // QGSplit::init()

  bool QGSplit::init(bool isSectorIn, int minMaxVar) {

    // Start by doing base class initialization
    isInit = AntennaFSR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna

    // MIN variation (zero finite terms)
    if (minMaxVar == -1) {
      Cs00 = 0;
      Cs10 = 0;
      Cs01 = 0;
      Cf00 = 0;
      Cf10 = 0;
      Cf01 = 0;      
    }

    // Global flag to set if there are non-zero finite terms
    // (can be used to speed up generation with zero finite terms slightly)
    hasFiniteTerms = (Cs00 != 0.0 || Cs10 != 0.0 || Cs01 != 0.0 ||
                      Cf00 != 0.0 || Cf10 != 0.0 || Cf01 != 0.0);
    
    // Return OK
    return isInit;
  }

  //------------
  // QGSplit::antennaFunction()
  double QGSplit::antennaFunction(double yij, double yjk, 
                                  double mui, double muj, double muk,
                                  int hA, int hB, int hi, int hj, int hk) {

    // Sanity check. Require positive invariants.
    if (yij <= 0.0 || yjk <= 0.0) return 0.0;
    
    // Require equal jk masses
    if ( abs(muj-muk) > TINY ) {
      if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
                                 "Warning: forcing mq = mqbar.");
      cout<<"QGSPLIT: y(ij,jk) = "<<yij<<" "<<yjk<<" mu(i,j,k) = "
          <<scientific<<mui<<" "<<muj<<" "<<muk<<endl;
      muk = muj;
    }

    // Massive antennae: hand off to unpolarized massive helper    
    if (mui > 0. || muj > 0.) 
      return antennaMassive(yij, yjk, mui, muj);

    // Massless antennae: treat unknown polarizations as unpolarized
    if ( hA != 1 && hA != -1 ) hA = 9;
    if ( hB != 1 && hB != -1 ) hB = 9;
    if ( hi != 1 && hi != -1 ) hi = 9;
    if ( hj != 1 && hj != -1 ) hj = 9;
    if ( hk != 1 && hk != -1 ) hk = 9;
    
    // How many configurations are we averaging over?
    int nAvg = 1;
    if (hA == 9) nAvg *= 2;
    if (hB == 9) nAvg *= 2;
    
    // Shorthands
    double cjk  = 1./yjk;
    double tf   = cjk * 0.5 * pow2(yij);
    double ts   = cjk * ( 0.5 - yij ) + tf;
    bool doFinite = useFiniteTermsSav && hasFiniteTerms;

    // Do helicity sum
    double hSum = 0.0;

    // S-wave antennae (++ and --)
    if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
      // ++ > +-+ && -- > -+- antennae (s)
      double term = ts;
      if (doFinite)    term += Cs00 + Cs01*yjk + Cs10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;

      // ++ > ++- && -- > --+ antennae (f)
      term = tf;
      if (doFinite)    term += Cf00 + Cf01*yjk + Cf10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;

    }
    
    // P-wave antennae (+- and -+)
    if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
      // +- > ++- && -+ > --+ antennae (s)
      double term = ts; 
      if (doFinite)    term += Cs00 + Cs01*yjk + Cs10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      
      // +- > +-+ && -+ > -+- antennae (f)
      term = tf;
      if (doFinite)    term += Cf00 + Cf01*yjk + Cf10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;

    }
    
    // Sector splitting antennae = 2 * Global
    if (isSectorSav) hSum *= 2.;    

    // Return helicity sum, averaged over initial helicities
    return hSum/nAvg;
  }

  //------------
  // QGSplit::antennaMassive()

  // Function to compute unpolarized massive antenna
  double QGSplit::antennaMassive(double yij, double yjk, 
                                    double mui, double muj) {

    // Compute third invariant
    double mu2i = pow2(mui);
    double mu2j = pow2(muj);
    double yik  = 1.0 - yij - yjk - mu2i - 2*mu2j;

    // Gluon propagator and sIK^2 denominator factors
    double prop  = yjk + 2*mu2j;
    double denom = pow2(1. - mu2i);

    // Compute eikonal, with mass correction
    double sum  = 0.5 / prop * ( (pow2(yik) + pow2(yij))/denom
                                 + 2*mu2j/prop );
                                 
    // Non-singular terms
    if (useFiniteTermsSav) {
      
      // Add average of summed helicity-dependent finite terms      
      // (Here written with explicit averaging and summing to 
      // avoid confusion)
      if (hasFiniteTerms) {
        double C00 = 0.5 * ( 2*Cs00 + 2*Cf00);
        double C01 = 0.5 * ( 2*Cs01 + 2*Cf01 );
        double C10 = 0.5 * ( 2*Cs10 + 2*Cf10 );
        sum += yjk / prop * (C00 + C10*yij + C01*yjk) / denom; 
      }

    } // end of if (useFiniteTermsSav)
    
    // Sector splitting antennae = 2 * Global
    if (isSectorSav) sum *= 2;
    
    // Return sum of antenna terms
    return sum;
    
  }

//======================================================================

  //------------
  // GGSplit::init()

  bool GGSplit::init(bool isSectorIn, int minMaxVar) {

    // Start by doing base class initialization
    isInit = AntennaFSR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna

    // MIN variation (zero finite terms)
    if (minMaxVar == -1) {
      Cs00 = 0;
      Cs10 = 0;
      Cs01 = 0;
      Cf00 = 0;
      Cf10 = 0;
      Cf01 = 0;      
    }

    // Global flag to set if there are non-zero finite terms
    // (can be used to speed up generation with zero finite terms slightly)
    hasFiniteTerms = (Cs00 != 0.0 || Cs10 != 0.0 || Cs01 != 0.0 ||
                      Cf00 != 0.0 || Cf10 != 0.0 || Cf01 != 0.0);
    
    // Return OK
    return isInit;
  }

  //------------
  // GGSplit::antennaFunction()

  double GGSplit::antennaFunction(double yij, double yjk, 
                                    double mui, double muj, double muk,
                                    int hA, int hB, int hi, int hj, int hk) {

    // Sanity check. Require positive invariants.
    if (yij <= 0.0 || yjk <= 0.0) return 0.0;
    
    // Ignore gluon mass
    if (mui != 0.0) {
      if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
                                 "Ignoring gluon virtuality");
      mui = 0.0;
    }

    // Require equal jk masses (if no k given, set k = j)
    if ( abs(muj-muk) > TINY ) {
      if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
                                 "Warning: forcing mq = mqbar.");
      muk = muj;
    }

    // Massive antennae: hand off to unpolarized massive helper    
    if (mui > 0. || muj > 0.) 
      return antennaMassive(yij, yjk, muj);

    // Massless antennae: treat unknown polarizations as unpolarized
    if ( hA != 1 && hA != -1 ) hA = 9;
    if ( hB != 1 && hB != -1 ) hB = 9;
    if ( hi != 1 && hi != -1 ) hi = 9;
    if ( hj != 1 && hj != -1 ) hj = 9;
    if ( hk != 1 && hk != -1 ) hk = 9;
    
    // How many configurations are we averaging over?
    int nAvg = 1;
    if (hA == 9) nAvg *= 2;
    if (hB == 9) nAvg *= 2;
    
    // Shorthands
    double cjk  = 1./yjk;
    double tf   = cjk * 0.5 * pow2(yij);
    double ts   = cjk * ( 0.5 - yij ) + tf;
    bool doFinite = useFiniteTermsSav && hasFiniteTerms;

    // Do helicity sum
    double hSum = 0.0;

    // S-wave antennae (++ and --)
    if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
      // ++ > +-+ && -- > -+- antennae (s)
      double term = ts;
      if (doFinite)    term += Cs00 + Cs01*yjk + Cs10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;

      // ++ > ++- && -- > --+ antennae (f)
      term = tf;
      if (doFinite)    term += Cf00 + Cf01*yjk + Cf10*yij;
      if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
    }
    
    // P-wave antennae (+- and -+)
    if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
      // +- > ++- && -+ > --+ antennae (s)
      double term = ts; 
      if (doFinite)    term += Cs00 + Cs01*yjk + Cs10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      
      // +- > +-+ && -+ > -+- antennae (f)
      term = tf;
      if (doFinite)    term += Cf00 + Cf01*yjk + Cf10*yij;
      if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
      if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;
    }
    
    // Sector splitting antennae = 2 * Global
    if (isSectorSav) hSum *= 2.;

    // Return helicity sum, averaged over initial helicities
    return hSum/nAvg;
  }


  //------------
  // GGSplit::antennaMassive()

  // Function to compute unpolarized massive antenna
  double GGSplit::antennaMassive(double yij, double yjk, double muj) {

   // Compute third invariant
    double mu2j = pow2(muj);
    double yik  = 1.0 - yij - yjk - 2*mu2j;

    // Gluon propagator factor
    double prop  = yjk + 2*mu2j;

    // Compute eikonal, with mass correction
    double sum  = 0.5 / prop * ( (pow2(yik) + pow2(yij)) + 2*mu2j/prop );
                                 
    // Non-singular terms
    if (useFiniteTermsSav) {
      
      // Add average of summed helicity-dependent finite terms      
      // (Here written with explicit averaging and summing to 
      // avoid confusion)
      if (hasFiniteTerms) {
        double C00 = 0.5 * ( 2*Cs00 + 2*Cf00);
        double C01 = 0.5 * ( 2*Cs01 + 2*Cf01 );
        double C10 = 0.5 * ( 2*Cs10 + 2*Cf10 );
        sum += yjk / prop * (C00 + C10*yij + C01*yjk); 
      }

    } // end of if (useFiniteTermsSav)
    
    // Sector splitting antennae = 2 * Global
    if (isSectorSav) sum *= 2;
    
    // Return sum of antenna terms
    return sum;
    
  } 

  // End inherited AntennaFSR classes
  //======================================================================

  //-----------
  // AntennaISR
  // Base class definitions not found in header
  
  bool AntennaISR::init(bool /*isSectorIn*/, int minMaxVar) {
      
    // Check whether pointers are initialized
    if (!isInitPtr) return false;

    // Verbose level
    verbose             = settingsPtr->mode("Vincia:verbose");
    
    // Antenna on/off
    isOnSav             = settingsPtr->flag(vinciaName());

    // Charge factor
    chargeFactorSav     = settingsPtr->parm(vinciaName()+":chargeFactor");
    
    // Prefactor of finite term and ant minimum
    preFacFiniteTermSav = 0.0;
         if (minMaxVar > 0) preFacFiniteTermSav =  1.0;
    else if (minMaxVar < 0) preFacFiniteTermSav = -1.0;
    antMinSav           = 1e-6;
    
    // Check if we are the MIN variation, only is this case
    // use antMinSav to get sensible results for ant checks
    isMinVar = (minMaxVar < 0);
    
    // Return OK
    isInit = true;
    return isInit;
    
  }
  
  // The antenna function. Base class returns zero.
  double AntennaISR::antennaFunction(double, double, double, 
                                     int, int, int, int, int) {
    printOut("Antenna::antennaFunction","Error. Dummy function in base "
             "class called.");
    return 0.0;
  }

  // The AP kernel. Base class returns -1.
  double AntennaISR::AltarelliParisi(double, double, double, 
                                     int, int, int, int, int) {
    return -1.0;
  }

  // Generic check of antenna function during initialization
  bool AntennaISR::check() {

    bool isOK = true;
      
    // For check for positivity
    int nPointsCheck = settingsPtr->mode("Vincia:nPointsCheck");
    bool isPositive  = true;
    bool isZero      = true;
    bool hasDeadZone = false;
    double deadZoneAvoidance = settingsPtr->parm("Vincia:deadZoneAvoidance");
    // Use 14 TeV as phase space max
    double shhMax    = pow2(14000.0);
    
    // Initial-Initial: sab = sAB + saj + sjb
    if (isIIant()) {
      
      // Z and Higgs as testing case
      double sAB[2] = { pow2(91.0), pow2(125.0) };
    
      // Check soft singularity against dipole factor
      if (idj() == 21 || idj() == 22) {
        // Test one symmetric and one asymmetric soft phase-space point
        for (int iTest = 0; iTest <= 1; ++iTest) {
          // Test two sAB values
          for (int isAB = 0; isAB <= 1; ++isAB) {
            // Test invariants
            double yaj = 1.0e-3 * pow(10.0,iTest); // saj/sab
            double yjb = 1.0e-3 / pow(10.0,iTest); // sjb/sab
            double sab = sAB[isAB]/(1.0-yaj-yjb);  // 1 = sAB/sab + yaj + yjb
            double saj = yaj*sab;
            double sjb = yjb*sab;
            // Compute eikonal
            double eik = 2.0*sab/saj/sjb;
            // Compute antenna
            double ant = antennaFunction(saj,sjb,sAB[isAB]);
            // Compare
            double ratio = ant/eik;
            if (abs(ratio-1.0) >= 0.001) {
              isOK = false;
              if (verbose >= 1) printOut(vinciaName()+":check","WARNING: FAILED "
                                         "(eikonal "+num2str(iTest,1)+" and sAB = "+
                                         num2str((int)sqrt(sAB[isAB]))+"^2)");
            } else if (verbose >= 6) {
              printOut(vinciaName()+":check","OK (eikonal "+num2str(iTest,1)+
                       " and sAB = "+num2str((int)sqrt(sAB[isAB]))+"^2)");
            }
          }
        }
      }
    
      // Check for collinearity (no helicity APs so far)
      // Test invariants, for a few points along collinear axis
      for (int iTest = 0; iTest <= 2; ++iTest) {
        // Test two sAB values
        for (int isAB = 0; isAB <= 1; ++isAB) {
          // Test invariants
          double y1   = 1.0e-5; // saj/sAB
          double y2   = 0.2 + iTest*0.3; // sjb/sAB
          double sab = sAB[isAB]/(1.0-y1-y2);  // 1 = sAB/sab + yaj + yjb
          double s1  = y1*sab;
          double s2  = y2*sab;
          // Compute AP kernels
          double AP1 = AltarelliParisi(s1,s2,sAB[isAB]);
          double AP2 = AltarelliParisi(s2,s1,sAB[isAB]);
          // Compute antennae
          double ant1 = antennaFunction(s1,s2,sAB[isAB]);
          double ant2 = antennaFunction(s2,s1,sAB[isAB]);
          // Check if only a singularity on one side exists (-> z = -1)
          double zs1 = z1(s1,s2,sAB[isAB]);
          double zs2 = z2(s1,s2,sAB[isAB]);
          if (zs1 != -1.0) {
            double ratio = ant1/AP1;
            if (abs(ratio-1.0) >= 0.01) {
              isOK = false;
              if (verbose >= 1) printOut(vinciaName()+":check","WARNING: FAILED "
                                         "(collinear aj "+num2str(iTest,1)+" and sAB = "+
                                         num2str((int)sqrt(sAB[isAB]))+"^2)");
              if (verbose >= 3) {
                setprecision(6);
                cout<<"    ant  = "<<num2str(ant1,9)<<" y1 = "<<num2str(y1,9)
                    <<" y2 = "<<num2str(y2,9)<<" "<<endl;
                cout<<"    P(z) = "<<num2str(AP1,9)<<"  z = "
                    <<num2str(zs1,9)<<endl;
              }
            } else if (verbose >= 6) {
              printOut(vinciaName()+":check","OK (collinear aj "+num2str(iTest,1)+
                       " and sAB = "+num2str((int)sqrt(sAB[isAB]))+"^2)");
            }
          }
          if (zs2 != -1.0) {
            double ratio = ant2/AP2;
            if (abs(ratio-1.0) >= 0.01) {
              isOK = false;
              if (verbose >= 1) printOut(vinciaName()+":check","WARNING: FAILED "
                                         "(collinear jb "+num2str(iTest,1)+" and sAB = "+
                                         num2str((int)sqrt(sAB[isAB]))+"^2)");
              if (verbose >= 3) {
                setprecision(6);
                cout<<"    ant  = "<<num2str(ant1,9)<<" y1 = "<<num2str(y1,9)
                    <<" y2 = "<<num2str(y2,9)<<" "<<endl;
                cout<<"    P(z) = "<<num2str(AP2,9)<<"  z = "
                    <<num2str(zs2,9)<<endl;
              }
            } else if (verbose >= 6) {
              printOut(vinciaName()+":check","OK (collinear jb "+num2str(iTest,1)+
                       " and sAB = "+num2str((int)sqrt(sAB[isAB]))+"^2)");
            }
          }
        }
      }
      
      // Test all helicity configurations for posivity and dead zones
      int hA(9), hB(9), ha(9), hj(9), hb(9);
      string helString = " 9 9 -> 9 9 9";
      for (int iHel = -1; iHel < 0/*16*/; ++iHel) { // Begin - loop over hels
        if (iHel >= 0) {
          hj = 2*(iHel%2) - 1;
          ha = 2*((iHel/2)%2) - 1;
          hb = 2*((iHel/4)%2) - 1;
          hA = 2*((iHel/8)%2) - 1;
          hB = 2*((iHel/16)%2) - 1;
          helString = num2str(hA,2) + num2str(hB,2) + " ->" +
                      num2str(ha,2) + num2str(hj,2) + num2str(hb,2);
        }
        
        // Gluon emission in global ants: 
        // do not allow external helicity flips
        if ( idj() == 21 || idj() == 22 )
          if ( hA != ha || hB != hb ) continue;
        
        // Check for positivity and dead zones
        for (int iTest = 0; iTest <= nPointsCheck; ++iTest) {
          // sAB^0.5 between 1 and 1000 GeV
          double sABnow = pow2(1.0+rndmPtr->flat()*999.0);
          // Generate a random point in phase-space triangle
          double yaj = rndmPtr->flat(); // saj/sab
          double yjb = rndmPtr->flat(); // sjb/sab
          if (yaj + yjb > 1.0) {
            yaj = 1.0-yaj;
            yjb = 1.0-yjb;
          }
          double sab = sABnow/(1.0-yaj-yjb);  // 1 = sAB/sab + yaj + yjb
          // Check that sab < shhMax
          if (sab > shhMax) continue;
          double saj = yaj*sab;
          double sjb = yjb*sab;
          // Compute antenna
          double ant = antennaFunction(saj,sjb,sABnow,hA,hB,ha,hj,hb);
          // Check positivity (strict)
          if (ant < 0.0) {
            isPositive = false;
            if (verbose >= 3) 
              printOut(vinciaName()+":check","ERROR ant("+num2str((int)sqrt(saj))+"^2,"+
                       num2str((int)sqrt(sjb))+"^2,"+num2str((int)sqrt(sABnow))+"^2 ; "+
                       helString+") < 0");
          } else if (ant > 0.0) {
            isZero = false;
          }
          // Check for dead zones away from phase-space boundary
          if ( (1.0-yaj-yjb > 0.05) && (yaj > 0.05) && (yjb > 0.05)
               && (ant*sABnow < deadZoneAvoidance) ) hasDeadZone = true;
        }
        
        // OK?
        isOK = isOK && isPositive;
        
        // Verbose output
        if (!isPositive && verbose >= 1) {
          printOut(vinciaName()+":check","ERROR (ant < 0 encountered "
                   +helString+" )");
        } else if (isPositive && !isZero && verbose >= 6) {
          printOut(vinciaName()+":check","OK (is positive "
                   +helString+" )");
        }
        if (!isZero) {
          if (hasDeadZone && verbose >= 1) {
            printOut(vinciaName()+":check","WARNING (dead zone encountered "
                     +helString+" )");
          } else if (!hasDeadZone && verbose >= 6) {
            printOut(vinciaName()+":check","OK (no dead zones "
                     +helString+" )");
          }
        }
      } // End - loop over hels
      
    }
    // Initial-Initial
      
    // Initial-Final: sAK + sjk = sak + saj
    else {
      
      // Two values for sAK
      double sAK[2] = { pow2(50.0), pow2(150.0) };
    
      // Check soft singularity against dipole factor
      if (idj() == 21 || idj() == 22) {
        // Test one symmetric and one asymmetric soft phase-space point
        for (int iTest = 0; iTest <= 1; ++iTest) {
          // Test two sAK values
          for (int isAK = 0; isAK <= 1; ++isAK) {
            // Test invariants
            double yaj = 1.0e-3 * pow(10.0,iTest); // saj/sAK
            double yjk = 1.0e-3 / pow(10.0,iTest); // sjk/sAK
            double sak = sAK[isAK]*(1.0+yjk-yaj); // 1 + yjk = sak/sAK + yaj
            double saj = yaj*sAK[isAK];
            double sjk = yjk*sAK[isAK];
            // Compute eikonal
            double eik = 2.0*sak/saj/sjk;
            // Compute antenna
            double ant = antennaFunction(saj,sjk,sAK[isAK]);
            // Compare
            double ratio = ant/eik;
            if (abs(ratio-1.0) >= 0.001) {
              isOK = false;
              if (verbose >= 1) printOut(vinciaName()+":check","WARNING: FAILED "
                                         "(eikonal "+num2str(iTest,1)+" and sAK = "+
                                         num2str((int)sqrt(sAK[isAK]))+"^2)");
            } else if (verbose >= 6) {
              printOut(vinciaName()+":check","OK (eikonal "+num2str(iTest,1)+
                       " and sAK = "+num2str((int)sqrt(sAK[isAK]))+"^2)");
            }
          }
        }
      }
    
      // Check for collinearity (no helicity APs so far)
      // Test invariants, for a few points along collinear axis
      for (int iTest = 0; iTest <= 2; ++iTest) {
        // Test two sAK values
        for (int isAK = 0; isAK <= 1; ++isAK) {
          // Test invariants
          double y1  = 1.0e-5; // saj/sAK
          double y2  = 0.2 + iTest*0.3; // sjk/sAK
          double s1  = y1*sAK[isAK];
          double s2  = y2*sAK[isAK];
          // Compute AP kernels
          double AP1 = AltarelliParisi(s1,s2,sAK[isAK]);
          double AP2 = AltarelliParisi(s2,s1,sAK[isAK]);
          // Compute antennae
          double ant1 = antennaFunction(s1,s2,sAK[isAK]);
          double ant2 = antennaFunction(s2,s1,sAK[isAK]);
          // Check if only a singularity on one side exists (-> z = -1)
          double zs1 = z1(s1,s2,sAK[isAK]);
          double zs2 = z2(s1,s2,sAK[isAK]);
          if (zs1 != -1.0) {
            double ratio = ant1/AP1;
            if (abs(ratio-1.0) >= 0.01) {
              isOK = false;
              if (verbose >= 1) printOut(vinciaName()+":check","WARNING: FAILED "
                                         "(collinear aj "+num2str(iTest,1)+" and sAK = "+
                                         num2str((int)sqrt(sAK[isAK]))+"^2)");
              if (verbose >= 3) {
                setprecision(6);
                cout<<"    ant  = "<<num2str(ant1,9)<<" y1 = "<<num2str(y1,9)
                    <<" y2 = "<<num2str(y2,9)<<" "<<endl;
                cout<<"    P(z) = "<<num2str(AP1,9)<<"  z = "
                    <<num2str(zs1,9)<<endl;
              }
            } else if (verbose >= 6) {
              printOut(vinciaName()+":check","OK (collinear aj "+num2str(iTest,1)+
                       " and sAK = "+num2str((int)sqrt(sAK[isAK]))+"^2)");
            }
          }
          if (zs2 != -1.0) {
            double ratio = ant2/AP2;
            if (abs(ratio-1.0) >= 0.01) {
              isOK = false;
              if (verbose >= 1) printOut(vinciaName()+":check","WARNING: FAILED "
                                         "(collinear jk "+num2str(iTest,1)+" and sAK = "+
                                         num2str((int)sqrt(sAK[isAK]))+"^2)");
              if (verbose >= 3) {
                setprecision(6);
                cout<<"    ant  = "<<num2str(ant1,9)<<" y1 = "<<num2str(y1,9)
                    <<" y2 = "<<num2str(y2,9)<<" "<<endl;
                cout<<"    P(z) = "<<num2str(AP2,9)<<"  z = "
                    <<num2str(zs2,9)<<endl;
              }
            } else if (verbose >= 6) {
              printOut(vinciaName()+":check","OK (collinear jk "+num2str(iTest,1)+
                       " and sAK = "+num2str((int)sqrt(sAK[isAK]))+"^2)");
            }
          }
        }
      }
      
      // Test all helicity configurations for posivity and dead zones
      int hA(9), hK(9), ha(9), hj(9), hk(9);
      string helString = " 9 9 -> 9 9 9";
      for (int iHel = -1; iHel < 0/*16*/; ++iHel) { // Begin - loop over hels
        if (iHel >= 0) {
          hj = 2*(iHel%2) - 1;
          ha = 2*((iHel/2)%2) - 1;
          hk = 2*((iHel/4)%2) - 1;
          hA = 2*((iHel/8)%2) - 1;
          hK = 2*((iHel/16)%2) - 1;
          helString = num2str(hA,2) + num2str(hK,2) + " ->" +
                      num2str(ha,2) + num2str(hj,2) + num2str(hk,2);
        }
        
        // Gluon emission: do not allow external helicity flips
        if ( idj() == 21 || idj() == 22 )
          if ( hA != ha || hK != hk ) continue;
        
        // Check for positivity and dead zones
        for (int iTest = 0; iTest <= nPointsCheck; ++iTest) {
          // sAK^0.5 between 1 and 1000 GeV
          double sAKnow = pow2(1.0+rndmPtr->flat()*999.0);
          // pick random xA
          double xA  = rndmPtr->flat();
          // sjk between 0 and sAK(1-xA)/xA, saj between 0 and sAK+sjk
          double sjk = rndmPtr->flat()*sAKnow*(1.0-xA)/xA;
          double saj = rndmPtr->flat()*(sAKnow+sjk);
          double sak = sAKnow+sjk-saj;
          // Check that sxy < shhMax
          if (saj > shhMax || sjk > shhMax || sak > shhMax || sAKnow > shhMax) continue;
          // Compute antenna
          double ant = antennaFunction(saj,sjk,sAKnow,hA,hK,ha,hj,hk);
          // Check positivity (strict)
          if (ant < 0.0) {
            isPositive = false;
            if (verbose >= 3) 
              printOut(vinciaName()+":check","ERROR ant("+num2str((int)sqrt(saj))+"^2,"+
                       num2str((int)sqrt(sjk))+"^2,"+num2str((int)sqrt(sAKnow))+"^2 ; "+
                       helString+") < 0");
          } else if (ant > 0.0) {
            isZero = false;
          }
          // Check for dead zones away from phase-space boundary
          if ( (sjk < 0.95*sAKnow*(1.0-xA)/xA) && (sjk > 1.0) && (saj > 1.0)
               && (sak > 1.0) && (ant*sAKnow < deadZoneAvoidance) ) hasDeadZone = true;
        }
        
        // OK?
        isOK = isOK && isPositive;
        
        // Verbose output
        if (!isPositive && verbose >= 1) {
          printOut(vinciaName()+":check","ERROR (ant < 0 encountered "
                   +helString+" )");
        } else if (isPositive && !isZero && verbose >= 6) {
          printOut(vinciaName()+":check","OK (is positive "
                   +helString+" )");
        }
        if (!isZero) {
          if (hasDeadZone && verbose >= 1) {
            printOut(vinciaName()+":check","WARNING (dead zone encountered "
                     +helString+" )");
          } else if (!hasDeadZone && verbose >= 6) {
            printOut(vinciaName()+":check","OK (no dead zones "
                     +helString+" )");
          }
        }
      } // End - loop over hels
      
    }
    // Initial-Final

    return isOK;
    
  }

  //**************************************************************************
  //
  // Inherited Antennae for Initital - Initial

  //--------------------------------------------------------------------------
  // class QQEmitII

  bool QQEmitII::init(bool isSectorIn, int minMaxVar) {
    // Start by doing base class initialization
    isInit = AntennaISR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna
    
    // Return OK
    return isInit;
  }

  double QQEmitII::antennaFunction(double saj, double sjb, double sAB,
                                   int hA, int hB, int ha, int hj, int hb) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;

    // Massless antennae: treat unknown polarizations as unpolarized
    if ( hA != 1 && hA != -1 ) hA = 9;
    if ( hB != 1 && hB != -1 ) hB = 9;
    if ( ha != 1 && ha != -1 ) ha = 9;
    if ( hj != 1 && hj != -1 ) hj = 9;
    if ( hb != 1 && hb != -1 ) hb = 9;
    
    // How many configurations are we averaging over?
    int nAvg = 1;
    if (hA == 9) nAvg *= 2;
    if (hB == 9) nAvg *= 2;
    
    // Shorthands
    double eik     = sAB/saj/sjb;
    double collQa1 = 1.0/saj;
    double collQb1 = 1.0/sjb;
    double collQa2 = sjb/sAB/saj;
    double collQb2 = saj/sAB/sjb;
    double finite  = 0.5*preFacFiniteTermSav/sAB;
    
    // Do helicity sum
    double hSum = 0.0;

    // S-wave antennae (++ and --)
    if ( hA * hB > 0 || hA == 9 || hB == 9) {
      // ++ > +++ && -- > --- antennae (MHV)
      double term1 = eik + finite;
      if ( RH[hA] && RH[hB] && RH[ha] && RH[hj] && RH[hb] ) hSum += term1; 
      if ( LH[hA] && LH[hB] && LH[ha] && LH[hj] && LH[hb] ) hSum += term1;
      
      // ++ > +-+ && -- > -+- antennae (NMHV)
      double term2 = eik + finite + 2.0*collQa1 + 2.0*collQb1
                                  + collQa2 + collQb2;
      if ( RH[hA] && RH[hB] && RH[ha] && LH[hj] && RH[hb] ) hSum += term2;
      if ( LH[hA] && LH[hB] && LH[ha] && RH[hj] && LH[hb] ) hSum += term2;
    }
    
    // P-wave antennae (+- and -+)
    if ( hA * hB < 0 || hA == 9 || hB == 9) {
      // +- > ++- && -+ > --+ antennae
      double term1 = eik + finite + 2.0*collQb1 + collQb2;
      if ( RH[hA] && LH[hB] && RH[ha] && RH[hj] && LH[hb] ) hSum += term1;
      if ( LH[hA] && RH[hB] && LH[ha] && LH[hj] && RH[hb] ) hSum += term1;
      
      // +- > +-- && -+ > -++ antennae
      double term2 = eik + finite + 2.0*collQa1 + collQa2;
      if ( RH[hA] && LH[hB] && RH[ha] && LH[hj] && LH[hb] ) hSum += term2;
      if ( LH[hA] && RH[hB] && LH[ha] && RH[hj] && RH[hb] ) hSum += term2;
    }
    
    // Check to make sure it does the right thing
    if (false) {
      double sab    = sAB+saj+sjb;
             eik    = 2.0*sab*sAB/saj/sjb;
      double collQa = sjb/saj;
      double collQb = saj/sjb;
      double ant    = ( eik + collQa + collQb + preFacFiniteTermSav )/sAB;
      if (hA+hB+ha+hb+hj == 45) 
        if ( abs(ant-(hSum/nAvg)) > 1.0e-10 ) 
          cout << "  sum+avg ant(9,9,9,9,9) != ant(no hels)" << endl;
    }
    
    // Return helicity sum, averaged over initial helicities
    if (!isMinVar) return hSum/nAvg;
    return max( hSum/nAvg, antMinSav );
  }

  double QQEmitII::AltarelliParisi(double saj, double sjb, double sAB, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;
    
    // Using smaller invariant for collinear limit
    double z  = ( saj < sjb ? z1(saj,sjb,sAB) : z2(saj,sjb,sAB) );
    double Q2 = min(saj,sjb);
    double AP = 1.0/z * (1.0+pow2(z))/(1.0-z);
    // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
    AP *= 1.0;
    // ant with quark reproduces full AP
    AP *= 1.0;
        
    return AP/Q2;
  }

  //--------------------------------------------------------------------------
  // class GQEmitII

  bool GQEmitII::init(bool isSectorIn, int minMaxVar) {
    // Start by doing base class initialization
    isInit = AntennaISR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna
    
    // Return OK
    return isInit;
  }

  double GQEmitII::antennaFunction(double saj, double sjb, double sAB, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;
    
    double sab    = sAB+saj+sjb;
    double eik    = 2.0*sab*sAB/saj/sjb;
    double collGa = (2.0*sjb/saj*sab/sAB + 2.0*sjb/saj*sAB/(sab+saj));
    double collQb = saj/sjb;
    double ant    = ( eik + collGa + collQb + preFacFiniteTermSav )/sAB;
    
    if (!isMinVar) return ant;
    return max( ant, antMinSav );
  }

  double GQEmitII::AltarelliParisi(double saj, double sjb, double sAB, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;
    
    // Using smaller invariant for collinear limit
    double z  = ( saj < sjb ? z1(saj,sjb,sAB) : z2(saj,sjb,sAB) );
    double Q2 = min(saj,sjb);
    double AP = 0.0;
    
    if (saj < sjb) { // collinear to initial gluon
      AP = 1.0/z*(pow(z,4.0)+1.0+pow(1-z,4.0))/z/(1.0-z);
      // AP uses alpha/2pi * CA, ant uses alpha/4pi * CA, with CA = 3
      AP *= 2.0;
      // gluon appears in 2 ants, hence ant reproduces half of AP
      AP *= 0.5;
    }
    else { // collinear to initial quark
      AP = 1.0/z * (1.0+pow2(z))/(1.0-z);
      // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
      AP *= 1.0;
      // ant with quark reproduces full AP
      AP *= 1.0;
    }
        
    return AP/Q2;
  }

  //--------------------------------------------------------------------------
  // class GGEmitII

  bool GGEmitII::init(bool isSectorIn, int minMaxVar) {
    // Start by doing base class initialization
    isInit = AntennaISR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna
    
    // Return OK
    return isInit;
  }

  double GGEmitII::antennaFunction(double saj, double sjb, double sAB, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;
    
    double sab    = sAB+saj+sjb;
    double eik    = 2.0*sab*sAB/saj/sjb;
    double collGa = (2.0*sjb/saj*sab/sAB + 2.0*sjb/saj*sAB/(sab+saj));
    double collGb = (2.0*saj/sjb*sab/sAB + 2.0*saj/sjb*sAB/(sab+sjb));
    double ant    = ( eik + collGa + collGb + preFacFiniteTermSav )/sAB;
    
    if (!isMinVar) return ant;
    return max( ant, antMinSav );
  }

  double GGEmitII::AltarelliParisi(double saj, double sjb, double sAB, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;
    
    // Using smaller invariant for collinear limit
    double z  = ( saj < sjb ? z1(saj,sjb,sAB) : z2(saj,sjb,sAB) );
    double Q2 = min(saj,sjb);
    double AP = 1.0/z*(pow(z,4.0)+1.0+pow(1-z,4.0))/z/(1.0-z);
    // AP uses alpha/2pi * CA, ant uses alpha/4pi * CA, with CA = 3
    AP *= 2.0;
    // gluon appears in 2 ants, hence ant reproduces half of AP
    AP *= 0.5;
        
    return AP/Q2;
  }

  //--------------------------------------------------------------------------
  // class QXSplitII
  // Note: Splitting in the forwards sense aka quark backwards evolving
  //       to a gluon and emitting a antiquark in the final state

  bool QXSplitII::init(bool isSectorIn, int minMaxVar) {
    // Start by doing base class initialization
    isInit = AntennaISR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna
    
    // Return OK
    return isInit;
  }

  double QXSplitII::antennaFunction(double saj, double sjb, double sAB, 
                                    int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;
    
    double sab = sAB+saj+sjb;
    double ant = ( -2.0*sjb*sAB/saj/(sab-saj) + sab/saj + preFacFiniteTermSav )/sAB;
    
    if (!isMinVar) return ant;
    return max( ant, antMinSav );
  }

  double QXSplitII::AltarelliParisi(double saj, double sjb, double sAB, 
                                    int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;
    
    // There is only the saj collinear limit
    double z  = z1(saj,sjb,sAB);
    double Q2 = saj;
    double AP = 1.0/z * (pow2(z)+pow2(1-z));
    // AP uses alpha/2pi * TR/2, ant uses alpha/4pi * TR, with TR = 1
    AP *= 1.0;
    // ant with quark reproduces full AP
    AP *= 1.0;
        
    return AP/Q2;
  }

  //--------------------------------------------------------------------------
  // class GXConvII
  // Note: Gluon evolves backwards into a quark and emits a quark
  //       in the final state

  bool GXConvII::init(bool isSectorIn, int minMaxVar) {
    // Start by doing base class initialization
    isInit = AntennaISR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna
    
    // Return OK
    return isInit;
  }

  double GXConvII::antennaFunction(double saj, double sjb, double sAB, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;
    
    double sab = sAB+saj+sjb;
    double ant = ( 0.5/saj*(sjb*sjb+sab*sab)/sAB + preFacFiniteTermSav )/sAB;
    
    if (!isMinVar) return ant;
    return max( ant, antMinSav );
  }

  double GXConvII::AltarelliParisi(double saj, double sjb, double sAB, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;
    
    // There is only the saj collinear limit
    double z  = z1(saj,sjb,sAB);
    double Q2 = saj;
    double AP = 1.0/z * (1.0+pow2(1.0-z))/z;
    // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
    AP *= 1.0;
    // gluon appears in 2 ants, hence ant reproduces half of AP
    AP *= 0.5;
        
    return AP/Q2;
  }

  //**************************************************************************
  //
  // Inherited Antennae for Initital - Final

  //--------------------------------------------------------------------------
  // class QQEmitIF

  bool QQEmitIF::init(bool isSectorIn, int minMaxVar) {
    // Start by doing base class initialization
    isInit = AntennaISR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna
    
    // Return OK
    return isInit;
  }

  double QQEmitIF::antennaFunction(double saj, double sjk, double sAK, 
                                   int hA, int hK, int ha, int hj, int hk) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;

    // Massless antennae: treat unknown polarizations as unpolarized
    if ( hA != 1 && hA != -1 ) hA = 9;
    if ( hK != 1 && hK != -1 ) hK = 9;
    if ( ha != 1 && ha != -1 ) ha = 9;
    if ( hj != 1 && hj != -1 ) hj = 9;
    if ( hk != 1 && hk != -1 ) hk = 9;
    
    // How many configurations are we averaging over?
    int nAvg = 1;
    if (hA == 9) nAvg *= 2;
    if (hK == 9) nAvg *= 2;
    
    // Shorthands
    double eik     = sAK/saj/sjk;
    double collQa1 = 1.0/saj;
    double collQk1 = 1.0/sjk;
    double collQa2 = sjk/sAK/saj;
    double collQk2 = saj/sAK/sjk;
    double finite  = 0.5*preFacFiniteTermSav/sAK;
    
    // Do helicity sum
    double hSum = 0.0;

    // S-wave antennae (++ and --)
    if ( hA * hK > 0 || hA == 9 || hK == 9) {
      // ++ > +++ && -- > --- antennae (MHV)
      double term1 = eik + finite;
      if ( RH[hA] && RH[hK] && RH[ha] && RH[hj] && RH[hk] ) hSum += term1; 
      if ( LH[hA] && LH[hK] && LH[ha] && LH[hj] && LH[hk] ) hSum += term1;
      
      // ++ > +-+ && -- > -+- antennae (NMHV)
      double term2 = eik + finite + 2.0*collQa1 - 2.0*collQk1
                                  + collQa2 + collQk2;
      if ( RH[hA] && RH[hK] && RH[ha] && LH[hj] && RH[hk] ) hSum += term2;
      if ( LH[hA] && LH[hK] && LH[ha] && RH[hj] && LH[hk] ) hSum += term2;
    }
    
    // P-wave antennae (+- and -+)
    if ( hA * hK < 0 || hA == 9 || hK == 9) {
      // +- > ++- && -+ > --+ antennae
      double term1 = eik + finite - 2.0*collQk1 + collQk2;
      if ( RH[hA] && LH[hK] && RH[ha] && RH[hj] && LH[hk] ) hSum += term1;
      if ( LH[hA] && RH[hK] && LH[ha] && LH[hj] && RH[hk] ) hSum += term1;
      
      // +- > +-- && -+ > -++ antennae
      double term2 = eik + finite + 2.0*collQa1 + collQa2;
      if ( RH[hA] && LH[hK] && RH[ha] && LH[hj] && LH[hk] ) hSum += term2;
      if ( LH[hA] && RH[hK] && LH[ha] && RH[hj] && RH[hk] ) hSum += term2;
    }
    
    // Check to make sure it does the right thing
    if (false) {
      double sak    = sAK+sjk-saj;
             eik    = 2.0*sak*sAK/saj/sjk;
      double collQa = sjk/saj;
      double collQk = saj/sjk;
      double ant    = ( eik + collQa + collQk + preFacFiniteTermSav )/sAK;
      if (hA+hK+ha+hk+hj == 45) 
        if ( abs(ant-(hSum/nAvg)) > 1.0e-10 ) 
          cout << "  sum+avg ant(9,9,9,9,9) != ant(no hels)" << endl;
    }
    
    // Return helicity sum, averaged over initial helicities
    if (!isMinVar) return hSum/nAvg;
    return max( hSum/nAvg, antMinSav );
  }

  double QQEmitIF::AltarelliParisi(double saj, double sjk, double sAK, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
    // Using smaller invariant for collinear limit
    double z  = ( saj < sjk ? z1(saj,sjk,sAK) : z2(saj,sjk,sAK) );
    double Q2 = min(saj,sjk);
    double AP = 0.0;
    
    if (saj < sjk) { // collinear to initial quark
      AP = 1.0/z * (1.0+pow2(z))/(1.0-z);
      // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
      AP *= 1.0;
      // ant with quark reproduces full AP
      AP *= 1.0;
    }
    else { // collinear to final quark
      AP = (1.0+pow2(z))/(1.0-z);
      // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
      AP *= 1.0;
      // ant with quark reproduces full AP
      AP *= 1.0;
    }
        
    return AP/Q2;
  }

  //--------------------------------------------------------------------------
  // class QGEmitIF

  bool QGEmitIF::init(bool isSectorIn, int minMaxVar) {
    // Start by doing base class initialization
    isInit = AntennaISR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna
    
    // Return OK
    return isInit;
  }

  double QGEmitIF::antennaFunction(double saj, double sjk, double sAK, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
    double sak    = sAK+sjk-saj;
    double eik    = 2.0*sak*sAK/saj/sjk;
    double collQa = sjk/saj;
    double collGk = saj/sjk*sak/sAK;
    double ant    = ( eik + collQa + collGk + preFacFiniteTermSav )/sAK;
    
    if (!isMinVar) return ant;
    return max( ant, antMinSav );
  }

  double QGEmitIF::AltarelliParisi(double saj, double sjk, double sAK, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
    // Using smaller invariant for collinear limit
    double z  = ( saj < sjk ? z1(saj,sjk,sAK) : z2(saj,sjk,sAK) );
    double Q2 = min(saj,sjk);
    double AP = 0.0;
    
    if (saj < sjk) { // collinear to initial quark
      AP = 1.0/z * (1.0+pow2(z))/(1.0-z);
      // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
      AP *= 1.0;
      // ant with quark reproduces full AP
      AP *= 1.0;
    }
    else { // collinear to final gluon
      AP = (2.0*z/(1.0-z)+z*(1.0-z));
      // AP uses alpha/2pi * CA, ant uses alpha/4pi * CA, with CA = 3
      AP *= 2.0;
      // gluon appears in 2 ants, hence ant reproduces half of AP
      AP *= 0.5;
    }
        
    return AP/Q2;
  }

  //--------------------------------------------------------------------------
  // class GQEmitIF

  bool GQEmitIF::init(bool isSectorIn, int minMaxVar) {
    // Start by doing base class initialization
    isInit = AntennaISR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna
    
    // Return OK
    return isInit;
  }

  double GQEmitIF::antennaFunction(double saj, double sjk, double sAK, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
    double sak    = sAK+sjk-saj;
    double eik    = 2.0*sak*sAK/saj/sjk;
    double collGa = (2.0*sjk/saj*sak/sAK + 2.0*sAK*sjk/saj/(sAK+sjk));
    double collQk = saj/sjk;
    double ant    = ( eik + collGa + collQk + preFacFiniteTermSav )/sAK;
    
    if (!isMinVar) return ant;
    return max( ant, antMinSav );
  }

  double GQEmitIF::AltarelliParisi(double saj, double sjk, double sAK, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
    // Using smaller invariant for collinear limit
    double z  = ( saj < sjk ? z1(saj,sjk,sAK) : z2(saj,sjk,sAK) );
    double Q2 = min(saj,sjk);
    double AP = 0.0;
    
    if (saj < sjk) { // collinear to initial gluon
      AP = 1.0/z*(pow(z,4.0)+1.0+pow(1-z,4.0))/z/(1.0-z);
      // AP uses alpha/2pi * CA, ant uses alpha/4pi * CA, with CA = 3
      AP *= 2.0;
      // gluon appears in 2 ants, hence ant reproduces half of AP
      AP *= 0.5;
    }
    else { // collinear to final quark
      AP = (1.0+pow2(z))/(1.0-z);
      // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
      AP *= 1.0;
      // ant with quark reproduces full AP
      AP *= 1.0;
    }
        
    return AP/Q2;
  }

  //--------------------------------------------------------------------------
  // class GGEmitIF

  bool GGEmitIF::init(bool isSectorIn, int minMaxVar) {
    // Start by doing base class initialization
    isInit = AntennaISR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna
    
    // Return OK
    return isInit;
  }

  double GGEmitIF::antennaFunction(double saj, double sjk, double sAK, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
    double sak    = sAK+sjk-saj;
    double eik    = 2.0*sak*sAK/saj/sjk;
    double collGa = (2.0*sjk/saj*sak/sAK + 2.0*sAK*sjk/saj/(sAK+sjk));
    double collGk = saj/sjk*sak/sAK;
    double ant    = ( eik + collGa + collGk + preFacFiniteTermSav )/sAK;
    
    if (!isMinVar) return ant;
    return max( ant, antMinSav );
  }

  double GGEmitIF::AltarelliParisi(double saj, double sjk, double sAK, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
    // Using smaller invariant for collinear limit
    double z  = ( saj < sjk ? z1(saj,sjk,sAK) : z2(saj,sjk,sAK) );
    double Q2 = min(saj,sjk);
    double AP = 0.0;
    
    if (saj < sjk) { // collinear to initial gluon
      AP = 1.0/z*(pow(z,4.0)+1.0+pow(1-z,4.0))/z/(1.0-z);
      // AP uses alpha/2pi * CA, ant uses alpha/4pi * CA, with CA = 3
      AP *= 2.0;
      // gluon appears in 2 ants, hence ant reproduces half of AP
      AP *= 0.5;
    }
    else { // collinear to final gluon
      AP = (2.0*z/(1.0-z)+z*(1.0-z));
      // AP uses alpha/2pi * CA, ant uses alpha/4pi * CA, with CA = 3
      AP *= 2.0;
      // gluon appears in 2 ants, hence ant reproduces half of AP
      AP *= 0.5;
    }
        
    return AP/Q2;
  }

  //--------------------------------------------------------------------------
  // class QXSplitIF
  // Note: Splitting in the forwards sense aka quark backwards evolving
  //       to a gluon and emitting a antiquark in the final state

  bool QXSplitIF::init(bool isSectorIn, int minMaxVar) {
    // Start by doing base class initialization
    isInit = AntennaISR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna
    
    // Return OK
    return isInit;
  }

  double QXSplitIF::antennaFunction(double saj, double sjk, double sAK, 
                                    int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
    double sak = sAK+sjk-saj;
    double ant = ( sak/saj - 2.0*sjk/saj*(sAK-saj)/(sAK+sjk)
                           + preFacFiniteTermSav )/sAK;
    
    if (!isMinVar) return ant;
    return max( ant, antMinSav );
  }

  double QXSplitIF::AltarelliParisi(double saj, double sjk, double sAK, 
                                    int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
    // There is only the saj collinear limit
    double z  = z1(saj,sjk,sAK);
    double Q2 = saj;
    double AP = 1.0/z * (pow2(z)+pow2(1-z));
    // AP uses alpha/2pi * TR/2, ant uses alpha/4pi * TR, with TR = 1
    AP *= 1.0;
    // ant with quark reproduces full AP
    AP *= 1.0;
        
    return AP/Q2;
  }

  //--------------------------------------------------------------------------
  // class GXConvIF
  // Note: Gluon evolves backwards into a quark and emits a quark
  //       in the final state

  bool GXConvIF::init(bool isSectorIn, int minMaxVar) {
    // Start by doing base class initialization
    isInit = AntennaISR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna
    
    // Return OK
    return isInit;
  }

  double GXConvIF::antennaFunction(double saj, double sjk, double sAK, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
    double sak = sAK+sjk-saj;
    double ant = ( 0.5/saj*(sak*sak+sjk*sjk)/sAK + preFacFiniteTermSav )/sAK;
    
    if (!isMinVar) return ant;
    return max( ant, antMinSav );
  }

  double GXConvIF::AltarelliParisi(double saj, double sjk, double sAK, 
                                   int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
    // There is only the saj collinear limit
    double z  = z1(saj,sjk,sAK);
    double Q2 = saj;
    double AP = 1.0/z * (1.0+pow2(1.0-z))/z;
    // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
    AP *= 1.0;
    // gluon appears in 2 ants, hence ant reproduces half of AP
    AP *= 0.5;
        
    return AP/Q2;
  }

  //--------------------------------------------------------------------------
  // class XGSplitIF
  // Note: Gluon splitting in the final state

  bool XGSplitIF::init(bool isSectorIn, int minMaxVar) {
    // Start by doing base class initialization
    isInit = AntennaISR::init(isSectorIn, minMaxVar);
    
    // Then initialize parts specific to this antenna
    
    // Return OK
    return isInit;
  }

  double XGSplitIF::antennaFunction(double saj, double sjk, double sAK, 
                                    int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
    double sak = sAK+sjk-saj;
    double ant = ( 0.5/sjk*(sak*sak+saj*saj)/sAK + preFacFiniteTermSav )/sAK;
    
    if (!isMinVar) return ant;
    return max( ant, antMinSav );
  }

  double XGSplitIF::AltarelliParisi(double saj, double sjk, double sAK, 
                                    int, int, int, int, int) {
    // Sanity check. Require positive invariants.
    if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
    // There is only the sjk collinear limit
    double z  = z2(saj,sjk,sAK);
    double Q2 = sjk;
    double AP = (pow2(z)+pow2(1-z));
    // AP uses alpha/2pi * TR/2, ant uses alpha/4pi * TR, with TR = 1
    AP *= 1.0;
    // gluon appears in 2 ants, hence ant reproduces half of AP
    AP *= 0.5;
        
    return AP/Q2;
  }

} // End Vincia Namespace
  
