// TimeAntShower.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2016 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the TimeAntShower class, 
// as follows:
// 1) inherited member functions from the Pythia 8 TimeShower classs
// 2) non-inherited member functions for the TimeAntShower class

// Include the Vincia headers 
#include "Vincia/TimeAntShower.h"
#include "Vincia/SpaceAntShower.h"

namespace Vincia {
  
  //**********************************************************************
  
  // The TimeAntShower class and its members
  
  //**********************************************************************
  
  // BEGIN INHERITED FUNCTIONS FOR INTERFACE WITH PYTHIA8
  //  : init     : global shower initialization function
  //  : shower   : perform shower off partonic system
  //  : prepare  : prepare partonic system for showering
  //  : update   : update the partonic system
  //  : rescatterUpdate : purely dummy
  //  : rescatterPropagateRecoil : purely dummy
  //  : pTnext   : generate next scale in downwards evolution
  //  : branch   : perform a branching, store results in event record
  
  //*********
  
  // Initialize Shower
  
  void TimeAntShower::init( BeamParticle* beamAPtrIn, 
                            BeamParticle* beamBPtrIn) {
    
    // Check if already initialized 
    if (isInit)
      if (settingsPtr->word("Merging:Process").compare("void") == 0)
        return;
    
    // Event counters and debugging
    nAccepted       = -1; // Set to -1, so prepare works properly
    nSelected       = -1;
    nVetoUserHooks  = 0;
    nFailHadLevel   = 0;
    nCallPythiaNext = 0;
    startDebug      = settingsPtr->mode("Vincia:startDebug");

    // Verbose level and version number
    verbose            = settingsPtr->mode("Vincia:verbose");
    version            = settingsPtr->parm("Vincia:versionNumber");
    
    // Verbose output
    if (verbose >= 6) 
      printOut("TimeAntShower::init","begin --------------");
    
    // Store input pointers for future use. 
    beamAPtr     = beamAPtrIn;
    beamBPtr     = beamBPtrIn;
    
    // Possibility to allow user veto of emission step.
    hasUserHooks      = (userHooksPtr != 0);
    canVetoEmission   = (hasUserHooks && userHooksPtr->canVetoFSREmission());

    // Shower parameters
    evolutionType     = settingsPtr->mode("Vincia:evolutionType");
    // Decide whether to force strong (evolution) ordering or not
    orderingMode      = settingsPtr->mode("Vincia:orderingMode");
    orderingModeMPI   = settingsPtr->mode("Vincia:orderingModeMPI");
    // Number of active quark flavours
    nGluonToQuarkF    = settingsPtr->mode("Vincia:nGluonToQuarkF");
    
    // Perturbative cutoff
    cutoffType         = settingsPtr->mode("Vincia:cutoffTypeF");
    if (cutoffType == 0) cutoffType = evolutionType;
    if (evolutionType == 3 && cutoffType != evolutionType) {
      printOut("TimeAntShower::init",
               "Warning - forced cutoffType = 1 to regulate energy ordering");
      cutoffType = 1;
    }
    cutoffScale        = settingsPtr->parm("Vincia:cutoffScaleFF");
    
    // Normalization factors for types 1 and 2 evolution variables
    pTnormalization    = settingsPtr->parm("Vincia:pTnormalization");
    mDnormalization    = settingsPtr->parm("Vincia:mDnormalization");
    
    // Number of flavours to be treated as massless
    // (Can be made user-specifiable in future if desired)
    nFlavMassless      = 5;    
    nFlavZeroMass      = settingsPtr->mode("Vincia:nFlavZeroMassF");
    massCorrectionEmit = settingsPtr->mode("Vincia:massCorrectionEmitF");
    massCorrectionSplit= settingsPtr->mode("Vincia:massCorrectionSplitF");
    
    // Ariadne Factor
    doPari             = settingsPtr->flag("Vincia:GluonSplittingCorrection");

    // Mass windows : specifies thresholds for each flavour.
    mt          = vinComPtr->mt;
    mb          = vinComPtr->mb;
    mc          = vinComPtr->mc;
    ms          = vinComPtr->ms;
    // Derived scales
    mtb         = sqrt(mt*mb);
    
    // Set shower alphaS pointer
    alphaSptr          = &vinComPtr->alphaStrong;
    alphaSCMWptr       = &vinComPtr->alphaStrongCMW;
    // Defaults
    alphaSdefPtr       = &vinComPtr->alphaStrongDef;
    alphaSCMWdefPtr    = &vinComPtr->alphaStrongDefCMW;
    
    // AlphaS parameters
    alphaSvalue        = settingsPtr->parm("Vincia:alphaSvalue");
    CMWtype            = settingsPtr->mode("Vincia:CMWtypeFF");
    alphaSorder        = settingsPtr->mode("Vincia:alphaSorder");
    aSkMu2             = pow2(settingsPtr->parm("Vincia:alphaSkMuF"));
    aSkMu2Split        = pow2(settingsPtr->parm("Vincia:alphaSkMuSplitF"));
    alphaSmax          = settingsPtr->parm("Vincia:alphaSmax"); 
    alphaSmuFreeze     = settingsPtr->parm("Vincia:alphaSmuFreeze");
    mu2freeze          = pow2(alphaSmuFreeze);
    // Defaults
    alphaSvalueDef     = settingsPtr->parmDefault("Vincia:alphaSvalue");
    CMWtypeDef         = settingsPtr->modeDefault("Vincia:CMWtypeFF");
    alphaSorderDef     = settingsPtr->modeDefault("Vincia:alphaSorder");
    aSkMu2Def          = pow2(settingsPtr->parmDefault("Vincia:alphaSkMuF"));
    aSkMu2SplitDef     = pow2(settingsPtr->parmDefault("Vincia:alphaSkMuSplitF"));
    alphaSmaxDef       = settingsPtr->parmDefault("Vincia:alphaSmax"); 
    alphaSmuFreezeDef  = settingsPtr->parmDefault("Vincia:alphaSmuFreeze");
    mu2freezeDef       = pow2(alphaSmuFreezeDef);
    
    // Smallest allow scale
    alphaSmuMin        = max( alphaSmuFreeze, 1.05 * max( alphaSptr->Lambda3(),
                                                          alphaSCMWptr->Lambda3() ) );
    mu2min             = pow2(alphaSmuMin);
    // Default
    alphaSmuMinDef     = max( alphaSmuFreezeDef, 1.05 * max( alphaSdefPtr->Lambda3(),
                                                             alphaSCMWdefPtr->Lambda3() ) );
    mu2minDef          = pow2(alphaSmuMinDef);
    
    // Check largest numerical value of alphaS we can have
    if (alphaSorder >= 1)
      alphaSmax = min( alphaSmax, min( alphaSptr->alphaS(mu2min),
                                       alphaSCMWptr->alphaS(mu2min) ) );
    // Default
    if (alphaSorderDef >= 1)
      alphaSmaxDef = min( alphaSmaxDef, min( alphaSdefPtr->alphaS(mu2minDef),
                                             alphaSCMWdefPtr->alphaS(mu2minDef) ) );
        
    // Initialize Trial Generators
    // Note: additional trial generators can be added and initialized by 
    // adding them in the header and adding an additional push_back here.
    vector<TrialGeneratorFSR*> trialGenerators;
    trialGenerators.push_back(&trialSoft);
    trialGenerators.push_back(&trialCollI);
    trialGenerators.push_back(&trialCollK);
    trialGenerators.push_back(&trialSplitI);
    trialGenerators.push_back(&trialSplitK);
    for (int iGen=0; iGen < int(trialGenerators.size()); ++iGen) {
      trialGenerators[iGen]->initPtr(rndmPtr,settingsPtr);      
      double evNormFactor = 1.0;
      if (evolutionType == 1) evNormFactor = pTnormalization;
      else if (evolutionType == 2) evNormFactor = mDnormalization;
      trialGenerators[iGen]->init(evolutionType,evNormFactor);
    }
    
    // Uncertainty Bands
    uncertaintyBands       = settingsPtr->flag("Vincia:uncertaintyBands");
    kMu2Var                = pow2(settingsPtr->parm("Vincia:uncertaintyBandsKmu"));
    uncertaintyBandsMaxDev = settingsPtr->parm("Vincia:uncertaintyBandsMaxDev");
    
    // Decide whether to impose sector ordering or not
    sectorShower       = settingsPtr->flag("Vincia:sectorShower");
    useSectorTerms     = settingsPtr->flag("Vincia:useSectorTerms");
    if (sectorShower) { // TODO no sector shower at the moment
      printOut("TimeAntShower::init","warning; switching sectorShower off "
               "since disabled at the moment.");
      settingsPtr->flag("Vincia:sectorShower",false);
      sectorShower = false;
    }
    if ( !settingsPtr->flag("Vincia:useCollinearTerms") && sectorShower) {
      printOut("TimeAntShower::init","Note - useCollinearTerms = off"
               ", so also forcing useSectorTerms = off");  
      useSectorTerms = false;      
    }
    
    // RESTRICT SECTOR SHOWER FLEXIBILITY
    if (sectorShower) {
      if (evolutionType != 1) {
        printOut("TimeAntShower::init",
                 "Note - forcing evolutionType = 1 for sector showers");
        evolutionType = 1;
      }
      if (orderingMode != 0) {
        printOut("TimeAntShower::init",
                 "Note - forcing orderingMode = 0 for sector showers");
        orderingMode = 0;
      }
    }
    
    // Uncertainty Bands not yet implemented for sector showers
    if (sectorShower && uncertaintyBands) {
      uncertaintyBands = false;
      settingsPtr->flag("Vincia:uncertaintyBands",false);
      if (verbose >= 2) printOut("TimeAntShower::init","uncertainty bands "
                                 "are switched off for sector showers.");
    }
    
    // Global flag for polarization
    helicityShower     = settingsPtr->flag("Vincia:helicityShower");
    if (helicityShower) { // TODO no helicity shower at the moment
      printOut("TimeAntShower::init","warning; switching helicityShower off "
               "since disabled at the moment.");
      settingsPtr->flag("Vincia:helicityShower",false);
      helicityShower = false;
    }
    
    // local flag for polarization and mass. Default values
    isMassless = true;    
    usePolarizationNow = helicityShower;

    // Hyperjet compatibility mode on/off
    hyperjet           = settingsPtr->flag("Vincia:hyperjet");

    // Colour reconnections
    CRmode        = settingsPtr->mode("Vincia:CRmode");
    if (CRmode > 0) { // TODO no CR at the moment
      printOut("TimeAntShower::init","warning; forcing CRmode == 0 "
               "since disabled at the moment.");
      settingsPtr->mode("Vincia:CRmode",0);
      CRmode = 0;
    }
    CRjunctions   = settingsPtr->flag("Vincia:CRjunctions");
    
    // Matching settings
    matchingLO         = settingsPtr->mode("Vincia:matchingLO");
    meCorrections      = matchingLO >= 1;
    matchingNLO        = min(matchingLO,
                             settingsPtr->mode("Vincia:matchingNLO"));
    if (matchingNLO > 0) { // TODO no NLO matching at the moment
      printOut("TimeAntShower::init","warning; forcing matchingNLO == 0 "
               "since disabled at the moment.");
      settingsPtr->mode("Vincia:matchingNLO",0);
      matchingNLO = 0;
    }
    // Matching beyond first order can only be done for smoothly ordered
    // showers. Switch off matching beyond first order if user has selected
    // strong ordering (since the goal is then presumably to look at pure
    // showers anyway).
    if (orderingMode == 1 && matchingLO >= 2) {
      printOut("TimeAntShower::init","warning; forcing matchingLO <= 1 "
               "for strongly ordered showers");
      matchingLO  = min(matchingLO,1);
      matchingNLO = min(matchingNLO,1);
    }
    // Similar if no hard jets are allowed.
    if (!settingsPtr->flag("Vincia:hardJets") && matchingLO >= 2) {
      printOut("TimeAntShower::init","warning: forcing matchingLO = 1 "
               "as no hard jets are allowed");
      matchingLO = min(matchingLO,1);
    }
    if (!settingsPtr->flag("Vincia:hardJetsQCD") && matchingLO >= 1) {
      printOut("TimeAntShower::init","warning: forcing matchingLO = 0 "
               "as no hard jets are allowed");
      matchingLO = 0;
    }
    matchingFullColor  = settingsPtr->flag("Vincia:matchingFullColor");
    matchingRegOrder   = settingsPtr->mode("Vincia:matchingRegOrder");
    // Matching scale can only be imposed at highest matched multiplicity
    // or not at all. (Matching assumes all previous multiplicities were
    // matched over all of phase space.)
    if (matchingRegOrder <= matchingLO-1) {
      printOut("TimeAntShower::init","warning; forcing matchingRegOrder >= "
               "matchingLO");
      matchingRegOrder = max(matchingLO,matchingRegOrder);
    }
    // NLO matching: virtual correction assumes real corrections are 
    // matched over all of phase space.  
    if (matchingRegOrder <= matchingNLO) {
      printOut("TimeAntShower::init","warning; forcing matchingRegOrder >= "
               "matchingNLO + 1");
      matchingRegOrder = max(matchingNLO+1,matchingRegOrder);
    }
    matchingRegShape   = settingsPtr->mode("Vincia:matchingRegShape");
    matchingScaleIsAbs = settingsPtr->flag("Vincia:matchingRegScaleIsAbsolute");
    matchingScale      = settingsPtr->parm("Vincia:matchingRegScale");
    matchingScaleRatio = settingsPtr->parm("Vincia:matchingRegScaleRatio");
    matchingIRcutoff   = settingsPtr->parm("Vincia:matchingIRcutoff");
    Hinterference      = settingsPtr->flag("Vincia:Hinterference");
    isMGOn = settingsPtr->flag("Vincia:MGInterface");
    // Initialize matching pointer
    matchingPtr->init(uncertaintyBands ? 11 : 1);
    
    // Number of uncertainty variations (including USR)
    // Hardcoded here, no sense to make user specifiable, since user 
    // can see a table of which variations are being performed, and can
    // then just choose the weight sets he or she wants to consider. 
    nWeights         = (uncertaintyBands ? 11 : 1);
    allowReweighting = settingsPtr->flag("Vincia:allowReweighting");
    
    // Resize Paccept and PME to the relevant number of elements
    Paccept.resize(nWeights);
    PME.resize(nWeights);
    
    // Hard system(s) and MEclass
    hardSys.clear();
    hardSys.resize(0);
    MEclass.clear();
    MEclass.resize(0);
    
    // Statistics
    nTrialsSum         = 0;
    nTrialsVeto        = 0;
    nTrialsHull        = 0;
    nTrialsHullMassive = 0;
    nTrialsCutM        = 0;
    nTrialsCutQ        = 0;
    nClosePSforHQ      = 0;
    nMCviolShower       = 0;
    nMCviolDidMEC       = 0;
    nMCviolPlannedMEC   = 0;
    maxMCviolShower     = 1.0;
    maxMCviolDidMEC     = 1.0;
    maxMCviolPlannedMEC = 1.0;
    nNegPaShower       = 0;
    nNegPaDidMEC       = 0;
    nNegPaPlannedMEC   = 0;
    maxNegPaShower     = 0.0;
    maxNegPaDidMEC     = 0.0;
    maxNegPaPlannedMEC = 0.0;
    
    // Internal Histograms
    if (verbose >= 2) initVinciaHistos();
    
    // Initialize headroom factors (optimized dynamically later)
    optimizeHeadRoom = settingsPtr->flag("Vincia:MCoptimization");
    iRegDependance   = false;
    initHeadRoom(1.0, 1.0, 1.0, 1.0);
    pAcceptFuss      = settingsPtr->parm("Vincia:PacceptFuss");
    pAcceptPT0       = settingsPtr->parm("Vincia:PacceptPT0");
    nReportMCviol    = 5;
    
    // Initialize properties of hard system
    isQCD        = false;
    Q2fac        = 1.0;
    Qfac         = 1.0;
    hardJets     = settingsPtr->flag("Vincia:hardJets");
    hardJetsQCD  = settingsPtr->flag("Vincia:hardJetsQCD");
    enhanceQ2fac = pow2(settingsPtr->parm("Vincia:QmaxEnhanceF"));
    
    // Merging?
    doMergeSav     = !(settingsPtr->word("Merging:Process").compare("void")
                       == 0);
    // (External) Merging only with strong ordering, no hard jets, no MECs, no Pari
    if (doMergeSav) {
      if (orderingMode != 1) {
        printOut("TimeAntShower::init","warning: forcing orderingMode = 1 "
                 "as (external) merging requested");
        orderingMode = 1;
      }
      if (matchingLO != 0) {
        printOut("TimeAntShower::init","warning: forcing matchingLO = 0 "
                 "as (external) merging requested");
        matchingLO = 0;
      }
      if (hardJets || hardJetsQCD) {
        printOut("TimeAntShower::init","warning: switching hardJets(QCD) off "
                 "as (external) merging requested");
        hardJets    = false;
        hardJetsQCD = false;
      }
      if (doPari) {
        printOut("TimeAntShower::init","warning: switching GluonSplittingCorrection off "
                 "as (external) merging requested");
        doPari = false;
      }
    }
    
    // Set ordering mode
    orderingModeSav = orderingMode;
    if (orderingModeSav == 3) {
      // Set to smooth / strong ordering depending on matchingLO
      if (matchingLO > 0) orderingMode = 2;
      else {
        orderingModeSav = 1;
        orderingMode    = 1;
      }
    }
    
    // Reset saved particle state
    particleNrs.resize(0);
    particles.resize(0);    

    // Initialize USR antennae. Not done before to make sure we catch if
    // something was changed in Settings before.
    if (verbose >= 6)
      cout<<" TimeAntShower(): initializing DEF antennaSet"<<endl;
    antennaSets[0]->init(false);
    antennaSetsSct[0]->init(true);

    // Print VINCIA header and list of parameters
    if (verbose >= 1 && !headerIsPrinted) header();
    
    // Verbose output
    if (verbose >= 6) 
      printOut("TimeAntShower::init","end --------------");
    
    // Is initialized now
    isInit = true;

  }
  
  //*********
  
  // Top-level routine to do a full time-like shower in resonance decay.
  
  int TimeAntShower::shower(int iBeg, int iEnd, Event& event, double pTmax, 
                            int nBranchMax) {
    
    // Verbose output
    if (verbose >= 6) 
      printOut("TimeAntShower::shower","begin --------------");
    
    // Add new system, automatically with two empty beam slots.
    int iSys = partonSystemsPtr->addSys();
    
    // Verbose output
    if (verbose >= 8) 
      printOut("TimeAntShower::shower","preparing to shower. System no. "
               +num2str(iSys));
    
    // Loop over allowed range to find all final-state particles.
    Vec4 pSum;
    for (int i = iBeg; i <= iEnd; ++i) if (event[i].isFinal()) {
      partonSystemsPtr->addOut( iSys, i);
      pSum += event[i].p();
    }
    partonSystemsPtr->setSHat( iSys, pSum.m2Calc() );
    
    // Let prepare routine do the setup.    
    prepare(iSys, event, false);
    
    // Begin evolution down in pT from hard pT scale. 
    int nBranch = 0;
    do {
      double pTtimes = pTnext(event, pTmax, 0.);
      
      // Do a final-state emission (if allowed).
      if (pTtimes > 0.) {
        if (branch(event)) ++nBranch; 
        pTmax = pTtimes;
      }
      
      // Keep on evolving until nothing is left to be done.
      else pTmax = 0.;
    } while (pTmax > 0. && (nBranchMax <= 0 || nBranch < nBranchMax)); 
    
    // Return number of emissions that were performed.
    return nBranch;
    
  }
  
  //*********
  
  // Prepare system for evolution.
  // Note: this function so far assumes that the system is input
  // in colour order. A generalization would be needed to organize
  // systems that are not in colour order.

  void TimeAntShower::prepare(int iSys, Event& event, bool limitPTmaxIn) {
    
    // Verbose output
    if (verbose >= 6)
      printOut("TimeAntShower::prepare","begin --------------");
    
    // If verbose, list event before showering
    if (verbose >= 4) {
      if (verbose >= 5) event.list();
      printOut("TimeAntShower::prepare","preparing system no. "
               +num2str(iSys)+" for showering. Size = "
               +num2str(partonSystemsPtr->sizeOut(iSys))+
               " limitPTmax = "+bool2str(limitPTmaxIn));
    }
    
    // Counter for accepted events
    nAccepted = max(nAccepted, infoPtr->nAccepted());
    
    // First time prepare is called for an event
    bool firstCallInEvent = (iSys == 0);
    // getCounter(3) number of times a Pythia::next() call has been begun.
    int nCallPythiaNextNow = infoPtr->getCounter(3);
    if (nCallPythiaNextNow > nCallPythiaNext) {
      nCallPythiaNext  = nCallPythiaNextNow;
      nVetoUserHooks   = 0;
      nFailHadLevel    = 0;
      firstCallInEvent = true;
    }
    // Last event got accepted.
    long nSelectedNow = infoPtr->nSelected();
    if (nSelectedNow > nSelected) {
      nSelected        = nSelectedNow;
      nVetoUserHooks   = 0;
      nFailHadLevel    = 0;
      firstCallInEvent = true;
    }
    // getCounter(10) number of times the selection of a new hard process has
    // been begun. = 1 if no user hooks veto, > 1 if user hooks veto.
    int nVetoUserHooksNow = (infoPtr->getCounter(10)-1);
    if (nVetoUserHooksNow > nVetoUserHooks) {
      nVetoUserHooks   = nVetoUserHooksNow;
      firstCallInEvent = true; 
    }
    // getCounter(14) number of times the loop over parton- and hadron-level processing 
    // has begun for a hard process. = 1 if everything after user hooks veto is
    // succesfull, > 1 eg if hadron level fails.
    int nFailHadLevelNow = (infoPtr->getCounter(14)-1);
    if (nFailHadLevelNow > nFailHadLevel) {
      nFailHadLevel    = nFailHadLevelNow;
      firstCallInEvent = true;
    }
    
    // Resetting for first time in new event
    if (firstCallInEvent) {
    //if (iSys == 0) {
      // Reset counters
      vinComPtr->resetCounters();
      
      // Set generated event weight from Pythia
      weightsPtr->resetWeights(infoPtr->weight()*ISRshowerPtr->PDFweight(),
                               infoPtr->nAccepted());
      
      // Reset all vectors for the first time we are called
      hardSys.clear();          hardSys.resize(0);
      MEclass.clear();          MEclass.resize(0);
      nQLneeded.clear();        nQLneeded.resize(0);
      Hint.clear();             Hint.resize(0);
      doPariNow.clear();        doPariNow.resize(0);
      particleNrs.clear();      particleNrs.resize(0);
      particles.clear();        particles.resize(0);
      branchelementals.clear(); branchelementals.resize(0);
      Q2hat.clear();            Q2hat.resize(0);
      iUnOrdered.clear();       iUnOrdered.resize(0);
      nBranchs.clear();         nBranchs.resize(0);
      firstTrial.clear();       firstTrial.resize(0);
      mSystem.clear();          mSystem.resize(0);
      isResonance.clear();      isResonance.resize(0);
      nG.clear();               nG.resize(0);
      nQQ.clear();              nQQ.resize(0);
    }
    
    // Debugging
    if (nAccepted == startDebug) {
      verbose = 9;
      vinCluPtr->setVerbose(9);
      vinComPtr->setVerbose(9);
      MEcorrectionsPtr->setVerbose(9);
    }
    
    // Sanity check: at least two particles in system
    int sizeSystem = partonSystemsPtr->sizeAll(iSys);
    if (sizeSystem <= 1) return;

    // Resize to make sure we can access the iSys'th element
    int currentSize = branchelementals.size();
    if ( currentSize < (iSys + 1) ) {
      hardSys.resize(iSys + 1);
      MEclass.resize(iSys + 1);
      nQLneeded.resize(iSys + 1);
      Hint.resize(iSys + 1);
      doPariNow.resize(iSys + 1);
      particleNrs.resize(iSys + 1);
      particles.resize(iSys + 1);
      branchelementals.resize(iSys + 1);
      Q2hat.resize(iSys + 1);
      iUnOrdered.resize(iSys + 1);
      nBranchs.resize(iSys + 1);
      firstTrial.resize(iSys + 1);
      mSystem.resize(iSys + 1);
      isResonance.resize(iSys + 1);
      nG.resize(iSys + 1);
      nQQ.resize(iSys + 1);
    }
    
    // We don't have a restart scale yet
    Q2hat[iSys] = 0.0;
    // After prepare we always have zero branchings
    nBranchs[iSys] = 0;
    // After prepare we are always in the first trial
    firstTrial[iSys] = true;
    
    // Note that we have not yet binned the first branching scale
    if (iSys == 0) firstQBranchBinned = false;
    
    // Vector to find total invariant mass of final state system
    Vec4 pSum;
    
    // Find and save all colors and anticolors
    // Not only FF, to find the neighbours
    map<int,int> indexOfAcol;
    map<int,int> indexOfCol;
    vector< pair<int,int> > antFF;
    
    // Loop over event record. Find indices of FS particles with color lines
    // Make new copies so any modifications by VINCIA 
    // (eg polarization assignment) are clearly recorded
    int iAncestor  = 0;
    for (int i = 0; i < sizeSystem; ++i) {    
      int i1 = partonSystemsPtr->getAll( iSys, i);
      // Skip if not present in final state
      if ( i1 <= 0 ) continue;
      // Save to colour maps (taking neg-index sextets into account)
      int col  = event[i1].col();
      int acol = event[i1].acol();
      if (!event[i1].isFinal()) { // Switch colours for initial partons
        col  = acol;
        acol = event[i1].col();
      }
      // Save colours
      if (col > 0) indexOfCol[col] = i1;
      else if (col < 0) indexOfAcol[-col] = i1;
      if (acol > 0) indexOfAcol[acol] = i1;
      else if (acol < 0) indexOfCol[-acol] = i1;
      // Check for ancestor only for final state particles
      if (event[i1].isFinal()) {
        // Check if all the partons have a single common mother
        if (iAncestor >= 0) {
          int iMot1 = event[i1].mother1();
          int iMot2 = event[i1].mother2();
          if (iMot2 == 0) iMot2 = iMot1;
          if (iAncestor == 0) iAncestor = iMot1;
          // If more than one ancestor found, set iAncestor = -1
          if (iMot1 != iAncestor || iMot2 != iMot1) iAncestor = -1; 
        }
        // Sum total FS momentum
        pSum += event[i1].p();
        // Update list of final-final connections
        if (col > 0 && indexOfAcol.count(col) == 1) {
          if (event[indexOfAcol[col]].isFinal())
            antFF.push_back( make_pair(i1,indexOfAcol[col]) );
        }
        if (acol < 0 && indexOfAcol.count(-acol) == 1) {
          if (event[indexOfAcol[-acol]].isFinal())
            antFF.push_back( make_pair(i1,indexOfAcol[-acol]) );
        }
        if (acol > 0 && indexOfCol.count(acol) == 1) {
          if (event[indexOfCol[acol]].isFinal())
            antFF.push_back( make_pair(indexOfCol[acol], i1) );
        }
        if (col < 0 && indexOfCol.count(-col) == 1) {
          if (event[indexOfCol[-col]].isFinal())
            antFF.push_back( make_pair(indexOfCol[-col], i1) );
        }
      }
      // Add to list of partons
      particleNrs[iSys].push_back( i1 );
      particles[iSys].push_back( event[i1] );
    }

    // Sanity check: If common ancestor found, check that it is a resonance
    if (iAncestor > 0) {
      if (! event[iAncestor].isResonance() ) iAncestor=-1;
      else { // Add resonance at the beginning
        particleNrs[iSys].insert( particleNrs[iSys].begin(), iAncestor );
        particles[iSys].insert( particles[iSys].begin(), event[iAncestor] );
      }
    }
    isResonance[iSys] = (iAncestor > 0);
    
    // Ariadne Factor for e+e- -> Z -> jets
    if (doPari && (iAncestor > 0))
      doPariNow[iSys] = (event[iAncestor].id() == 23);
    else doPariNow[iSys] = false;
    
    // Check the process class for MECs and whether we are in the
    // hardest system
    bool hardSysNow  = false;
    int nColIn       = 0;
    int nColOut      = 0;
    int nGluTot      = 0;
    int nBosIn       = 0;
    int nBosOut      = 0;
    int nZW          = 0;
    int nHiggs       = 0;
    int nPartsTot = (int)particles[iSys].size();
    for (int ip=0; ip<nPartsTot; ip++) {
      // Is this a coloured object?
      int col  = particles[iSys][ip].col();
      int acol = particles[iSys][ip].acol();
      if (col != 0 || acol != 0) {
        ( particles[iSys][ip].isFinal() ? nColOut++
                                        : nColIn++ );
      }
      // Is this a boson?
      else if ( (particles[iSys][ip].idAbs() > 21) && 
                (particles[iSys][ip].idAbs() < 38) ) {
        ( particles[iSys][ip].isFinal() ? nBosOut++
                                        : nBosIn++ );
        if ( (particles[iSys][ip].idAbs() == 23) || 
             (particles[iSys][ip].idAbs() == 24) ) nZW++;
        else if (particles[iSys][ip].idAbs() == 25) nHiggs++;
      }
      // Is this a gluon?
      if (particles[iSys][ip].id() == 21) nGluTot++;
      // If status code is 21,22,23 we are in the hard system
      if ( abs(particles[iSys][ip].statusAbs()-22) < 2 ) hardSysNow = true;
    }
    // Ask ISR if the system is hard, if so adjust hardSysNow
    // NOTE this needs to be done if ISR made q's massless
    if (ISRshowerPtr->isSysHard(iSys)) hardSysNow = true;
    int nColTot = nColIn + nColOut;
    int nBosTot = nBosIn + nBosOut;
    // Check what process class for MEC we should use
    int MEclassNow = 0;
    if (nColTot == nPartsTot) MEclassNow = pureQCD;
    else if ( (nBosTot == 1) && (nColTot == 2) && 
              (nBosTot+nColTot == nPartsTot) ) {
           if ((nColOut==0) && (nBosOut==1)) MEclassNow = singleBosonProd;
      else if ((nColIn ==0) && (nBosIn ==1)) MEclassNow = singleBosonDecay;
    }
    // Set properties
    hardSys[iSys]   = hardSysNow;
    MEclass[iSys]   = MEclassNow;
    nQLneeded[iSys] = nZW;
    // For Higgs check if we should do interference
    if (nHiggs > 0 && !Hinterference) {
      if (MEclassNow == singleBosonProd || MEclassNow == singleBosonDecay)
        Hint[iSys] = (nGluTot == 0 ? 1 : 2);
    } else Hint[iSys] = 0;
    
    // Set properties of hard system
    if (hardSysNow) {
      Q2fac       = infoPtr->Q2Fac();
      Qfac        = sqrt(Q2fac);
      
      // Are we a QCD process? Colour objects in the final state, unless they
      // they arise from a decay, i.e. gg -> H -> gg is NOT a QCD process
      // -> check if coloured particles in initial AND final state
      isQCD = (nColOut > 0 && nColIn > 0);
      
      // For QCD processes with hard jets allowed we need
      // factorization scale as Q2hat
      if (isQCD && hardJetsQCD)
        Q2hat[iSys] = Q2fac;
    }
    
    // Check if 1->n or 2->n
    int nIn = 0;
    if (isResonance[iSys]) nIn = 1;
    else if (partonSystemsPtr->hasInAB(iSys)) nIn = 2;

    // Sanity check: can only shower systems with more than 0 FF connection
    if (antFF.size() <= 0) return;
    
    // Store total invariant mass of the final state
    mSystem[iSys]  = m(pSum);
    
    // VINCIA starts by creating a copy of the particles it will work with
    // Colour and polarization information may be changed or added,
    // respectively, and masses may be set to zero for partons VINCIA
    // wants to treat as massless.
    treatAsMassless = true;
    int nISmassless = 0;
    int nISmassive  = 0;
    int nFS         = 0;
    vector<int> jMakeMassless;
    for (int i=0; i<(int)particleNrs[iSys].size(); i++) {
      int nr = particleNrs[iSys][i];
      if ( event[nr].m2Calc() > TINYMASS ) {
        // For resonance decays allow massive final state
        if (isResonance[iSys]) {
          if (event[nr].idAbs() <= nFlavMassless)
            jMakeMassless.push_back(i);
          // Ignore resonance itself
          else if (i>nIn) treatAsMassless = false;
        }
        // For everything else map everyone with colors to massless
        else if (abs(event[nr].col())+abs(event[nr].acol()) > 0)
          jMakeMassless.push_back(i);
        if (!event[nr].isFinal()) nISmassive++;
      }
      else if (!event[nr].isFinal()) nISmassless++;
      if (event[nr].isFinal()) nFS++;
    }
    // Special case, 2->n, with only massless incoming guys
    bool doSpecial    = ( (nISmassive == 0) && (nISmassless = 2) );
    int nToMassless   = (int)jMakeMassless.size();
    bool madeMassless = (jMakeMassless.size() > 0);
    
    // Sanitize input. 
    // Bottom and lighter quarks are always bookkept as massless by Vincia
    // (Top quarks may be made massless as well, for ultra-high energies)
    while (jMakeMassless.size() > 0) {
      // Special case, 2->n, with only massless incoming guys and either
      // 2 partons to make massless or >1 FS guys with 1 to make massless
      if ( doSpecial && ( (nToMassless == 2) || 
                          (nToMassless == 1 && nFS > 1) ) ) {
        int i1 = jMakeMassless[0];
        int i2 = jMakeMassless[nToMassless-1];
        if (i2 == i1) { // Need massive partner
          for (int ip=0; ip<(int)particles[iSys].size(); ip++) {
            if (particles[iSys][ip].isFinal() && (ip != i1)) i2 = ip;
          }
        }
        // Go to rest frame of pair
        Vec4 p1 = particles[iSys][i1].p();
        Vec4 p2 = particles[iSys][i2].p();
        RotBstMatrix M;
        M.toCMframe(p1,p2);
        p1.rotbst(M);
        p2.rotbst(M);
        double eCMs = p1.e() + p2.e();
        // Check if both to be massless
        if (nToMassless == 2 || particles[iSys][i2].m2() <= TINY) {
          p1 = Vec4(0.0, 0.0,  eCMs/2.0, eCMs/2.0);
          p2 = Vec4(0.0, 0.0, -eCMs/2.0, eCMs/2.0);
          // PYTHIA stores mass as "5th" component, to preserve precision
          particles[iSys][i1].m(0.0);
          particles[iSys][i2].m(0.0);
        }
        // Else just make i1 massless
        else {
          double e1 = (eCMs - particles[iSys][i2].m2()/eCMs)/2.;
          double e2 = eCMs - e1;
          p1 = Vec4(0.0, 0.0,  e1, e1);
          p2 = Vec4(0.0, 0.0, -e1, e2);
          // PYTHIA stores mass as "5th" component, to preserve precision
          particles[iSys][i1].m(0.0);
        }
        M.invert();
        p1.rotbst(M);
        p2.rotbst(M);
        particles[iSys][i1].p(p1);
        particles[iSys][i2].p(p2);
        jMakeMassless.clear();
        jMakeMassless.resize(0);
        continue;
      }
      // Special case, 2->n, with only massless incoming guys
      // use mapToMassless
      if (doSpecial) {
        vector<Vec4> momenta;
        for (int i=0; i<(int)jMakeMassless.size(); i++)
          momenta.push_back(particles[iSys][jMakeMassless[i]].p());
        // Mapping
        vinCluPtr->mapToMassless(momenta, 0);
        // Set new momenta
        for (int i=0; i<(int)jMakeMassless.size(); i++) {
          particles[iSys][jMakeMassless[i]].p(momenta[i]);
          // PYTHIA stores mass as "5th" component, to preserve precision
          particles[iSys][jMakeMassless[i]].m(0.0);
        }
        jMakeMassless.clear();
        jMakeMassless.resize(0);
        continue;
      }
      // Start with last parton
      // (-> easy to use pop_back to erase it when we are done with it)
      int i1 = jMakeMassless[jMakeMassless.size()-1];
      Particle& parton1 = particles[iSys][i1];
      // If this particle was already made massless, erase and continue
      if (parton1.m2() <= TINY) {
        jMakeMassless.pop_back();
        continue;
      }
      // Just make this parton massless
      if (parton1.idAbs() > 6) {
        double facInv = parton1.pAbs()/parton1.e();
        if (facInv > 0.0) parton1.rescale3(1./facInv);
        // PYTHIA stores mass as "5th" component, to preserve precision
        particles[iSys][i1].m(0.0);
        particles[iSys][i1].p(parton1.p());
      } else { // Find matching partner
        bool isIn1  = (!parton1.isFinal());
        bool isAnti = (parton1.id() < 0);
        int col1 = (isAnti ? parton1.acol() : parton1.col());
        for (int i2=0; i2<(int)particles[iSys].size(); ++i2) {
          // Skip self
          if (i2 == i1) continue;
          Particle& parton2 = particles[iSys][i2];
          bool isIn2  = (!parton2.isFinal());
          // Look for opposite-colour 
          int col2 = ( ((isIn1 && isIn2) || (!isIn1 && !isIn2)) ? 
                       (isAnti ? parton2.col() : parton2.acol()) :
                       (isAnti ? parton2.acol() : parton2.col()) );
          if (col1 == col2) {
            // Go to rest frame of parton pair
            Vec4 p1 = parton1.p();
            Vec4 p2 = parton2.p();
            RotBstMatrix M;
            M.toCMframe(p1,p2);
            p1.rotbst(M);
            p2.rotbst(M);
            double eCMs = p1.e() + p2.e();
            // Check if both to be massless
            if (parton2.m2() <= TINY) {
              p1 = Vec4(0.0, 0.0,  eCMs/2.0, eCMs/2.0);
              p2 = Vec4(0.0, 0.0, -eCMs/2.0, eCMs/2.0);
              // PYTHIA stores mass as "5th" component, to preserve precision
              particles[iSys][i1].m(0.0);
              particles[iSys][i2].m(0.0);
            }
            // Else just make this parton massless
            else {
              double e1 = (eCMs - parton2.m2()/eCMs)/2.;
              double e2 = eCMs - e1;
              p1 = Vec4(0.0, 0.0,  e1, e1);
              p2 = Vec4(0.0, 0.0, -e1, e2);
              // PYTHIA stores mass as "5th" component, to preserve precision
              particles[iSys][i1].m(0.0);
            }
            M.invert();
            p1.rotbst(M);
            p2.rotbst(M);
            particles[iSys][i1].p(p1);
            particles[iSys][i2].p(p2);
            break;
          }
        }
      }
      // Done with this parton
      jMakeMassless.pop_back();
    }
    
    // Verbose output
    if (verbose >= 6) {
      if (treatAsMassless) 
        printOut("TimeAntShower::prepare","Will treat input particles as relativistic");
      else 
        printOut("TimeAntShower::prepare","Will treat input particles as massive");
    }
    
    // Polarization: check if VINCIA should (attempt to) assign helicities
    bool doPolarize = false;
    usePolarizationNow = (treatAsMassless && helicityShower);
    if (usePolarizationNow) {
      // Check whether event is already polarized
      bool isPolarized = true;
      for (int i=0; i<(int)particles[iSys].size(); ++i) 
        if (event[i].pol() == 9) isPolarized = false;
      // If not, see if we can polarize it
      if (!isPolarized) {
        usePolarizationNow = matchingPtr->polarize(nIn,particles[iSys]);
        if (usePolarizationNow) doPolarize = true;
      }
    }
    if (verbose >= 7) printOut("TimeAntShower::prepare",
                               "doPolarize = "+bool2str(doPolarize));
    
    // Assign each dipole a random color index between 1 and 9.
    // (For use in colour reconnections)
    map<int,int> colorMap;
    bool doColorize = (CRmode >= 1);
    if (doColorize) {
      // TODO once this is used we need to make sure that Pythia knows how
      // many colTags we used
      int nextTag = event.nextColTag();
      doColorize = reconnectionsPtr->getColourMap(colorMap,particles[iSys],nextTag);
    }
    if (verbose >= 7) printOut("TimeAntShower::prepare",
                               "doColorize = "+bool2str(doColorize));
    
    // Create new (polarized, massless, color-indexed) copies of
    // original parton system;
    // only if we changed anything to avoid with changing the
    // system after ISR already did
    if (madeMassless || doPolarize || doColorize) {
      map<int,int> iChanged;
      indexOfCol.clear();
      indexOfAcol.clear();
      for (int i=0; i<(int)particles[iSys].size(); i++) {
        int iOld = particleNrs[iSys][i];
        int iNew = iOld;
        if (event[iOld].isFinal()) iNew = event.copy(iOld,51);
        else {
          iNew = event.copy(iOld,-event[iOld].statusAbs());
          // Everyone who as iOld has daughter needs to get
          // iNew as daughter now
          for (int ip=0; ip<(int)event.size(); ip++) if (ip < iOld) {
            if (event[ip].daughter1() == iOld) {
              // iNew > daughter2, that's fine
              event[ip].daughters(iNew, event[ip].daughter2());
            }
            if (event[ip].daughter2() == iOld) {
              // iNew < daughter2, need swap
              event[ip].daughters(iNew, event[ip].daughter1());
            }
          }
        }
        iChanged[iOld] = iNew;
        // Update 4-momentum and mass (eg if made massless)
        event[iNew].p(particles[iSys][i].p());
        event[iNew].m(particles[iSys][i].m());
        // Set polarizations and colors if new ones defined
        if (usePolarizationNow) event[iNew].pol(particles[iSys][i].pol());
        if (doColorize) {
          event[iNew].col(colorMap[event[iNew].col()]);
          event[iNew].acol(colorMap[event[iNew].acol()]);
        }
        // Update properties in particleNrs[iSys] and particles[iSys]
        particleNrs[iSys][i] = iNew;
        particles[iSys][i]   = event[iNew];
        // Update the list of systems.
        partonSystemsPtr->replace(iSys, iOld, iNew); 
        // Update color maps
        int col  = event[iNew].col();
        int acol = event[iNew].acol();
        if (!event[iNew].isFinal()) { // Switch colours for initial partons
          col  = acol;
          acol = event[iNew].col();
        }
        if (col  != 0) indexOfCol[col]   = iNew;
        if (acol != 0) indexOfAcol[acol] = iNew;
      }
      // Let ISR know we did changes to the system
      ISRshowerPtr->updateAfterFSR(iSys, event, iChanged);
    }
    
    if (verbose >= 6) {
      printOut("TimeAntShower::prepare", "saved the following "
               "particles: nr id col acol p");
      for (int i=0; i<(int)particles[iSys].size(); i++) {
        cout << "  " << setw(3) << particleNrs[iSys][i];
        cout << " " << setw(2) << particles[iSys][i].id();
        cout << " " << setw(3) << particles[iSys][i].col();
        cout << " " << setw(3) << particles[iSys][i].acol();
        cout << " " << particles[iSys][i].p();
      }
    }
    
    // Now loop over colored particles to create branchelementals (=antennae)
    for (map<int,int>::iterator it=indexOfCol.begin(); 
         it != indexOfCol.end(); ++it) {
      int col  = it->first;
      int i1   = it->second;        // iOld1/iNew1 is colour
      int i2   = indexOfAcol[col];  // iOld2/iNew3 is anticolour
      if (col == 0 || i1 == 0 || i2 == 0) continue;
      // Don't include not FF antennae
      if (!event[i1].isFinal() || !event[i2].isFinal() ) continue;
      if (verbose >= 7) {
        printOut("TimeAntShower::prepare",
                 "Creating antenna between "+num2str(i1,4)+", "
                 +num2str(i2,4)+": col = "+num2str(col,5));
      }
      
      // Store trial QCD antenna and add trial generators depending on type
      BranchelementalFSR trial(event,i1,i2);
      resetTrialGenerators(&trial);

      // Save system and ancestor
      trial.system    = iSys;
      trial.iAncestor = iAncestor;
    
      trial.renewTrial();
      trial.mCalc();

      // Set the unordered (re)start scale to the phase space limit.
      double Q2maxGemi   = trial.s() *
                           ( evolutionType == 1 ? pTnormalization/4.0
                                                : mDnormalization/2.0 );
      double Q2maxGsplit = trial.s();
      trial.setQ2reUnOrd(Q2maxGemi,Q2maxGsplit,-1);
      
      // Set the ordered (re)start scale to enhanceFactor * Q2fac.
      // For MPI, decays from hadronization, ... and fo (external) merging
      // we use the scale Pythia is providing us with (in pTnext).
      trial.setQ2reOrd(enhanceQ2fac*Q2fac,enhanceQ2fac*Q2fac,-1);
      
      // Set the (un)ordering:
      // 1) Hard process
      if (hardSysNow) {
        // a) QCD -> unordered if first hard jet is allowed,
        //             ordered if it is not allowed
        if (isQCD) trial.unOrdering = (hardJetsQCD);
        // b) non QCD (for FF ants this can only mean we are in 
        //             the decay of the hard process) 
        //            -> ordered
        else       trial.unOrdering = false;
      }
      // For antennae in the hard process use Q2re(Un)Ord
      // 2) MPI, decays from hadronization, ..., merging -> ordered
      //    use scale Pythia provides
      else trial.unOrdering = false;

      // Find neigbor on i1 side
      if (event[i1].acol() != 0) {
        int colL = event[i1].acol();
        trial.iL = indexOfCol[colL];
      }
      
      // Find neigbor on i2 side
      if (event[i2].col() != 0) {
        int colR = event[i2].col();
        trial.iR = indexOfAcol[colR];        
      }
      
      // Save branchelemental
      branchelementals[iSys].push_back(trial);
    }
    
    // Count up number of gluons and quark pairs
    nG[iSys]  = 0;
    nQQ[iSys] = 0;
    for (int i=0; i<(int)particles[iSys].size(); i++) {
      if (particles[iSys][i].id()==21) nG[iSys]++;
      else if (abs(particles[iSys][i].id()) < 7) nQQ[iSys]++;
    }
    // Half the quarks to get quark pairs
    nQQ[iSys] = nQQ[iSys]/2;

    // Sanity check
    if (branchelementals[iSys].size() <= 0) {
      if (verbose >= 5) printOut("TimeAntShower::prepare",
                                 "did not find any antennae: exiting.");
      return; 
    }
      
    //If verbose output, list dipoles
    if (verbose >= 5) list(); 
    
    // Verbose output
    if (verbose >= 6) 
      printOut("TimeAntShower::prepare","finished --------------");

  }
  
  //*********
  
  // Update list of branchelementals after ISR radiation
  
  void TimeAntShower::update( int iSys, Event& event, bool) {
    
    // Return here in case ISR wants to veto event anyway
    if (vetoHooksPtr->doVeto) {
      if (verbose >= 8) printOut("TimeAntShower::update","Returning since ISR vetoed event!");
      return;
    }
    
    // Verbose output
    if (verbose >= 6) 
      printOut("TimeAntShower::update","begin --------------");
    
    // If verbose, list event before showering
    if (verbose >= 5) event.list();
    
    // Count number of branches
    nBranchs[iSys]++;
    
    // Reset ordering mode to strong?
    if ( hardSys[iSys] && (orderingModeSav == 3) &&
         meCorrections && (nBranchs[iSys]==matchingLO) )
      orderingMode = 1;
    
    // Particles in the list are already updated by ISR
    // Find and save all colors and anticolors
    // Not only FF, to find the neighbours
    map<int,int> indexOfAcol;
    map<int,int> indexOfCol;
    vector< pair<int,int> > antFF;
    int sizeSystem = partonSystemsPtr->sizeAll(iSys);
    for (int i = 0; i < sizeSystem; ++i) {   
      int i1 = partonSystemsPtr->getAll( iSys, i);
      if ( i1 <= 0 ) continue;
      // Save to colour maps
      int col  = event[i1].col();
      int acol = event[i1].acol();
      if (!event[i1].isFinal()) { // Switch colours for initial partons
        col  = acol;
        acol = event[i1].col();
      }
      // Save colours
      if (col > 0) indexOfCol[col] = i1;
      else if (col < 0) indexOfAcol[-col] = i1;
      if (acol > 0) indexOfAcol[acol] = i1;
      else if (acol < 0) indexOfCol[-acol] = i1;
      if (event[i1].isFinal()) {
        // Update list of final-final connections
        if (col > 0 && indexOfAcol.count(col) == 1) {
          if (event[indexOfAcol[col]].isFinal())
            antFF.push_back( make_pair(i1,indexOfAcol[col]) );
        }
        if (acol < 0 && indexOfAcol.count(-acol) == 1) {
          if (event[indexOfAcol[-acol]].isFinal())
            antFF.push_back( make_pair(i1,indexOfAcol[-acol]) );
        }
        if (acol > 0 && indexOfCol.count(acol) == 1) {
          if (event[indexOfCol[acol]].isFinal())
            antFF.push_back( make_pair(indexOfCol[acol], i1) );
        }
        if (col < 0 && indexOfCol.count(-col) == 1) {
          if (event[indexOfCol[-col]].isFinal())
            antFF.push_back( make_pair(indexOfCol[-col], i1) );
        }
      }
    }

    // Sanity check: can only shower systems with more than 1 FF connection
    if (antFF.size() <= 0) return;
    
    // Count up number of gluons and quark pairs
    nG[iSys]  = 0;
    nQQ[iSys] = 0;
    for (int i=0; i<(int)particles[iSys].size(); i++) {
      if (particles[iSys][i].id()==21) nG[iSys]++;
      else if (abs(particles[iSys][i].id()) < 7) nQQ[iSys]++;
    }
    // Half the quarks to get quark pairs
    nQQ[iSys] = nQQ[iSys]/2;
    
    if (verbose >= 6) {
      printOut("TimeAntShower::update", "saved the following "
               "particles: nr id col acol p");
      for (int i=0; i<(int)particles[iSys].size(); i++) {
        cout << "  " << setw(3) << particleNrs[iSys][i];
        cout << " " << setw(2) << particles[iSys][i].id();
        cout << " " << setw(3) << particles[iSys][i].col();
        cout << " " << setw(3) << particles[iSys][i].acol();
        cout << " " << particles[iSys][i].p();
      }
    }
    
    // Loop over the antennae, we need to reset neighbours and check
    // antenna partons
    for (int i=0; i<(int)branchelementals[iSys].size(); i++) {
      BranchelementalFSR* antPtr = &branchelementals[iSys][i];
      bool changed   = false;
      // Copy old ids
      int id1        = antPtr->old1.id();
      int id2        = antPtr->old2.id();
      // Antenna partons
      if (!event[antPtr->iOld1].isFinal()) {
        // Renew trials if partons participated in a branching in IF
        // Status 44 = outgoing shifted by a branching
        if (event[antPtr->iOld1].status() != 44) antPtr->renewTrial();
        // iOld1 is always the one with the colour
        antPtr->iOld1 = indexOfCol[antPtr->old1.col()];
        changed = true;
      }
      if (!event[antPtr->iOld2].isFinal()) {
        if (event[antPtr->iOld2].status() != 44) antPtr->renewTrial();
        // iOld1 is always the one with the anticolour
        antPtr->iOld2 = indexOfAcol[antPtr->old2.acol()];
        changed = true;
      }
      antPtr->old1 = event[antPtr->iOld1];
      antPtr->old2 = event[antPtr->iOld2];
      // Neighbours
      if (antPtr->old1.id() == 21)
        antPtr->iL = indexOfCol[antPtr->old1.acol()];
      else antPtr->iL = 0;
      if (antPtr->old2.id() == 21)
        antPtr->iR = indexOfAcol[antPtr->old2.col()];
      else antPtr->iR = 0;
      // Recalculate mass and check we need to reset trial generators
      if (changed) {
        antPtr->mCalc();
        if (id1 != antPtr->old1.id() || id2 != antPtr->old2.id())
          resetTrialGenerators(antPtr);
      }
      // Remove the antennae out of the list
      // This way we can check later if ISR added a new FF antenna
      // iOld1 / iOld2 is colour / anticolour
      pair<int,int> pairNow = make_pair(antPtr->iOld1,antPtr->iOld2);
      vector< pair<int,int> >::iterator iter;
      iter = find (antFF.begin(), antFF.end(), pairNow);
      if (iter != antFF.end()) antFF.erase(iter);
    }
    
    // Is there a FF connection left?
    for (int i=0; i<(int)antFF.size(); i++) {
      int i1   = antFF[i].first;   // iOld1/iNew1 is colour
      int i2   = antFF[i].second;  // iOld2/iNew3 is anticolour
      // Don't include not FF antennae
      if (!event[i1].isFinal() || !event[i2].isFinal() ) continue;
      if (verbose >= 7) {
        printOut("TimeAntShower::update",
                 "Creating antenna between "+num2str(i1,4)+", "
                 +num2str(i2,4)+": col = "+num2str(event[i1].col(),5));
      }
      // Store trial QCD antenna and add trial generators depending on type
      BranchelementalFSR trial(event,i1,i2);
      resetTrialGenerators(&trial);
      // Find neigbor on i1 side
      if (event[i1].acol() != 0) {
        int colL = event[i1].acol();
        trial.iL = indexOfCol[colL];
      }
      // Find neigbor on i2 side
      if (event[i2].col() != 0) {
        int colR = event[i2].col();
        trial.iR = indexOfAcol[colR];        
      }
      // Save system
      trial.system = iSys;
      // This an antenna created by ISR, hence not a resonance decay
      trial.iAncestor = 0;
      branchelementals[iSys].push_back(trial); 
    }
    
    // Update is called after ISR branching, ISR sets
    // Markovian restart scale and who is unordered for us
    double Q2restart    = Q2hat[iSys];
    vector<int> iUnOrd  = iUnOrdered[iSys];
    int orderingModeNow = (hardSys[iSys] ? orderingMode : orderingModeMPI);
    
    for (int i=0; i<(int)branchelementals[iSys].size(); i++) {
      BranchelementalFSR* trial = &branchelementals[iSys][i];
      trial->mCalc();
      // Starting scale for unordered antennae = phase space limit
      double Q2maxGemi   = trial->s()*( evolutionType == 1 ? pTnormalization/4.0
                                                           : mDnormalization/2.0 );
      double Q2maxGsplit = trial->s();
      trial->setQ2reUnOrd(Q2maxGemi,Q2maxGsplit,-1);
      // Starting scale for ordered antennae = Q2restart
      trial->setQ2reOrd(Q2restart,Q2restart,-1);
      // Now check if antenna is ordered or unordered
      if (orderingModeNow == 0) { // no ordering
        trial->unOrdering = true;
        trial->renewTrial();
      } else if (orderingModeNow == 2) { // smooth ordering
        trial->unOrdering = false;
        for (int j=0; j<(int)iUnOrd.size(); j++) {
          if ( (trial->iOld1 == particleNrs[iSys][iUnOrd[j]]) ||
               (trial->iOld2 == particleNrs[iSys][iUnOrd[j]]) ) {
            trial->unOrdering = true;
            trial->renewTrial();
          }
        }
      } else if (orderingModeNow == 1) { // strong ordering
        trial->unOrdering = false;
      }
      // If we went from matched to non-matched order renew all trials
      if (meCorrections && (nBranchs[iSys]==matchingLO) && hardSys[iSys])
        trial->renewTrial();
    }
    
    // If verbose output, print
    if (verbose >= 5 && orderingModeNow == 2) {
      string toPrint = "Qhat for the Pimp factor is " 
                       + num2str(sqrt(Q2restart)) + " and ";
      for (int i=0; i<(int)iUnOrd.size(); i++) 
        toPrint = toPrint + num2str(particleNrs[iSys][iUnOrd[i]]) + " ";
      toPrint += " are unordered";
      printOut("TimeAntShower::update", toPrint);
      printOut("TimeAntShower::update", "therefore the following "
               "antennae are unordered:");
      for (int i=0; i<(int)branchelementals[iSys].size(); i++) {
        BranchelementalFSR ant = branchelementals[iSys][i];
        if (ant.system != iSys) continue;
        if (ant.unOrdering) cout << "  " << ant.iOld1 << " " << ant.iOld2 << endl;
      }
    }

    // Sanity check
    if (branchelementals[iSys].size() <= 0) {
      if (verbose >= 5) printOut("TimeAntShower::update",
                                 "did not find any antennae: exiting.");
      return; 
    }
      
    //If verbose output, list dipoles
    if (verbose >= 5) list();
    
    // Verbose output
    if (verbose >= 6) 
      printOut("TimeAntShower::update","finished --------------");

  }

  //*********
  
  // Select next pT in downwards evolution of the existing dipoles.
  
  double TimeAntShower::pTnext(Event& event, double pTevolBegAll,
                               double pTevolEndAll,
                               bool /*isFirstTrial*/, bool /*doTrialIn*/) {
    
    // Sanity checks
    if (branchelementals.size() <= 0) return 0.0;
    if (!settingsPtr->flag("Vincia:FSR")) return 0.0;
    
    // Verbose output
    if (verbose >= 6) 
      printOut("TimeAntShower::pTnext","begin --------------");
    
    // Verbose output
    if (verbose >= 4)
      printOut("TimeAntShower::pTnext",
               "(re)starting evolution between pTevolBegAll = " + num2str(pTevolBegAll) +
               " and pTevolEndAll = " + num2str(pTevolEndAll));
    
    // Check if we skipped all antenna since MPI/ISR already found high scale
    bool allSkipped        = true;
    // Check if we have cases where the qEnd < all cutoffs
    bool qEndsmallerCutoff = true;
    // Check if we have > 0 antennae
    bool zeroAnts          = true;
    
    // Denote VINCIA scales by "q", PYTHIA ones by "pTevol". 
    double qOld         = pTevolBegAll;    
    double qEndAll      = pTevolEndAll;
    double evNormFactor = 1.0;
         if ( evolutionType == 1 ) evNormFactor = pTnormalization;
    else if ( evolutionType == 2 ) evNormFactor = mDnormalization;

    // Initialize winner scale
    double qWin = 0.0;
    
    // Verbose output
    if (verbose >= 6) list();
    
    // Loop over the systems we have ----------- BEGIN
    int nrSystems = (int)branchelementals.size();
    for (int iSyst = 0; iSyst < nrSystems; iSyst++) {
    
      int nrAnts = (int)branchelementals[iSyst].size();
      if (nrAnts <= 0) continue; // Sanity check
      zeroAnts = false;
      
      // Loop over antennae ----------- BEGIN
      for (int iAnt = 0; iAnt < nrAnts; ++iAnt) {
      
        // Shorthand for current dipole-antenna
        BranchelementalFSR* trialPtr = &branchelementals[iSyst][iAnt];
      
        // Verbose output
        if (verbose >= 6) {
          printOut("TimeAntShower::pTnext",
                   "* processing antenna no"+num2str(iAnt,4)
                   +" col = "+num2str(trialPtr->old1.col())
                   +" nTrialGens = "+num2str(trialPtr->nTrialGenerators(),2));
        }
      
        // Generate new trial branchings, starting from qOld
        double qBegin = qOld;

        // All evolution normalizations are limited by invariant mass of dipole
        double sAnt    = trialPtr->s();
        double mAnt    = trialPtr->m();
        qBegin = min(qBegin,mAnt);
      
        // Lowest evolution boundary: impose hadronization scale.
        double qEnd = qEndAll;
      
        // Translate the cutoff scale to a lower bound on the evolution variable
        double qCutoff = cutoffScale;
        if (cutoffType == 0 || cutoffType == evolutionType) {
          qCutoff = sqrt(evNormFactor) * cutoffScale;
        } 
        // mD evolution with a cutoff in pT: larger mD hull
        else if (cutoffType == 1 && evolutionType == 2) {
          if (4*pow2(cutoffScale) > sAnt) qCutoff = qBegin;
          else
            qCutoff = sqrt( evNormFactor * sAnt 
                            * (1.0 - sqrt(1.0 - 4.0*pow2(cutoffScale)/sAnt)));
        } 
        // pT evolution with a cutoff in mD: larger pT hull
        else if (cutoffType == 2 && evolutionType == 1) {
          qCutoff = sqrt( evNormFactor * pow(cutoffScale,4.) / sAnt);
        }
      
        double qCutoffSplit = 0.0;
        if (cutoffType == 1  || (cutoffType == 0 && evolutionType == 1)) {
          if (4*pow2(cutoffScale) > sAnt) qCutoffSplit = qBegin;
          else
            qCutoffSplit = mAnt*sqrt(0.5*(1.0 - 
                                     sqrt(1.0 - 4.0*pow2(cutoffScale)/sAnt)));
        }
        else if (cutoffType == 2 || (cutoffType == 0 && evolutionType == 2)) {
          qCutoffSplit = qCutoff;
        }
      
        qEnd = max(qEndAll, min(qCutoff, qCutoffSplit));
      
        // No trial generators for this antenna
        if (trialPtr->nTrialGenerators() == 0) continue;

        // Sector showers with pT-evolution: check smallest unconnected 
        // gluon-emission pT to use as absolute upper bound on Emit evolution.
        double qT2sectorMin = pow2(qBegin);
        if (sectorShower && evolutionType == 1) {
          int iOld1 = trialPtr->iOld1;
          int iOld2 = trialPtr->iOld2;
          for (int iSct = 0; iSct < nrAnts; ++iSct) {
            // Skip self
            if (iSct == iAnt) continue;
              // Only compute Emit scales: skip if ID2 is not g
            if (branchelementals[iSyst][iSct].old2.id() != 21) continue;          
            // Skip connected 
            if (branchelementals[iSyst][iSct].iOld2 == iOld1 
                || branchelementals[iSyst][iSct].iR == iOld1
                || branchelementals[iSyst][iSct].iOld1 == iOld2) continue;
            // Check if this sector has smallest sector measure
            int ia = branchelementals[iSyst][iSct].iOld1;
            int ir = branchelementals[iSyst][iSct].iOld2;
            int ib = branchelementals[iSyst][iSct].iR;
            qT2sectorMin = 
              min(qT2sectorMin, resolutionPtr->Q2E(event[ia],event[ir],event[ib],1));
          }
        }

        // Loop over trial functions. Find and save max trial scale.
        double qTrialMax    = 0.0;
        int    indxMax      = -1;
      
        // Loop over trial generators ----------- BEGIN
        for (int indx = 0; indx < (int)trialPtr->nTrialGenerators(); ++indx) {

          // Pointer to trial generator for this trial:
          TrialGeneratorFSR* trialGenPtr = trialPtr->trialGenPtrSav[indx];
          string trialGenName            = trialGenPtr->name();
          int iAntPhys                   = trialPtr->getPhysIndex(indx);
        
          double qBeginNow = qBegin;
          double qEndNow   = qEnd;
        
          // Determine what the hadronization cutoff is for this trial
          double qCutoffNow = qCutoff;
          if (iAntPhys == iQGsplit || iAntPhys == iGGsplit) qCutoffNow = qCutoffSplit;
          
          // Only use Pythia's scale for the first emission in not hard systems
          // and if (external) merging is requested
          if ( ! ( !hardSys[iSyst] && firstTrial[iSyst] ) && 
               ! (  doMergeSav     && firstTrial[iSyst] ) ) {
            if (trialPtr->unOrdering) qBeginNow = sqrt(trialPtr->Q2reUnOrd[indx]);
            else                      qBeginNow = sqrt(trialPtr->Q2reOrd[indx]);
          }
        
          // Phase space limit can never be exceeded
          qBeginNow = min(qBeginNow,mAnt);
          
          if (qEndAll > qCutoffNow) qEndsmallerCutoff = false;
          
          // Check if any phase space still open
          if ( (qBeginNow <= qCutoffNow) || (qBeginNow <= qEndNow) ) {
            // Verbose output
            if (verbose >= 9)
              printOut("TimeAntShower::pTnext",
                       "skipping this trial since qBeginNow = " + num2str(qBeginNow) +
                       " qCutoffNow = " + num2str(qCutoffNow) +
                       " qEndNow = " + num2str(qEndNow));
            continue;
          }
          allSkipped = false;
          
          double qTrial = qBeginNow;
        
          bool needNewTrial = true;
          if (trialPtr->hasTrial(indx)) {
            if (trialPtr->getTrialScale(indx) < (qBeginNow+TINY)) {
              qTrial = trialPtr->getTrialScale(indx);   
              // Verbose output
              if (verbose >= 9)
                printOut("TimeAntShower::pTnext",
                         "Trial " + num2str(indx) + " " + trialGenName +
                         " Retrieving saved qTrial = " + num2str(qTrial));
              needNewTrial = false;
            }
          }
      
          // Generate new q value / no saved trial ----------- BEGIN
          if (needNewTrial) {
          
            // MR 12.7.2010: normalize by factor
            //    Ecm / sqrt( lambda( Ecm^2, m_parent1^2, m_parent2^2 ) ) 
            // with Lambda the Kallen function, to account for the 
            // normalisation to the two-particle phase space volume
            double psMassCorr = 1.0;
            double s1 = trialPtr->old1.m2();
            double s2 = trialPtr->old2.m2();
            // Are we doing mass corrections for flavours bookkepts as massless?
            if (massCorrectionEmit >= 3 && trialPtr->old1.idAbs() > nFlavZeroMass && trialPtr->old1.idAbs() <= nFlavMassless)
              s1 = pow2(trialPtr->old1.m0());
            if (massCorrectionEmit >= 3 && trialPtr->old2.idAbs() > nFlavZeroMass && trialPtr->old2.idAbs() <= nFlavMassless)
              s2 = pow2(trialPtr->old2.m0());
            // Compute phase-space correction
            if (s1 > TINY || s2 > TINY) {
              psMassCorr = sAnt/sqrt(pow2(sAnt - s1 - s2) - 4*s1*s2);
              if (verbose >= 8 && psMassCorr > 1.0)
                printOut("TimeAntShower::pTnext", "massive phase space volume factor is "+ num2str( psMassCorr ) );
            }
          
            // Sector showers: improve efficiency by reducing max trial scale 
            // for gluon emissions if we already know sector vetos will kill 
            // region above qT2sectorMin
            if (sectorShower && (iAntPhys == iQQemit || iAntPhys == iQGemit || iAntPhys == iGGemit))
              qTrial = min(qTrial,sqrt(qT2sectorMin));

            // Impose evolution windows
            bool acceptRegion = false;
            int iRegion = getRegion(qTrial);

            // Go through regions ----------- BEGIN
            while (!acceptRegion) {

              // Set overestimated Z range for trial generation
              double qMinNow = getQmin(iRegion);
              double yMinNow = pow2(qMinNow)/sAnt;
              double zMinNow = trialGenPtr->getZmin(yMinNow);
              double zMaxNow = trialGenPtr->getZmax(yMinNow);
              
              bool gluonSplit = (iAntPhys == iQGsplit || iAntPhys == iGGsplit);
              
              // Set color factor for trial
              double colFac = getAnt(iAntPhys)->getChargeFactor();
              int nF        = getNf(iRegion);
              if (gluonSplit) colFac *= min(nF,nGluonToQuarkF);
              
              // Set headRoom factor
              double headRoomFac = getHeadRoom( ( gluonSplit ? 1 : 0 ), iRegion,
                                                nG[iSyst], nQQ[iSyst], iSyst, qMinNow );
              // Add extra headRoom for Pari for gluon splitting
              if (gluonSplit && doPariNow[iSyst]) headRoomFac *= 1.3;
            
              // Normalization factor (headRoom * phaseSpaceMassCorrection)
              double normFac = headRoomFac * psMassCorr;

              // Verbose output
              if (verbose >= 9)
                  printOut("TimeAntShower::pTnext",
                           "Trial " + num2str(indx) + " " + trialGenName +
                           ": Generating new qTrial starting from " + num2str(qTrial) +
                           " (in evolution window " + num2str(iRegion) + " with qMin = " +
                           num2str(qMinNow) + "; nF = " + num2str(nF));
              
              // Effective renormalization-scale prefactor. 
              double kR = sqrt( gluonSplit ? aSkMu2Split : aSkMu2 );

              // Check if we can use a running alphaS
              bool runAlpha = false;
              if (alphaSorder >= 1 && evolutionType == 1) {
                runAlpha = true;
                // If we are close to lambdaQCD, use constant instead
                if (qMinNow < 2.5*getLambdaCMW(nF)/kR) runAlpha = false;
              }

              // Sanity check for zero branching probability
              if (colFac < TINY || headRoomFac < TINY) {
                // Save old scale
                double qTmp = qTrial;
                qTrial = 0.0;              
                trialPtr->saveTrial(indx,qTmp,qTrial,0.,0.,0.,0.);
              }

              // 1) AlphaS running inside trial integral
              // Only used for pT-evolution with muR propto pT 
              // and for mQQ-evolution with muR propto mQQ
              else if (runAlpha) { 
                // Save old scale
                double qTmp = qTrial;
                // One-loop beta function (two-loop imposed by veto, below)
                double b0 = (33.0 - 2.0*nF) / (12.0 * M_PI);
                // Use CMW alphaS for overestimate
                double lambdaEff = getLambdaCMW(nF);
                // Generate new q value, with alphaS running inside trial integral
                double q2trial = trialGenPtr->genQ2(pow2(qTrial),
                                                    sAnt,zMinNow,zMaxNow,colFac,
                                                    b0,kR,lambdaEff,normFac);
                qTrial = sqrt(q2trial);
                // Save trial information
                double evNormNow = evNormFactor;
                if (iAntPhys == iQGsplit || iAntPhys == iGGsplit) evNormNow = 1.0;
                double muEff    = kR*qTrial/sqrt(evNormNow)/lambdaEff;
                double alphaEff = 1./b0/log(pow2(muEff));
                trialPtr->saveTrial(indx,qTmp,qTrial,zMinNow,zMaxNow,colFac,
                                    alphaEff,headRoomFac);
              }
            
              // 2) AlphaS outside trial integral
              else {
                // Save old scale
                double qTmp = qTrial;
              
                // 0) constant alphaS
                double facAlphaS = alphaSvalue;
                
                // 1) running with muR not proportional to evolution variable
                if (alphaSorder >= 1) {
                  facAlphaS = alphaSmax;
                  // Special case for mD evolution with muR proportional to pT
                  if (evolutionType == 2) {
                    // Lowest pT scale that can be reached in this mD window
                    double pT2min = pow(qMinNow,4.)/sAnt/pow2(evNormFactor);
                    double mu2 = max(mu2min,pow2(kR)*pT2min);
                    // Use CMW alphaS for overestimate
                    facAlphaS = alphaSCMWptr->alphaS(mu2);
                    facAlphaS = min(alphaSmax,facAlphaS);
                    // Special for 2-loop running below 1 GeV: extra headroom
                    if (alphaSorder >= 2 && mu2 < 1.) facAlphaS *= 1.1;
                  }
                }
              
                // Generate new q value, with alphaS outside trial integral
                double q2trial = trialGenPtr->genQ2(pow2(qTrial),
                                                    sAnt,zMinNow,zMaxNow,colFac,
                                                    facAlphaS,normFac); 
                qTrial = sqrt(q2trial);
                // Save trial information
                trialPtr->saveTrial(indx,qTmp,qTrial,zMinNow,zMaxNow,colFac,
                                    facAlphaS,headRoomFac);
              } 

              // Check evolution window boundaries 
              if (qTrial > qMinNow) acceptRegion = true;
              else if (qMinNow < qWin || qMinNow < qTrialMax) {
                if (verbose >= 8)
                  printOut("TimeAntShower::pTnext",
                           "stopping evolution, already found scale that is bigger");
                acceptRegion = true;
                trialPtr->renewTrial(indx); // Don't use this trial (outside the boundaries)
                // Set the restart scales
                // Use tiny to prevent ending up in the same evolution window next time
                double Q2restart = pow2(qMinNow)-TINY;
                trialPtr->setQ2reUnOrd(Q2restart,Q2restart,indx,true);
                trialPtr->setQ2reOrd(Q2restart,Q2restart,indx,true);
                qTrial = 0.0;
              } 
              else if (iRegion == 0) { 
                acceptRegion = true;
                qTrial       = 0.0;
              }
              else { 
                qTrial = qMinNow;
                iRegion--;
              }

            } 
            // Go through regions ----------- END
            
          }
          // Generate new q value / no saved trial ----------- END
        
          if (qTrial > qTrialMax) {
            qTrialMax = qTrial;
            indxMax   = indx;
          }
          
        }
        // Loop over trial generators ----------- END
      
        // Did I win?
        if (qTrialMax >= qWin || qWin <= 0.0) {      
          winnerPtr = trialPtr;
          qWin      = qTrialMax;
          indxWin   = indxMax;
          iSysWin   = iSyst;
        } 
      
        // Verbose output
        if (verbose >= 7)
          printOut("TimeAntShower::pTnext",
                   "qTrialMax = " + num2str(qTrialMax) + 
                   ((indxMax >= 0) ? " ("+trialPtr->trialGenPtrSav[indxMax]->name()+")" : "") +
                   ((iAnt+1 == nrAnts) ? "; final qWin = " : "; current qWin = ") +
                   num2str(qWin) + " in system " + num2str(iSyst,4));
            
      }
      // Loop over antennae ----------- END
      
      // We are certainly not in the first trial anymore      
      firstTrial[iSyst] = false;
      
    }
    // Loop over the systems we have ----------- END
    
    // Set the restart scales: Unordered antennae get the winning scale, ordered antennae
    // get the winning scale only if it is smaller than their current restart scale!
    // We only use that if we get the branching and it gets rejected 
    // (will be overwritten if we actually carry out a branching or if ISR gets the branching)
    if (!allSkipped) {
      for (int iSyst = 0; iSyst < nrSystems; iSyst++) {
        for (int iAnt = 0; iAnt < (int)branchelementals[iSyst].size(); ++iAnt) {
          branchelementals[iSyst][iAnt].setQ2reUnOrd(pow2(qWin),pow2(qWin),-1,true);
          branchelementals[iSyst][iAnt].setQ2reOrd(pow2(qWin),pow2(qWin),-1,true);
        }
      }
    }

    // If non-zero branching scale found: continue
    if (qWin > qEndAll) {
      // Verbose output  
      if (verbose >= 5) {
        if (verbose >= 7)  list();
        printOut("TimeAntShower::pTnext", "=== Winner at scale qWin = "
                 + num2str(qWin)
                 + " trial type " + winnerPtr->trialGenPtrSav[indxWin]->name()
                 + "   col = " + num2str(winnerPtr->old1.col(),5));
      }
    }
    
    else {
      qWin = 0.0;
      // Check if qEnd < all cutoffs
      if (qEndsmallerCutoff && !zeroAnts) {
        // Verbose output
        if (verbose >= 5) {
          printOut("TimeAntShower::pTnext","=== All trials now below cutoff "
                   "qEndAll = "+num2str(qEndAll,3)+".");
          if (verbose >= 6) {
            printOut("TimeAntShower::pTnext","Final configuration was:");
            event.list();
          }
        }
        
        for (int iSyst = 0; iSyst < nrSystems; iSyst++) {
          for (int iAnt = 0; iAnt < (int)branchelementals[iSyst].size(); ++iAnt) {
            branchelementals[iSyst][iAnt].setQ2reUnOrd(pow2(qWin),pow2(qWin),-1,true);
            branchelementals[iSyst][iAnt].setQ2reOrd(pow2(qWin),pow2(qWin),-1,true);
          }
        }
        
        // FSR is done (at least for this system), set the weights
        weightsPtr->doWeighting();
        
        // If we have not yet binned a branching scale, bin 0 as the first
        // branching scale
        if (verbose >= 2 && !firstQBranchBinned) {
          if (vinciaHistos.find("1stBranchingQE/eCM") != vinciaHistos.end())
            vinciaHistos["1stBranchingQE/eCM"].fill(0);
          firstQBranchBinned = true;
        }
      
        // Optionally, before passing the event back to Pythia, apply
        // non-perturbative colour reconnections
        if (CRmode >= 1) {
          // Find branchelementals of hard system
          int iSysHard = -1;
          for (int i=0; i<(int)hardSys.size(); i++)
            if (hardSys[i]) iSysHard = i;
          if (iSysHard > -1) {
            // First try junction-string reconnections, if switched on
            if (CRjunctions) {
              int nReconnectionsJun 
                = reconnectionsPtr->doJunctions(event, branchelementals[iSysHard]);
              if (verbose >= 4 && nReconnectionsJun >= 1) 
                printOut("pTnext()","Junction(s) created before HAD");
            }
            int measureType    = 1;
            int nReconnections = 
              reconnectionsPtr->doReconnect(event, branchelementals[iSysHard],
                                            0.0, measureType);
              if (verbose >= 3 && nReconnections >= 1) 
                printOut("pTnext()","Antenna reconnection(s) created before HAD");
          }
        }
      }
    }
    
    // End of pTnext: finalize and return to PYTHIA
    
    // Verbose output
    if (verbose >= 6) 
      printOut("TimeAntShower::pTnext","finished --------------");
    
    // Done! Unordered values should not spoil the evolution!   
    return qWin;
    
  }
  
  //*********
  
  // Perform a branching
  
  bool TimeAntShower::branch(Event& event, bool isInterleaved) {
    
    // Verbose output
    if (verbose >= 6) 
      printOut("TimeAntShower::branch","begin --------------");
    
    // Count up global number of attempted trials
    ++nTrialsSum;

    // The winning branchelemental and trial type are stored in winner
    // and indxWin respectively. Extract current QE and sAnt scales
    double qNew    = winnerPtr->scaleSav[indxWin];
    int iTrial     = winnerPtr->getTrialIndex();
    int iAntPhys   = winnerPtr->getPhysIndex(iTrial);
    bool isSwapped = winnerPtr->getIsSwapped(iTrial);
    iSysSel        = winnerPtr->system;
    
    // Verbose output
    if (verbose >= 5)
      printOut("TimeAntShower::branch","* Processing Branching at scale Q = " 
               +num2str(qNew)+"   isInterleaved = "+bool2str(isInterleaved));
    
    //------------------------------------------------------------------------
    // GENERATE FULL TRIAL KINEMATICS (AND REJECT IF OUTSIDE PHASE SPACE)
    if (not generateKinematics(winnerPtr)) {
      // Mark this trial as "used", will need to generate a new one.
      winnerPtr->renewTrial(iTrial);
      return false;
    }

    //------------------------------------------------------------------------
    // ASSIGN COLOUR FLOW
    // NOTE: Using lastColTag here, if branching accepted we tell Pythia
    bool usedColTag = assignColourFlow(event, winnerPtr);

    //------------------------------------------------------------------------
    // UPDATE LIST OF PARTONS
    vector<Particle> partsNow = particles[iSysSel];
    vector<int> antPos; 
    antPos.resize(3);
    // Replace the old ones...
    for (int i = 0; i < (int)particleNrs[iSysSel].size(); ++i) { 
      if (particleNrs[iSysSel][i] == winnerPtr->iOld1) {
        partsNow[i] = winnerPtr->new1;
        antPos[0]   = i;
      }
      else if (particleNrs[iSysSel][i] == winnerPtr->iOld2) {
        partsNow[i] = winnerPtr->new3;
        antPos[1]   = i;
      }
    }
    // ... and add the new one.
    partsNow.push_back(winnerPtr->new2);
    antPos[2] = (int)partsNow.size()-1;

    //------------------------------------------------------------------------
    // PHASE SPACE CLOSED FOR GETTING RID OF HEAVY QUARKS?
    if (!ISRshowerPtr->checkHeavyQuarkPhaseSpace(partsNow,iSysSel)) {
      if (verbose >= 6)
        printOut("TimeAntShower::branch", "branching rejected because phase "
                 "space after branching does not allow forced splittings");
      // Mark this trial as "used", will need to generate a new one.
      winnerPtr->renewTrial(iTrial);
      ++nClosePSforHQ;
      return false;
    }
    
    //------------------------------------------------------------------------
    // CALCULATE MARKOVIAN RESTART SCALE
    // Ordering mode for this system
    int orderingModeNow = orderingModeMPI;
    if (hardSys[iSysSel]) {
      if ( (orderingModeSav == 3) && meCorrections && 
           (nBranchs[iSysSel]+1==matchingLO) )
        orderingModeNow = 1;
      else orderingModeNow = orderingMode;
    }
    double Q2restart = 0.0;
    vector<int> iUnOrd;
    // For smooth ordering calculate the Markovian scale
    // otherwise just use the scale of the branching
    if (orderingModeNow == 2) {
      MEcorrectionsPtr->setClustProps(nQLneeded[iSysSel], Hint[iSysSel]);
      MEcorrectionsPtr->unorderingScale(partsNow, &Q2restart, &iUnOrd);
    } else {
      Q2restart = pow2(qNew);
      iUnOrd.resize(0);
    }

    //------------------------------------------------------------------------
    // VETO STEP: DECIDE WETHER TO ACCEPT OR REJECT BRANCHING
    if (not acceptTrial(event, winnerPtr, partsNow, antPos, Q2restart)) {
      // Mark this trial as "used", will need to generate a new one.
      winnerPtr->renewTrial(iTrial);
      return false;
    }
    
    // Flag to tell whether a general update of antenna parents is necessary
    bool updateMothers = false;
    
    //------------------------------------------------------------------------
    // PUT NEW PARTICLES INTO EVENT RECORD
    // Store a copy of event, to be used if branching vetoed by userHooks
    Event evtOld   = event;
    int evtSizeOld = evtOld.size();
    int iOld1      = winnerPtr->iOld1;
    int iOld2      = winnerPtr->iOld2;
    int iNew1      = event.append(winnerPtr->new1);
    int iNew2      = event.append(winnerPtr->new2);
    int iNew3      = event.append(winnerPtr->new3);
    // Update event pointers if necessary
    event.restorePtrs();
    // Check if we went from polarized to unpolarized state
    // If so, depolarize parton state
    // (A more complete alternative here would be to create depolarized 
    // copies of all partons and then update everything, but deemed
    // unnecessary for now.)
    if ( event[iOld1].pol() != 9 && event[iOld2].pol() != 9 && 
        (event[iNew1].pol() == 9 || event[iNew2].pol() == 9 ||
         event[iNew3].pol() == 9)) {
      if (verbose >= 5) 
        printOut("TimeAntShower::branch","Depolarizing parton state");
      // Depolarize parton state (except the branching mothers, 
      // which will be replaced by unpolarized daughters) 
      int sizeSystem = partonSystemsPtr->sizeOut(iSysSel);
      for (int i = 0; i < sizeSystem; ++i) {    
        int i1 = partonSystemsPtr->getOut( iSysSel, i);
        // Skip if not present in final state
        if ( i1 <= 0 || !event[i1].isFinal()) continue;
        // Skip if mother parton to be replaced by daughter
        if ( i1 == winnerPtr->iOld1 || i1 == winnerPtr->iOld2) continue;
        // Else depolarize
        if ( event[i1].pol() != 9 ) event[i1].pol(9);
        // Tell that antenna mothers should be updated in update loop below
        updateMothers = true;
      }
      // Also make sure all daughters are depolarized as well
      winnerPtr->new1.pol(9);
      winnerPtr->new2.pol(9);
      winnerPtr->new3.pol(9);
    }
    
    //------------------------------------------------------------------------
    // SET MOTHERS AND DAUGHTERS
    // Mark original dipole partons as branched
    // iOld1=I -> iNew1=i
    //         -> iNew2=j
    // iOld2=K -> iNew3=k
    event[iOld1].statusNeg();
    event[iOld2].statusNeg();
    // Gluon emission
    if (iAntPhys == iQQemit || iAntPhys == iQGemit || iAntPhys == iGGemit) {
      event[iNew1].mothers(iOld1,0);
      event[iNew2].mothers(iOld1,iOld2);
      event[iNew3].mothers(iOld2,0);
      event[iOld1].daughters(iNew2,iNew1); 
      event[iOld2].daughters(iNew3,iNew2);
    }
    // Gluon splitting in old2
    else if (isSwapped) {
      event[iNew1].mothers(iOld1,0);
      event[iNew2].mothers(iOld1,0);
      event[iNew3].mothers(iOld2,iOld2);
      event[iOld1].daughters(iNew2,iNew1); 
      event[iOld2].daughters(iNew3,iNew3);
    }
    // Gluon splitting in old1
    else {
      event[iNew1].mothers(iOld1,iOld1);
      event[iNew2].mothers(iOld2,0);
      event[iNew3].mothers(iOld2,0);
      event[iOld1].daughters(iNew1,iNew1); 
      event[iOld2].daughters(iNew3,iNew2);
    }
    event[iNew1].daughters(0,0); 
    event[iNew2].daughters(0,0);
    event[iNew3].daughters(0,0);
    
    //------------------------------------------------------------------------
    // VETO BY USERHOOKS
    // Possibility to allow user veto of emission step.
    if (canVetoEmission)
    if (userHooksPtr->doVetoFSREmission(evtSizeOld, event, iSysSel)) {
      // Restore old event
      event = evtOld;
      return false;
    }

    //------------------------------------------------------------------------
    // EVERYTING ACCEPTED
    
    // If diagnostic histograms are on, write in the branching scale
    if (verbose >= 2 && !firstQBranchBinned) {
      if (vinciaHistos.find("1stBranchingQE/eCM") != vinciaHistos.end())
        vinciaHistos["1stBranchingQE/eCM"].fill( qNew/event[0].e() );
      firstQBranchBinned = true;
    }
    
    // Update the list of systems.
    partonSystemsPtr->replace(iSysSel, iOld1, iNew1);  
    partonSystemsPtr->addOut(iSysSel, iNew2);
    partonSystemsPtr->replace(iSysSel, iOld2, iNew3); 
    
    // Update number of gluons and quark pairs
    if (winnerPtr->new2.id() == 21) ++nG[iSysSel];
    else {
      nG[iSysSel] -= 1;
      ++nQQ[iSysSel];
    }
    
    // Check merging functions
    /*int iEmis = iNew2; // Emission
    int iRadi = iNew1; // Radiator
    if (event[iNew2].id() != 21 && !isSwapped) iRadi = iNew3;
    int iReco = (iRadi == iNew1 ? iNew3 : iNew1);
    cout << " iRad, iEmi, iRec = " << iRadi << ", " << iEmis << ", " << iReco << endl;
    // 3) Get name of branching
    string branchName = getSplittingName( event, iRadi, iEmis, iReco );
    cout << " iAntPhys = " << iAntPhys << endl;
    cout << " Branching name = " << branchName << endl;
    // 4) Get branching probability
    cout << " Branching probability = " 
         << getSplittingProb(event, iRadi, iEmis, iReco, branchName) << endl;
    // 5) Get scale
    cout << " Branching scales = " << pow2(qNew) << " and "
         << getStateVariables(event, iRadi,  iEmis, iReco, branchName)[0] << endl;
    // 6) Get clustered event
    cout << " Event before clustering: " << endl;
    event.list();
    cout << " Event after clustering: " << endl;
    clustered(event, iRadi,  iEmis, iReco, branchName).list();
    cout << endl;*/
    
    //------------------------------------------------------------------------
    // UPDATE LIST OF PARTONS AND PARTON NUMBERS
    // Save the first particle (might be ancestor)
    int iAncestor     = particleNrs[iSysSel][0];
    Particle ancestor = particles[iSysSel][0];
    particleNrs[iSysSel].clear(); particleNrs[iSysSel].resize(0);
    particles[iSysSel].clear();   particles[iSysSel].resize(0);
    if (winnerPtr->iAncestor > 0) {
      particleNrs[iSysSel].push_back( iAncestor );
      particles[iSysSel].push_back( ancestor );
    }
    int sizeSystem = partonSystemsPtr->sizeAll(iSysSel);
    for (int i = 0; i < sizeSystem; ++i) {    
      int i1 = partonSystemsPtr->getAll( iSysSel, i);
      if ( i1 <= 0 ) continue;
      particleNrs[iSysSel].push_back( i1 );
      particles[iSysSel].push_back( event[i1] );
    }
    if (verbose >= 6) {
      printOut("TimeAntShower::branch", "updated system, now we have "
               "the following particles: nr id col acol p");
      for (int i=0; i<(int)particles[iSysSel].size(); i++) {
        cout << "  " << setw(3) << particleNrs[iSysSel][i];
        cout << " " << setw(2) << particles[iSysSel][i].id();
        cout << " " << setw(3) << particles[iSysSel][i].col();
        cout << " " << setw(3) << particles[iSysSel][i].acol();
        cout << " " << particles[iSysSel][i].p();
      }
    }
    
    if (orderingModeNow == 2) {
      MEcorrectionsPtr->setClustProps(nQLneeded[iSysSel], Hint[iSysSel]);
      MEcorrectionsPtr->unorderingScale(particles[iSysSel], &Q2restart, &iUnOrd);
    }
    // Update restart scale and who is unordered
    Q2hat[iSysSel]      = Q2restart;
    iUnOrdered[iSysSel] = iUnOrd;
    // Tell ISR about the restart scale and who is unordered
    ISRshowerPtr->setMarkovianProps(iSysSel, Q2restart, iUnOrd,
                                    particles[iSysSel], particleNrs[iSysSel]);
    
    // Update old and add new dipoles
    if (verbose >= 5)
      printOut("TimeAntShower::branch",
               "=== Branching Accepted. Updating dipole-antenna(e)");
    
    //************************************************************************
    // Updates of the branched antennae
    //************************************************************************
    
    // Store old neighbours
    int iL  = winnerPtr->iL;
    int iR  = winnerPtr->iR;
    
    if (winnerPtr->new1.colType() > 0) {        
      // Normal case, where iNew1-iNew2 has an LC color connection
      winnerPtr->reset(event,iNew1,iNew2);
      winnerPtr->old1 = event[iNew1];
      winnerPtr->old2 = event[iNew2];    

      // Update trial generators
      resetTrialGenerators(winnerPtr);

      // Update neighbours
      winnerPtr->iL = iL;
      winnerPtr->iR = iNew3;
      
      // If second parton is octet, insert new antenna (iNew2,iNew3)
      if (event[iNew2].colType() > 0) {
        // Define new dipole-antenna
        BranchelementalFSR trial(event,iNew2,iNew3);
        trial.system = iSysSel;
        trial.old1   = event[iNew2];
        trial.old2   = event[iNew3];
        
        // Set trial generators
        resetTrialGenerators(&trial);

        // Set tri- and quadrupole neighbors
        trial.iL     = iNew1;
        trial.iR     = iR;
        
        // Keep family name for matching
        trial.iAncestor = winnerPtr->iAncestor;
        
        // Add dipole-antenna to list, insert after iNew1-Inew2 to keep order
        for (int iAnt=0; iAnt<int(branchelementals[iSysSel].size()); iAnt++) {          
          if (&branchelementals[iSysSel][iAnt] == winnerPtr) {
            branchelementals[iSysSel].insert(branchelementals[iSysSel].begin()+iAnt+1,trial); 
            break;
          }
        }
      }
    } else {
      // Special case for qbar q-x final state, where only second pair is dipole
      winnerPtr->reset(event,iNew2,iNew3);
      winnerPtr->old1 = event[iNew2];
      winnerPtr->old2 = event[iNew3];
      
      // Update trial generators
      resetTrialGenerators(winnerPtr);

      // Update neighbors
      winnerPtr->iL = iNew1;
      winnerPtr->iR = iR;
    }
    
    //************************************************************************
    // Updates of the other antennae
    //************************************************************************
    
    //------------------------------------------------------------------------
    // Updates
    for (int iAnt = 0; iAnt < (int)branchelementals[iSysSel].size(); ++iAnt) {
      BranchelementalFSR* antPtr = &branchelementals[iSysSel][iAnt];
      
      // Update mother partons in case a depolarization took place
      if (updateMothers) {
        antPtr->old1 = event[antPtr->iOld1];
        antPtr->old2 = event[antPtr->iOld2];
      }
      
      // The post-branching branchelementals
      if (antPtr->iOld1 == iNew1 || antPtr->iOld1 == iNew2) {
        // No update necessary for this pair (handled above)
      }
      
      else {
        // Neighbour update 
        if (antPtr->iL == iOld2) antPtr->iL = iNew3;
        if (antPtr->iR == iOld1) antPtr->iR = iNew1;
        
        if (antPtr->iOld1 == iOld2) {
          antPtr->iOld1       = iNew3;
          antPtr->old1        = event[iNew3];
          // Update neighbors
          antPtr->iL          = iNew2;
          antPtr->renewTrial();
          // Update trial generators
          resetTrialGenerators(antPtr);
        }
        
        if (antPtr->iOld2 == iOld1) {
          antPtr->iOld2       = iNew1;
          antPtr->old2        = event[iNew1];
          // Update neighbors
          antPtr->iR          = iNew2;
          antPtr->renewTrial();
          // Update trial generators
          resetTrialGenerators(antPtr);
        }
      }
      
      antPtr->mCalc();
      
      // Starting scale for unordered antennae = phase space limit
      double Q2maxGemi   = antPtr->s()*( evolutionType == 1 ? pTnormalization/4.0
                                                            : mDnormalization/2.0 );
      double Q2maxGsplit = antPtr->s();
      antPtr->setQ2reUnOrd(Q2maxGemi,Q2maxGsplit,-1);
      // Starting scale for ordered antennae = Q2restart
      antPtr->setQ2reOrd(Q2restart,Q2restart,-1);
      
      // Now check if antenna is ordered or unordered
      if (orderingModeNow == 0) { // no ordering
        antPtr->unOrdering = true;
        antPtr->renewTrial();
      } else if (orderingModeNow == 2) { // smooth ordering
        antPtr->unOrdering = false;
        for (int j=0; j<(int)iUnOrd.size(); j++) {
          if ( (antPtr->iOld1 == particleNrs[iSysSel][iUnOrd[j]]) ||
               (antPtr->iOld2 == particleNrs[iSysSel][iUnOrd[j]]) ) {
            antPtr->unOrdering = true;
            antPtr->renewTrial();
          }
        }
      } else if (orderingModeNow == 1) { // strong ordering
        antPtr->unOrdering = false;
      }
      
      // If we went from matched to non-matched order renew all trials
      if (meCorrections && (nBranchs[iSysSel]+1==matchingLO) && hardSys[iSysSel])
        antPtr->renewTrial();
    }
    
    //------------------------------------------------------------------------
    // Optionally allow perturbative colour reconnections 
    // (for the time being, apply only beyond matched orders, since 
    // matched orders already LO correct at full colour)
    if (hardSys[iSysSel] && CRmode >= 2) {
      int nQCDnow = nBranchs[iSysSel];
      bool doMEC  = (meCorrections && nQCDnow < matchingLO); 
      if (!doMEC) {
        int measureType    = 2;
        int nReconnections = 
          reconnectionsPtr->doReconnect(event, branchelementals[iSysSel], 
                                        pow2(qNew), measureType);
        if (nReconnections >= 1 && verbose >= 6) 
          cout << " I did a perturbative CR, at order = " << nQCDnow << endl;
      }
    }
    
    // If verbose output, print
    if (verbose >= 5 && orderingModeNow == 2) {
      string toPrint = "Qhat for the Pimp factor is " 
                       + num2str(sqrt(Q2restart)) + " and ";
      for (int i=0; i<(int)iUnOrd.size(); i++) 
        toPrint = toPrint + num2str(particleNrs[iSysSel][iUnOrd[i]]) + " ";
      toPrint += " are unordered";
      printOut("TimeAntShower::branch", toPrint);
      printOut("TimeAntShower::branch", "therefore the following "
               "antennae are unordered:");
      for (int i=0; i<(int)branchelementals[iSysSel].size(); i++) {
        BranchelementalFSR ant = branchelementals[iSysSel][i];
        if (ant.system != iSysSel) continue;
        if (ant.unOrdering) cout << "  " << ant.iOld1 << " " << ant.iOld2 << endl;
      }
    }
    
    if (verbose >= 5) {
      event.list();
      list();
      if (verbose >= 7) {
        cout<<" Masses (mCalc): "<<scientific
            <<event[iNew1].m()<<" ("<<event[iNew1].mCalc()<<")  "
            <<event[iNew2].m()<<" ("<<event[iNew2].mCalc()<<")  "
            <<event[iNew3].m()<<" ("<<event[iNew3].mCalc()<<")  "<<endl;
      }
    }
    
    // Count the number of branchings in the system
    nBranchs[iSysSel]++;
    
    // Reset ordering mode to strong?
    if ( hardSys[iSysSel] && (orderingModeSav == 3) &&
         meCorrections && (nBranchs[iSysSel]==matchingLO) )
      orderingMode = 1;
    
    // Let Pythia know how many color tags we used
    if (usedColTag) {
      event.nextColTag();
      if (event[iNew2].id() == 21) {
        int lastTag = event.lastColTag();
        int colMax  = max(event[iNew2].col(),event[iNew2].acol());
        while (colMax > lastTag) lastTag = event.nextColTag();
      }
    }
    
    // Check the event after each branching
    vinComPtr->showerChecks(event, true);
    
    // Verbose output
    if (verbose >= 6) 
      printOut("TimeAntShower::branch","finished --------------");
    
    // Done. 
    return true;
    
  }
  
  //*********
  
  // Update list of branchelementals after MPI rescattering 
  
  void TimeAntShower::rescatterUpdate( int iSys, Event& event) {
    if (verbose >= 1)
      printErr("TimeAntShower::rescatterUpdate","system "+num2str(iSys)
               +" (function not implemented)");
    if (verbose >= 5) event.list();
  }
  
  
  //**********************************************************************
  
  // END OF INHERITED FUNCTIONS FOR INTERFACE TO PYTHIA8
  // BEGIN VINCIA INTERNAL FUNCTIONS
  
  //**********************************************************************
  
  //*********
  
  // Update antennae if ISR changed system in prepare
  
  void TimeAntShower::updateAfterISR(int iSys, Event& event, map<int,int> iChanged) {
    
    // Don't need to do anything
    if ((int)branchelementals.size() <= iSys ) return;
    if (branchelementals[iSys].size() <= 0) return;
    
    // Update the particles in my list
    for (int i=0; i<(int)particleNrs[iSys].size(); i++) {
      int iOld = particleNrs[iSys][i];
      particleNrs[iSys][i] = iChanged[iOld];
      particles[iSys][i]   = event[iChanged[iOld]];
    }
    
    // Loop over the antennae to reset antenna partons and neighbours
    for (int i=0; i<(int)branchelementals[iSys].size(); i++) {
      BranchelementalFSR* antPtr = &branchelementals[iSys][i];
      int iOld1 = antPtr->iOld1;
      int iOld2 = antPtr->iOld2;
      antPtr->iOld1 = iChanged[iOld1];
      antPtr->iOld2 = iChanged[iOld2];
      antPtr->old1  = event[iChanged[iOld1]];
      antPtr->old2  = event[iChanged[iOld2]];
      // Recalculate the mass
      antPtr->mCalc();
      // Trialgenerators are still the same
      // Neighbours
      if (antPtr->iL != 0) antPtr->iL = iChanged[antPtr->iL];
      if (antPtr->iR != 0) antPtr->iR = iChanged[antPtr->iR];
      // Update daughter partons
      iOld1 = antPtr->iOld1;
      iOld2 = antPtr->iOld2;
      antPtr->new1 = Particle(0,51,iOld1,iOld2,0,0,0,0,0.);
      antPtr->new2 = Particle(0,51,iOld1,iOld2,0,0,0,0,0.);
      antPtr->new3 = Particle(0,51,iOld1,iOld2,0,0,0,0,0.); 
      antPtr->new1.setEvtPtr(&event);
      antPtr->new2.setEvtPtr(&event);
      antPtr->new3.setEvtPtr(&event);
    }
    
  }
  
  //*********
  
  // Initialize internal and diagnostics histograms
  
  void TimeAntShower::initVinciaHistos() {
    
    // Verbose output
    if (verbose >= 7) {
      printOut("TimeAntShower::initVinciaHistos","begin --------------");
    }
    
    vinciaHistos.clear();
    
    // Don't book diagnostics histograms if verbose output not switched on.
    if (verbose <= 1) return;
    
    // Global
    int nBin;
    double binWid, wid, c, xMin, xMax;
    
    // MATCHING WEIGHTS
    nBin   = 61;
    binWid = 0.1;
    wid    = nBin * binWid;
    c      = 0.0;
    xMin   = c - wid/2.;
    xMax   = c + wid/2.;
    Hist HPME2LogUSR("Matching weight (USR)",nBin,xMin,xMax);
    
    // Histos for Paccept : gluon emission
    for (int nGtmp = 0; nGtmp <= 9; ++nGtmp) {
      for (int nQQtmp = 0; nQQtmp <= 5; ++nQQtmp) {
        string baseState = num2str(2*nQQtmp,1)+"q"+num2str(nGtmp,1)+"g";
        for (int iBranchType = 1; iBranchType <= 3; ++iBranchType) {
          string state = baseState;
          if (iBranchType == 3) state +="SplitB";
          else if (iBranchType == 2) state +="SplitA";
          else state +="Emit";
          string hisTitle="Log10(ME/AntPhys):" + state;
          vinciaHistos[hisTitle]=HPME2LogUSR;
          hisTitle="Log10(ME/AntTrial):" + state;
          vinciaHistos[hisTitle]=HPME2LogUSR;
        }
      }
    }

    // Limit highest multiplicity histograms by matching or verbose
    int nMultHi = 2 + matchingLO;
    if (verbose >= 7) nMultHi = max(nMultHi,10);
    if (verbose >= 8) nMultHi = max(nMultHi,100); 

    // Histos for pTevol scales and alphaS
    nBin   = 100;
    binWid = 0.1;
    wid    = nBin * binWid;
    xMax   = 0.;
    xMin   = xMax - wid;
    Hist HqLn("Ln(q2/sSystem) distribution (USR)",nBin,xMin,xMax);
    Hist HalphaS("alphaS distribution (USR)", nBin, 0., 2.);
    Hist HalphaSratio("alphaS/alphaTrial distribution (USR)", nBin, 0., 2.);
    Hist Hari("pAri distribution", 10, 0.0, 2.0);
    for (int nGtmp = 0; nGtmp <= nMultHi; ++nGtmp) {
      for (int nQQtmp = 0; 2*nQQtmp + nGtmp <= nMultHi; ++nQQtmp) {
        string baseState = num2str(2*nQQtmp,1)+"q"+num2str(nGtmp,1)+"g";
        for (int iBranchType = 1; iBranchType <= 3; ++iBranchType) {
          string state = baseState;
          if (iBranchType == 3) state +="SplitB";
          else if (iBranchType == 2) state +="SplitA";
          else state +="Emit";
          string hisTitle="Ln(q2trial/sSystem):" + state;
          vinciaHistos[hisTitle]       = HqLn;
          hisTitle="Ln(q2/sSystem):" + state;
          vinciaHistos[hisTitle]       = HqLn;
          string hisTitleAlphaS = "alphaS:" + state;
          vinciaHistos[hisTitleAlphaS] = HalphaS;
          string hisTitleAlphaSratio = "alphaSratio:" + state;
          vinciaHistos[hisTitleAlphaSratio] = HalphaSratio;
          string hisTitleAri = "pAri:" + state;
          vinciaHistos[hisTitleAri]    = Hari;
        }
      }
    }
    
    // Histogram of QE/eCMHad for the first branching
    Hist hQ2E("QE/eCM",1000,0.0,1.0);
    vinciaHistos["1stBranchingQE/eCM"] = hQ2E;
  }
  
  //*********
  
  // Add trial functions to a Branchelemental
  void TimeAntShower::resetTrialGenerators(BranchelementalFSR* trial) {
    
    // Reset
    trial->clearTrialGenerators();
    
    // Parent ID
    int id1      = trial->old1.id();
    int id2      = trial->old2.id();
    int colType1 = trial->old1.colType();
    int colType2 = trial->old2.colType();
    // (Treat color sextets as triplets here)
    if (abs(colType1) >= 3) colType1 = colType1 % 2;
    if (abs(colType2) >= 3) colType2 = colType2 % 2;
    
    // Gluon emission: soft-eikonal trial terms
    int iAntPhys = -1;
    bool isOn = false;
    // QQbar -> QGQbar
    if ( colType1 == 1 && colType2 == -1 ) {
      iAntPhys = iQQemit;
      isOn     = settingsPtr->flag("Vincia:QQemit");
      if (isOn) trial->addTrialGenerator(iAntPhys, false, &trialSoft);
    }
    // QG -> QGG
    else if ( colType1 == 1 && colType2 == 2 ) { 
      iAntPhys = iQGemit;
      isOn     = settingsPtr->flag("Vincia:QGemit");
      if (isOn) {
        trial->addTrialGenerator(iAntPhys, false, &trialSoft);
        if (sectorShower) 
          trial->addTrialGenerator(iAntPhys, false, &trialCollK);
      }
    }
    // GQbar -> GGQbar (physical antenna is obtained by swap of QG -> QGG)
    else if ( colType1 == 2 && colType2 == -1 ) { 
      iAntPhys = iQGemit;
      isOn     = settingsPtr->flag("Vincia:QGemit");
      if (isOn) {
        trial->addTrialGenerator(iAntPhys, true, &trialSoft);
        if (sectorShower) 
          trial->addTrialGenerator(iAntPhys, true, &trialCollI);        
      }
    }
    // GG -> GGG
    else if (colType1 == 2 && colType2 == 2) {
      iAntPhys = iGGemit;
      isOn     = settingsPtr->flag("Vincia:GGemit");
      if (isOn) {
        trial->addTrialGenerator(iAntPhys, false, &trialSoft);
        if (sectorShower) {
          trial->addTrialGenerator(iAntPhys, false, &trialCollI);
          trial->addTrialGenerator(iAntPhys, false, &trialCollK);
        }
      }
    }
    
    // XG -> XQbarQ
    if (id2 == 21) {
      if (colType1 == 2) {
        iAntPhys = iGGsplit;
        isOn = nGluonToQuarkF >= 1 && settingsPtr->flag("Vincia:GGsplit");
      } else {
        iAntPhys = iQGsplit;
        isOn = nGluonToQuarkF >= 1 && settingsPtr->flag("Vincia:QGsplit");
      }
      if (isOn) trial->addTrialGenerator(iAntPhys, false, &trialSplitK);
    }    
    // GX -> QbarQX (physical antenna obtained from swap of XG -> XQbarQ)
    if (id1 == 21) {
      if (colType2 == 2) {
        iAntPhys = iGGsplit;
        isOn = nGluonToQuarkF >= 1 && settingsPtr->flag("Vincia:GGsplit");
      } else {
        iAntPhys = iQGsplit;
        isOn = nGluonToQuarkF >= 1 && settingsPtr->flag("Vincia:QGsplit");
      }
      if (isOn) trial->addTrialGenerator(iAntPhys, true, &trialSplitI);      
    }  
  }
  
  // ******
  // 
  // generateKinematics(): generates the post-branching momenta 
  //               assumes branching scale trialPtr->qNew already generated
  //     Checks that require knowing only the local (2->3) branching 
  //     kinematics (without referring to the full event) belong here.
  // Steps:
  //     *) Generate second branching invariant, given qNew
  //     *) Set flavours and masses
  //     *) Check local hadronization veto inside trial antenna
  //        and construct kinematics in LAB frame
  //        Construct new partons
  // Return true if physical kinematics could be generated, otherwise false

  bool TimeAntShower::generateKinematics(BranchelementalFSR* trialPtr) {
    
    // Verbose output
    if (verbose >= 7)
      printOut("TimeAntShower::generateKinematics","begin --------------");

    //************************************************************************
    // Basic info about trial function and scale
    //************************************************************************

    int iTrial     = trialPtr->getTrialIndex();
    if (iTrial < 0) return false;
    int iAntPhys   = trialPtr->getPhysIndex(iTrial);
    bool isSwapped = trialPtr->getIsSwapped(iTrial);

    double qNew  = trialPtr->getTrialScale(iTrial);
    double sAnt  = trialPtr->s();

    // Consistency check: antenna mass and branching scale > TINY
    if ( sAnt < TINY) {
      if (verbose >= 3 && sAnt < TINY) {
        printOut("TimeAntShower::generateKinematics",
                 "Warning: s ="+num2str(sAnt)+"  Q2 = "+num2str(pow2(qNew)));
      }
      // Count up local and global failures due to hadronization and return
      trialPtr->nHadr++;
      ++nTrialsCutM;
      return false;
    }
    
    //************************************************************************
    // Generate zeta variable, then work out y12, y23
    //************************************************************************
    
    double y12, y23;
    bool pass = trialPtr->genTrialInvariants(y12, y23, iTrial);
    // Check phase-space hull
    if (! pass) {
      if (verbose >= 6) printOut("TimeAntShower::generateKinematics","failed "
                                 "genTrialInvariants, returning.");
      // Count up local and global failure rate due to hull and return false
      trialPtr->nHull++;
      ++nTrialsHull;
      return false;
    }

    //************************************************************************
    // Set flavors and masses
    //************************************************************************

    // Gluon emission: inherit parent flavors and add middle gluon 
    if (iAntPhys == iQQemit || iAntPhys == iQGemit || iAntPhys == iGGemit) {
      trialPtr->new1.id(trialPtr->old1.id());
      trialPtr->new2.id(21);
      trialPtr->new3.id(trialPtr->old2.id());
      trialPtr->new1.m(trialPtr->old1.m());
      trialPtr->new2.m(0.0);
      trialPtr->new3.m(trialPtr->old2.m());
    }

    // Gluon splitting:
    else if ( iAntPhys == iQGsplit || iAntPhys == iGGsplit) {

      // Set flavor of splitting
      double nF = trialPtr->getColFac(iTrial);
      int splitFlavor = int(rndmPtr->flat() * nF) + 1;

      // QG->QQbar'Q' and GG->GQbarQ
      if ( !isSwapped ) {
        trialPtr->new1.id(trialPtr->old1.id());
        trialPtr->new2.id(-splitFlavor);
        trialPtr->new3.id( splitFlavor);
        trialPtr->new1.m(trialPtr->old1.m());
        if (splitFlavor <= nFlavMassless) {
          trialPtr->new2.m(0.);
          trialPtr->new3.m(0.);
        } else {
          trialPtr->new2.m(trialPtr->new2.m0());
          trialPtr->new3.m(trialPtr->new3.m0());
        }
      } 
      // GQbar->Qbar'Q'Qbar and GG->QbarQG
      else {
        trialPtr->new1.id(-splitFlavor);
        trialPtr->new2.id( splitFlavor);
        trialPtr->new3.id(trialPtr->old2.id());
        if (splitFlavor <= nFlavMassless) {
          trialPtr->new1.m(0.);
          trialPtr->new2.m(0.);
        } else {
          trialPtr->new1.m(trialPtr->new1.m0());
          trialPtr->new2.m(trialPtr->new2.m0());
        }
        trialPtr->new3.m(trialPtr->old2.m());
      } 
      
    }
    
    // Check phase-space boundaries
    bool insidePS = true;
    if (y12 + y23 > 1.) insidePS = false;

    // Check whether we are doing any approximate mass corrections to phase space
    if ( insidePS && (iAntPhys == iQGsplit || iAntPhys == iGGsplit) 
         && massCorrectionSplit >= 1) {

      // g->QQ splittings cut off at Q2 = 4MQ 
      if (trialPtr->new2.idAbs() > nFlavZeroMass) {
        double yQQ = (isSwapped) ? y12 : y23;
        double sQQ = sAnt * yQQ;
        double mQ  = trialPtr->new2.m0();
        if (sQQ <= 4.0 * pow2(mQ)) {
          insidePS = false;
        }

        // When adding positive term to antenna function, also be more
        // restrictive with phase space, checking the other invariants
        if (insidePS && massCorrectionSplit >= 2) {
          double yQN = (isSwapped) ? y23 : y12;
          double sQN1 = sAnt * yQN;
          double sQN2 = sAnt * (1.0-yQQ-yQN);
          double mN = (isSwapped) ? trialPtr->new3.m0() : trialPtr->new1.m0();
          double sMin = pow2(mN + mQ);
          if (sQN1 < sMin || sQN2 < sMin) insidePS = false;
        }

        // Dalitz check
        if (insidePS && massCorrectionSplit >= 2) {
          double m1 = trialPtr->new1.m0();
          double m2 = trialPtr->new2.m0();
          double m3 = trialPtr->new3.m0();
          double s1 = pow2(m1);
          double s2 = pow2(m2);
          double s3 = pow2(m3);
          // Check if we could make massive momenta
          double d12 = y12*sAnt - s1 - s2;
          double d23 = y23*sAnt - s2 - s3;
          double d13 = sAnt - d12 - d23 - s1 - s2 - s3;
          double mAnt = sqrt(sAnt);
          double e1chk = (sAnt*(1 - y23) + s1)/(2*mAnt);
          double e3chk = (sAnt*(1 - y12) + s3)/(2*mAnt);
          double e2chk = mAnt - e1chk - e3chk;
          if (e1chk < m1 || e2chk < m2 || e3chk < m3) insidePS = false;

          double Delta = (d12*d23*d13 - pow2(d12)*s3 - pow2(d23)*s1 - pow2(d13)*s2
                          + 4*s1*s2*s3) / 4.;
          if ( insidePS && (d12 < 0. || d23 < 0. || d13 < 0. || Delta < 0.)  ) 
            insidePS = false;

        }
        
      }
      
    } else  if (insidePS && massCorrectionEmit >= 2) {
      double m1 = trialPtr->new1.m0();
      double m2 = trialPtr->new2.m0();
      double m3 = trialPtr->new3.m0();
      if (trialPtr->new1.idAbs() <= nFlavZeroMass) m1 = 0.0;
      if (trialPtr->new2.idAbs() <= nFlavZeroMass) m2 = 0.0;
      if (trialPtr->new3.idAbs() <= nFlavZeroMass) m3 = 0.0;
      if (m1 > 0.0 || m2 > 0.0 || m3 > 0.0) {
        double s1 = pow2(m1);
        double s2 = pow2(m2);
        double s3 = pow2(m3);
        // If we are doing a gluon emission, treat trial generator as if it
        // had generated dot products (captures Q->Qg propagator): add m back
        // to get invariants
        if (iAntPhys == iQQemit || iAntPhys == iQGemit) {
          y12 += s1/sAnt;
          y23 += s3/sAnt;
        }
        // Check user threshold for g->qq
        else if (iAntPhys == iQGsplit || iAntPhys == iGGsplit) {
          double sQQ = sAnt * ( (isSwapped) ? y12 : y23 ) ;
          int idQ = trialPtr->new2.idAbs();
          if (idQ == 5 && sQQ < 4 * pow2(mb)) insidePS = false;
          else if (idQ == 4 && sQQ < 4 * pow2(mc)) insidePS = false;
          else if (idQ == 3 && sQQ < 4 * pow2(ms)) insidePS = false;
        }
        // Check if we could make massive momenta
        double d12 = y12*sAnt - s1 - s2;
        double d23 = y23*sAnt - s2 - s3;
        double d13 = sAnt - d12 - d23 - s1 - s2 - s3;
        double mAnt = sqrt(sAnt);
        double e1chk = (sAnt*(1 - y23) + s1)/(2*mAnt);
        double e3chk = (sAnt*(1 - y12) + s3)/(2*mAnt);
        double e2chk = mAnt - e1chk - e3chk;
        if (e1chk < m1 || e2chk < m2 || e3chk < m3) insidePS = false;
        double Delta = (d12*d23*d13 - pow2(d12)*s3 - pow2(d23)*s1 - pow2(d13)*s2
                        + 4*s1*s2*s3) / 4.;
        if ( d12 < 0. || d23 < 0. || d13 < 0. || Delta < 0.) 
          insidePS = false;
      }
    }
    
    if (not insidePS) {
      // Count up local and global failure rate due to hull and return false
      trialPtr->nHull++;
      ++nTrialsHull;
      return false;      
    }
    
    //************************************************************************
    // Check local hadronization veto(s) : inside this Branchelemental
    //************************************************************************

    // Require 12 and 23 invariants at least above lowest physical meson mass 
    // (consistent with length > thickness assumption of string model)
    // Note: means we check m(qq) for g->qq splitting as well, but that is
    // presumably not a significant effect.
    
    // 12
    double mMin12 = vinComPtr->mHadMin(trialPtr->new1.id(), trialPtr->new2.id());
    if (y12 < pow2(1.01*mMin12)/sAnt) {
      if (verbose >= 5) printOut("TimeAntShower::generateKinematics",
                                 "=== Branching Vetoed. m12 < 1.01*mMes.");
      // Count up local and global failures due to hadronization and return
      trialPtr->nHadr++;
      ++nTrialsCutM;
      return false;
    }
    double mMin32 = vinComPtr->mHadMin(trialPtr->new3.id(), trialPtr->new2.id());
    if (y23 < pow2(1.01*mMin32)/sAnt) {
      if (verbose >= 5) printOut("TimeAntShower::generateKinematics",
                                 "=== Branching Vetoed. m23 < 1.01*mMes.");
      // Count up local and global failures due to hadronization and return
      trialPtr->nHadr++;
      ++nTrialsCutM;
      return false;
    }

    // Check ordinary hadronization cutoff: Ariadne pT or alternatives 
    double q2forHad = y12 * y23 * sAnt;
    // Alternative: Cutoff in Dipole dot product
    if (cutoffType == 2) q2forHad = min(y12,y23) * sAnt;
    // Check cutoffScale
    if (q2forHad < pow2(cutoffScale)) {
      if (verbose >= 9) {
        printOut("TimeAntShower::generateKinematics",
                 "qHad = " + num2str(sqrt(q2forHad), 0)
                 + " < qCut = " + num2str(cutoffScale, 0) + ", veto.");
      }
      // Count up local and global failures due to hadronization and return
      trialPtr->nHadr++;
      ++nTrialsCutQ;
      return false;
    }
    
    //***********************************************************************
    // Generate full kinematics for this trial branching
    //***********************************************************************
    
    // Define pointer to Global or Sector antenna, depending on shower setting
    // At this point, we only really need the kinematics map from antUsrPtr
    AntennaFSR* antUsrPtr = getAnt(iAntPhys,sectorShower,0);

    // Generate random (uniform) phi angle
    double phi = 2 * M_PI * rndmPtr->flat();
    
    // Generate branching kinematics, starting from dipole-antenna parents
    vector<Vec4> pOld, pNew;
    pOld.push_back(trialPtr->old1.p());
    pOld.push_back(trialPtr->old2.p());
    int kineMapType  = antUsrPtr->getKineMapType();
    vinCluPtr->map2to3(pNew, pOld, kineMapType, y12, y23, phi);
    
    // Save momenta
    trialPtr->new1.p(pNew[0]);
    trialPtr->new2.p(pNew[1]);
    trialPtr->new3.p(pNew[2]);
    
    // Set default polarizations: unpolarized (may be assigned later)
    trialPtr->new1.pol(9); 
    trialPtr->new2.pol(9); 
    trialPtr->new3.pol(9); 
    
    // Note: colour flow is not assigned here, since this requires 
    // knowledge of which is the next global colour tag available in
    // the event.
    
    // Set mothers
    trialPtr->new1.mothers(trialPtr->iOld1,trialPtr->iOld2);
    trialPtr->new2.mothers(trialPtr->iOld1,trialPtr->iOld2);
    trialPtr->new3.mothers(trialPtr->iOld1,trialPtr->iOld2);
        
    // All clear
    return true;
  }

  // ******
  // 
  // acceptTrial : main function to accept or reject a trial branching 
  //               after full branching kinematics have been constructed.
  //     Checks and corrections that require knowing the full event kinematics
  //     belong here.
  //     *) check sector ordering (if doing sector showers) 
  //     *) Optional: check global hadronization vetos
  //     *) Optional: apply beyond-LL (beyond 2->3) improvements 
  //     *) Assign helicities 
  //     *) Optional: match to fixed-order QCD matrix elements (from MadGraph)
  //     *) Optional: post-processing of uncertainty variations
  //     *) Accept or Reject Trial Emission
  // 
  
  bool TimeAntShower::acceptTrial(Event& event, BranchelementalFSR* trialPtr,
                                  vector<Particle> partsNow, vector<int> antPos,
                                  double Q2hatNew) {
    
    // Verbose output
    if (verbose >= 7)
      printOut("TimeAntShower::acceptTrial","begin --------------");
    
    // Basic info about trial function and scale
    int iTrial     = trialPtr->getTrialIndex();
    if (iTrial < 0) return false;
    double qNew    = trialPtr->getTrialScale(iTrial);
    double q2New   = pow2(qNew);
    int iAntPhys   = trialPtr->getPhysIndex(iTrial);
    bool isSwapped = trialPtr->getIsSwapped(iTrial);
    
    // How many variations for this trial?
    bool uncertaintyBandsNow = (uncertaintyBands && hardSys[iSysWin]);
    int nVariationsNow       = (uncertaintyBandsNow ? nWeights : 1);

    //***********************************************************************
    // Optional: start by checking section veto (if doing sector showers) 
    //***********************************************************************
 
    if (sectorShower) {
      if (!sectorAccept(event,trialPtr)) {
        if (verbose >= 5) printOut("TimeAntShower::acceptTrial",
                                   "Branching Vetoed. Failed sector veto.");
        // Count up local and global failure rate due to hull and return false
        trialPtr->nHull++;
        ++nTrialsHull;
        return false;
      } else { 
        if (verbose >= 7) printOut("TimeAntShower::acceptTrial",
                                   "Trial passed sector veto. Continuing...");
      }    
    }
    
    //************************************************************************
    // Check global hadronization veto(s) (requiring full event)
    // Note: here only check neighbor invariants. Branching invariants were
    // already checked above. 
    // Note 2: for neighbors, we are only checking invariant mass > pion mass.
    // Could in principle be made more elaborate by checking cutoffScale
    // as well, given sufficient reason to do so. 
    //**********************************************************************

    // Require all neighbor dot products at least above pion mass
    // (consistent with length > thickness assumption of string model)
    double sAnt  = trialPtr->s();
    double m2min = sAnt;
    double m2pi  = pow2(particleDataPtr->m0(111));
    if (trialPtr->iL != 0)
      m2min = m2(event[trialPtr->iL],trialPtr->new1);
    if (trialPtr->iR != 0) 
      m2min = min(m2min,m2(event[trialPtr->iR],trialPtr->new3));
    if (m2min < 1.05*m2pi) {
      if (verbose >= 5) printOut("TimeAntShower::acceptTrial","=== Branching Vetoed."
                                 " m2neighbor < 1.05*m2pi.");
      trialPtr->nHadr++;
      ++nTrialsCutM;
      return false;
    }

    //**********************************************************************
    // Masses
    //**********************************************************************

    // Dot products and mass-corrected scaled invariants
    double s12 = 2.0*dot4(trialPtr->new1,trialPtr->new2);
    double s23 = 2.0*dot4(trialPtr->new2,trialPtr->new3);
    double y12 = s12/sAnt;
    double y23 = s23/sAnt;
    // If y12 and y23 were shifted with mass terms, shift back before
    // evaluation trial and physical antenna terms
    if (massCorrectionEmit >= 3
        && (iAntPhys == iQQemit || iAntPhys == iQGemit)) {
      double m1 = trialPtr->new1.m0();
      double m3 = trialPtr->new3.m0();
      if (trialPtr->new1.idAbs() <= nFlavZeroMass) m1 = 0.0;
      if (trialPtr->new3.idAbs() <= nFlavZeroMass) m3 = 0.0;
      if (m1 > 0.0 || m3 > 0.0) {
        double s1 = pow2(m1);
        double s3 = pow2(m3);
        y12 -= s1/sAnt;
        y23 -= s3/sAnt;
      }      
    }
    
    //**********************************************************************
    // Helicity of mother partons 
    //**********************************************************************

    double hA = trialPtr->old1.pol();
    double hB = trialPtr->old2.pol();
    bool isPolarized = usePolarizationNow && (hA != 9 && hB != 9);
    
    //**********************************************************************
    // Compute spin-summed trial antenna sum for these variables
    // including colour and headroom factors
    //**********************************************************************

    // Sum over individual trial terms
    // Also, check whether all trial generators are phrased in terms
    // of the dot products (which is what we compute above) or whether 
    // eg the gluon-splitting ones are phrased in terms of the real 
    // invariant masses?
    double antTrialSum = 0.0;
    for (int iSum = 0; iSum < (int)trialPtr->nTrialGenerators(); ++iSum) {
      // Only include terms that correspond to the current physical antenna
      if (trialPtr->getPhysIndex(iSum) != iAntPhys ||
          trialPtr->getIsSwapped(iSum) != isSwapped ) continue; 
      // Evaluate trial term
      double antTrial    = trialPtr->trialGenPtrSav[iSum]->aTrial(y12,y23);
      // Get HeadRoom and Color factors
      double headRoomFac = trialPtr->getHeadRoomFactor(iSum);
      double colFac      = trialPtr->getColFac(iSum);
      // (color factor for splitings is just nF, which is not included in
      // physical antenna normalization nor in matrix elements)
      if (iAntPhys == iQGsplit || iAntPhys == iGGsplit) 
        colFac = getAnt(iAntPhys)->getChargeFactor();
      antTrial *= colFac * headRoomFac;
      antTrialSum += antTrial;
    }

    //**********************************************************************
    // If physical antenna function is mirror of current trial, translate to
    // swapped invariants for antenna-function evaluation
    //**********************************************************************

    double hAant  = hA;
    double hBant  = hB;
    double y12ant = y12;
    double y23ant = y23;
    if ( isSwapped ) {
      hAant  = hB;
      hBant  = hA;
      y12ant = y23;
      y23ant = y12;
    }
    
    //**********************************************************************
    // Check if we are going to do MECs, makes accept probs with CMW easier
    //**********************************************************************
    
    // QCD order as in Born*QCD
    int nQCDnow = nBranchs[iSysWin]+1;
    // Only up to matchingLO
    bool doMEC = (meCorrections && nQCDnow <= matchingLO);
    // Only for hard process
    if (!hardSys[iSysWin]) doMEC = false;
    // We are already below the cutoff (only checking for last order)
    if ( (nQCDnow == matchingLO) && (Q2hatNew < pow2(matchingIRcutoff)) ) 
      doMEC = false;

    // Current number of quarks and gluon
    int nGnow  = nG[iSysWin];
    int nQQnow = nQQ[iSysWin];
    if (trialPtr->new2.id() == 21) ++nGnow; 
    else {
      ++nQQnow;
      nGnow -= 1;
    }
    
    // Compute matching scale dampening factor (if imposed)
    double DeltaDamp = 1.0;
    double q2damp    = Q2hatNew;
    double q2match   = Q2fac;
    if (doMEC && nQCDnow >= matchingRegOrder && matchingRegOrder != 0) {
      // If matching scale is abolute use given scale
      if (matchingScaleIsAbs) q2match = pow2(matchingScale);
      // If matching scale is relative use given ratio
      else q2match = pow2(matchingScaleRatio)*q2match;
      // MatchingRegShape = 0: Step Function
      if (matchingRegShape == 0) {
        if (q2damp < q2match) {
          DeltaDamp = 0.0;
          doMEC = false;
        }
        else DeltaDamp = 1.0;
      }
      // MatchingRegShape = 1: linear in log(Q) between 1/2 QM2 < Q2 < 2 QM2
      else {
        double log2R = log(q2damp/q2match)/log(2.);
        if (log2R < -1) {
          DeltaDamp = 0.0;
          doMEC = false;
        } else if (log2R > 1.) {
          DeltaDamp = 1.0;
        } else {            
          DeltaDamp = 0.5*log2R + 0.5 ;
        }
      }
    }
    
    // Set all relevant properties and check if we can do the MEC
    if (doMEC) {
      MEcorrectionsPtr->setPari(doPariNow[iSysWin]);
      MEcorrectionsPtr->setProcClass(MEclass[iSysWin]);
      MEcorrectionsPtr->setClustProps(nQLneeded[iSysWin], Hint[iSysWin]);
      doMEC = MEcorrectionsPtr->canDoMEC(partsNow, hardSys[iSysWin]);
    }
        
    //***********************************************************************
    // AlphaS
    // Impose default choice. Can differ slighly from trial even when running
    // inside trial integral, due to flavor thresholds. Here, alphaS(mu) is
    // returned directly, with the number of flavors active at mu, whereas 
    // the number of flavors in the trial integral is controlled by the value
    // of the trial scale.
    //***********************************************************************
    
    double alphaTrial = trialPtr->getAlphaTrial(iTrial);
    if (isnan(alphaTrial)) cout<<" alphaTRIAL IS NAN!!!"<<endl;
    if (alphaTrial != alphaTrial) cout<<" alphaTRIAL IS != alphaTRIAL"<<endl;
    // Scale is pT for gluon emission and mQQ for gluon splitting
    double mu2     = s12*s23/sAnt;
    double kMu2Def = aSkMu2Def;
    double kMu2Usr = aSkMu2;
    if (trialPtr->new2.colType() == -1) {
      mu2     = s23;
      kMu2Def = aSkMu2SplitDef;
      kMu2Usr = aSkMu2Split;
    } else if (trialPtr->new2.colType() == 1) {
      mu2    = s12;
      kMu2Def = aSkMu2SplitDef;
      kMu2Usr = aSkMu2Split;
    }
    double mu2Def       = max(mu2minDef, mu2freezeDef + mu2*kMu2Def);
    double mu2Usr       = max(mu2min   , mu2freeze    + mu2*kMu2Usr);
    // alphaS values
    double alphaSdef    = min(alphaSmaxDef, alphaSdefPtr->alphaS(mu2Def));
    double alphaSusr    = min(alphaSmax,    alphaSptr->alphaS(mu2Usr));
    double alphaSCMWdef = min(alphaSmaxDef, alphaSCMWdefPtr->alphaS(mu2Def));
    double alphaSCMWusr = min(alphaSmax,    alphaSCMWptr->alphaS(mu2Usr));

    // Verbose output
    if (verbose >= 8)
      printOut("TimeAntShower::acceptTrial",
               "alphaTrial = "+num2str(alphaTrial)+", "
               "alphaSdef = "+num2str(alphaSdef)+", alphaSCMWdef = "+num2str(alphaSCMWdef)+", "
               "alphaSusr = "+num2str(alphaSusr)+", alphaSCMWusr = "+num2str(alphaSCMWusr));
        
    //************************************************************************
    // Compute spin-summed physical antennae (spin selection below)
    // Store result in vector representing uncertainty variations
    // iVar = 0 : User settings
    //        1 : Default settings
    //        2 : AlphaS-Hi (user settings, alphaS(Q/k))
    //        3 : AlphaS-Lo (user settings, alphaS(Q*k))
    //        4 : Ant-Hi    (user settings, MAX ants)
    //        5 : Ant-Lo    (user settings, MIN ants)
    //        6 : NLO-Hi    (user settings, *(1+alphaS))
    //        7 : NLO-Lo    (user settings, /(1+alphaS))
    //        8 : NLC-Hi    (user settings, CA for all emissions)
    //        9 : NLC-Lo    (user settings, 2CF for all emissions)
    //       10 : Pimp      (user settings, Pimp = 1/(1+q^4/Q^4)
     //************************************************************************

    // Define pointers to user, default, and uncertainty-variation antennae
    AntennaFSR* antUsrPtr = getAnt(iAntPhys,sectorShower,0);
    AntennaFSR* antDefPtr = getAnt(iAntPhys,sectorShower,1);
    AntennaFSR* antMaxPtr = getAnt(iAntPhys,sectorShower,2);
    AntennaFSR* antMinPtr = getAnt(iAntPhys,sectorShower,3);

    // Compute spin-summed antenna function 
    if (verbose >= 8) 
      printOut("TimeAntShower::acceptTrial","Evaluating antenna function");

    vector<double> helSum;
    helSum.resize(nVariationsNow);    
    // Unpolarized case: ignore parent helicities
    if (!isPolarized) {
      helSum[0] = antUsrPtr->antennaFunction(y12ant, y23ant, 0., 0. ,0.,
                                             9, 9, 9, 9, 9);
      if (uncertaintyBandsNow) {
        helSum[1] = antDefPtr->antennaFunction(y12ant, y23ant, 0., 0. ,0.,
                                               9, 9, 9, 9, 9);
        // Use USR antennae as basis for uncertainty variations
        for (int iVar=2; iVar<nVariationsNow; ++iVar) 
          helSum[iVar] = helSum[0];
        helSum[4] = antMaxPtr->antennaFunction(y12ant, y23ant, 0., 0. ,0.,
                                               9, 9, 9, 9, 9);
        helSum[5] = antMinPtr->antennaFunction(y12ant, y23ant, 0., 0. ,0.,
                                               9, 9, 9, 9, 9);
      }
    }
    // Polarized case: use parent helicities (but restrict to massless)
    else {       
      helSum[0] = antUsrPtr->antennaFunction(y12ant, y23ant, 0., 0., 0.,
                                             hAant, hBant, 9, 9, 9);
      if (uncertaintyBandsNow) {
        helSum[1] = antDefPtr->antennaFunction(y12ant, y23ant, 0., 0., 0.,
                                               hAant, hBant, 9, 9, 9);
        // Use USR antennae as basis for uncertainty variations
        for (int iVar=2; iVar<nVariationsNow; ++iVar) 
          helSum[iVar] = helSum[0];
        helSum[4] = antMaxPtr->antennaFunction(y12ant, y23ant, 0., 0., 0.,
                                               hAant, hBant, 9, 9, 9);
        helSum[5] = antMinPtr->antennaFunction(y12ant, y23ant, 0., 0., 0., 
                                               hAant, hBant, 9, 9, 9);
      }
    }

    // Verbose output
    if (verbose >= 8)
      printOut("TimeAntShower::acceptTrial","helSum  = "+num2str(helSum[0]/sAnt));
    
    //************************************************************************
    // Store (partial) helicity sum for each variation
    //************************************************************************

    vector<double> antPhys;
    for (int iVar=0; iVar<nVariationsNow; ++iVar) 
      antPhys.push_back(helSum[iVar]);
    
    //************************************************************************
    // Antenna Mass Correction Terms
    //************************************************************************
    
    if (massCorrectionEmit >= 3 && (iAntPhys == iQQemit || iAntPhys == iQGemit) ) {
      // Mass corrections to gluon emission (negative)
      int id1 = trialPtr->new1.idAbs();
      double m1 = trialPtr->new1.m0();
      if (id1 <= nFlavZeroMass) m1 = 0.0;
      int id3 = trialPtr->new3.idAbs();
      double m3 = trialPtr->new3.m0();
      if (id3 <= nFlavZeroMass) m3 = 0.0;
      if (m1 > 0. || m3 > 0.) {
        double s1 = pow2(m1);
        double s3 = pow2(m3);
        double d12 = y12 * sAnt;
        double d23 = y23 * sAnt;
        for (int iVar = 0; iVar < nVariationsNow; ++iVar) {
          // Eikonal mass corrections
          antPhys[iVar] -= (2*s1/pow2(d12) + 2*s3/pow2(d23)) * sAnt;
        }
      }
    }

    // Mass corrections to gluon splitting (positive)
    else if (massCorrectionSplit >= 2 && trialPtr->new2.idAbs() > nFlavZeroMass
             && (iAntPhys == iQGsplit || iAntPhys == iGGsplit) ) {
        double sQQ = y23ant * sAnt;
        double yij = y12ant;
        double yik = 1.0 - y23ant - y12ant;
        double mQ = trialPtr->new2.m0();
        double m2q = pow2(mQ);
        for (int iVar = 0; iVar < nVariationsNow; ++iVar) {
          // Phrase as multiplicative correction, defined so as not to become
          // singular in the point (yij,yik) = 0.
          antPhys[iVar] *= 1.0 + 2 * m2q / sQQ / (1.0 - 2*yij*yik);
        }
    }
    
    //************************************************************************
    // HYPERJET: special mode for processing events from HEJ 
    // => matching to soft-resummed matrix elements
    // Ignore otherwise
    //************************************************************************

    if ( hyperjet && (iAntPhys == iQQemit || iAntPhys == iQGemit
                      || iAntPhys == iGGemit) ) {
      // If running in hyper-jet mode, subtract soft Eikonal for gluon emission
      // for emissions off the hard interaction
      if (verbose >= 8) 
        printOut("TimeAntShower::acceptTrial","HyperJet mode: Subtracting Eikonal");
      for (int iVar=0; iVar<nVariationsNow; ++iVar) 
        antPhys[iVar] -= 2.0*(1.-y12-y23)/(y12*y23);
    }
    
    //**********************************************************************
    // CMW rescaling:
    // alphaS' = CMW alphaS, a = eik+coll, a' = alphaS'/alphaS*eik+coll
    // generated branchings according to trial = alphaS'*aHat
    // accept probability without MECs: alphaS*a'/(alphaS'*aHat)
    // accept probability with    MECs: alphaS*a/(alphaS'*aHat) * PME
    // -> PME includes the CMW improvement, no CMW here
    // TODO check for helicities (also in assignment part)
    //**********************************************************************
    
    if ( (trialPtr->new2.id() == 21) && !doMEC ) {
      
      double eik1       = 2.0/y12ant/y23ant;
      double eik2       = 2.0*(1.0-y12ant-y23ant)/y12ant/y23ant;
      // User
      double eikUsr     = 0.0;
      double subTermUsr = 0.0;
      if (CMWtype > 0) {
        eikUsr     = (CMWtype == 1 ? eik1 : eik2);
        subTermUsr = ((alphaSCMWusr/alphaSusr)-1.0)*eikUsr;
      }
      antPhys[0] = ( antPhys[0] + subTermUsr );
      
      if (uncertaintyBandsNow) {
        // Default
        double eikDef     = 0.0;
        double subTermDef = 0.0;
        if (CMWtypeDef > 0) {
          eikDef          = (CMWtypeDef == 1 ? eik1 : eik2);
          subTermDef      = ((alphaSCMWdef/alphaSdef)-1.0)*eikDef;
        }
        antPhys[1]        = ( antPhys[1] + subTermDef );
        // Additional scaling factor
        double mu2V2      = max(mu2min, mu2freeze + mu2*kMu2Usr/kMu2Var);
        double aSratioV2  = min(alphaSmax, alphaSCMWptr->alphaS(mu2V2)) /
                            min(alphaSmax, alphaSptr->alphaS(mu2V2));
        antPhys[2]        = ( antPhys[2] + (aSratioV2-1.0)*eikUsr );
        double mu2V3      = max(mu2min, mu2freeze + mu2*kMu2Usr*kMu2Var);
        double aSratioV3  = min(alphaSmax, alphaSCMWptr->alphaS(mu2V3)) /
                            min(alphaSmax, alphaSptr->alphaS(mu2V3));
        antPhys[3]        = ( antPhys[3] + (aSratioV3-1.0)*eikUsr );
        // USR choice for everyone else
        for (int iVar=4; iVar<nVariationsNow; ++iVar)
          antPhys[iVar] = ( antPhys[iVar] + subTermUsr );
      }
      
    }
    
    //************************************************************************
    // Apply color (charge) factor
    //************************************************************************

    for (int iVar=0; iVar<nVariationsNow; ++iVar) {
      double chargeFactor = antUsrPtr->getChargeFactor();
      if (iVar == 1) chargeFactor = antDefPtr->getChargeFactor();
      else if (iVar == 4) chargeFactor = antMaxPtr->getChargeFactor();
      else if (iVar == 5) chargeFactor = antMinPtr->getChargeFactor();
      else if (iVar == 8 && trialPtr->new2.id() == 21) chargeFactor = CA;
      else if (iVar == 9 && trialPtr->new2.id() == 21) chargeFactor = CF;
      antPhys[iVar] *= chargeFactor;
    }

    //***********************************************************************
    // Compute Pimp modification (only applied to global showers)
    //***********************************************************************

    // Pimp (allows suppressed unordered branchings)
    double Pimp = 1.0;
    if ( !sectorShower && trialPtr->unOrdering && (Q2hat[iSysWin] > 0.0) ) {
      Pimp = 1.0/(1.0 + q2New/Q2hat[iSysWin]);
      for (int iVar = 0; iVar < nVariationsNow; ++iVar) {
        // Modified Pimp factor
        if (iVar == 10) {
          double Pimp4 = 1.0/(1.0 + pow2(q2New/Q2hat[iSysWin]));
          antPhys[iVar] *= Pimp4;
        }
        else antPhys[iVar] *= Pimp;
      }
    }

    //***********************************************************************
    // Compute Pari modification (only applied to global showers)
    //***********************************************************************
    
    // Pari (ARIADNE factor for gluon splittings in e+e- -> Z -> jets)
    // Only for global showers, when neighbor is a final-state quark
    double Pari = 1.0;
    double sP = sAnt; 
    if ( !sectorShower && doPariNow[iSysWin] &&
         (iAntPhys == iQGsplit || iAntPhys == iGGsplit) ) {
      // Find initial particles in dipole branching.
      int iOld1 = trialPtr->iOld1;
      int iOld2 = trialPtr->iOld2;

      // Impose ARIADNE factor for g->qqbar.
      //       Pari = 2/(1+sN/sP)    (sN = s(neighbor), sP = s(parent))
      // If s(neighbor) << s(parent) -> splitting suppressed (screened?)
      // If s(neighbor) >> s(parent) -> splitting enhanced (up to factor 2)  

      double sN;
      int iN;
      // Which side of dipole-antenna is splitting? 
      if (event[iOld1].id() != 21 || trialPtr->new1.id() == 21) {
        iN = trialPtr->iR;
        if (iN != 0) { // Only apply if neighbor exists
          sN = m2(event[iOld2],event[iN]);
          Pari = 2.0*sN/(sP + sN);      
        }
      } else { 
        iN = trialPtr->iL;
        if (iN != 0) { // Only apply if neighbor exists
          sN = m2(event[iN],event[iOld1]);        
          Pari = 2.0*sN/(sP + sN);      
        }
      }
      // Check for negative Pari factor (normally should not happen!)
      if (Pari < 0.0) {
        if (verbose >= 1) 
          printErr("TimeAntShower::acceptTrial","Pari factor = "+num2str(Pari)+" < 0");
        Pari = 0.0;
      }
      // Apply Pari factor
      for (int iVar = 0; iVar < nVariationsNow; ++iVar)
        antPhys[iVar] *= Pari;            
    }
    
    //**********************************************************************
    // Compute accept probability : Physical/Trial 
    //**********************************************************************
    
    // Compute basic accept probability : Physical/Trial 
    for (int iVar = 0; iVar < nVariationsNow; ++iVar)
      Paccept[iVar] = antPhys[iVar]/antTrialSum;
    
    // Verbose output
    if (verbose >= 8) {
      printOut("TimeAntShower::acceptTrial", " ");
      cout << "   AntTrial/s = " << setprecision(6) << antTrialSum/sAnt << endl;
      cout << "   AntPhys/s  = " << setprecision(6) << antPhys[0]/sAnt
           << "      ( Pimp = "  << Pimp << " ; Pari = " << Pari << " )" << endl;
      cout << "   Paccept    = " << setprecision(6) << Paccept[0] << endl;
    }
        
    //***********************************************************************
    // Choose helicities for daughter particles 
    // (so far only for massless polarized shower)
    //***********************************************************************

    if ( isPolarized ) {
      // Sum over final-state helicities
      // (Slightly inefficient to first evaluate sum and then reevaluate
      // individual terms below, but perhaps acceptable for increased 
      // clarity. Some optimization is attempted by starting with the 
      // dominant (MHV) contributions below. Further optimization could
      // be achieved by implementing some form of caching of the individual
      // contributions.)
      // Generate random number
      double randHel = rndmPtr->flat() * helSum[0];
      // Select helicity
      int hi(0), hj(0), hk(0);
      double aHel(0.);
      for (hi = hAant; abs(hi) <= 1; hi -= 2*hAant) {
        for (hk = hBant; abs(hk) <= 1; hk -= 2*hBant) {
          for (hj = hAant; abs(hj) <= 1; hj -= 2*hAant) {
            aHel = antUsrPtr->antennaFunction(y12ant, y23ant, 0., 0. ,0.,
                                              hAant, hBant ,hi, hj, hk);
            randHel -= aHel;
            if (verbose >= 8) {
              printOut("TimeAntShower::acceptTrial", "antPhys("+
                       num2str(int(hAant))+" "+num2str(int(hBant))+">"+num2str(hi)+" "+
                       num2str(hj)+" "+num2str(hk)+") = "+num2str(aHel/sAnt)+
                       ", m(ar,rb) = "+num2str(sqrt(y12ant*sAnt))+", "+num2str(sqrt(y23ant*sAnt))+
                       ", isSwapped = "+bool2str(isSwapped)+
                       "; sum = "+num2str(helSum[0]/sAnt));
            }
            if (randHel < 0.) break;
          }
          if (randHel < 0.) break;
        }
        if (randHel < 0.) break;
      }
      // Uncertainty variations (would have picked these helicities with 
      // a different probability if using different antennae). 
      if (uncertaintyBandsNow) {
        double PhelUsr = aHel / helSum[0];
        Paccept[1] *= antDefPtr->antennaFunction(y12ant, y23ant, 0., 0. ,0.,
                                                 hAant, hBant ,hi, hj, hk)
                      / helSum[1] / PhelUsr;
        Paccept[4] *= antMaxPtr->antennaFunction(y12ant, y23ant, 0., 0. ,0.,
                                                 hAant, hBant ,hi, hj, hk)
                      / helSum[4] / PhelUsr;
        Paccept[5] *= antMinPtr->antennaFunction(y12ant, y23ant, 0., 0. ,0.,
                                                 hAant, hBant ,hi, hj, hk) 
                      / helSum[5] / PhelUsr;
      }
      if (verbose >= 7) 
        printOut("TimeAntShower::acceptTrial", "selected "+num2str((int)(hAant))+" "+
                 num2str(int(hBant))+" "+" -> "+num2str(hi)+" "+num2str(hj)+" "+
                 num2str(hk)+", isSwapped = "+bool2str(isSwapped));
      // Assign helicities (taking swapped invariants into account)
      if (!isSwapped) {
        //cout<<" Selected : " <<hA<<" "<<hB<<" > "<<hi<<" "<<hj<<" "<<hk<<endl;
        trialPtr->new1.pol(hi);
        trialPtr->new2.pol(hj);
        trialPtr->new3.pol(hk);
      } 
      else {
        //cout<<" Selected : " <<hB<<" "<<hA<<" > "<<hk<<" "<<hj<<" "<<hi<<endl;
        trialPtr->new1.pol(hk);
        trialPtr->new2.pol(hj);
        trialPtr->new3.pol(hi);        
      }
    }
    // If not polarized
    else {      
      trialPtr->new1.pol(9);
      trialPtr->new2.pol(9);
      trialPtr->new3.pol(9);
      usePolarizationNow = false;
    }

    // Store chosen polarizations also in vector representing system state
    partsNow[antPos[0]].pol(trialPtr->new1.pol());
    partsNow[antPos[2]].pol(trialPtr->new2.pol());
    partsNow[antPos[1]].pol(trialPtr->new3.pol());
    
    //**********************************************************************
    // Apply alphaS choices
    //**********************************************************************
    
    // Reweight central accept probability with user choice   
    Paccept[0] *= alphaSusr / alphaTrial;
    
    // Do uncertainty variations
    if (uncertaintyBandsNow) {  
      // Up/down variatons on USR alphaS
      double mu2V2    = max(mu2min, mu2freeze + mu2*kMu2Usr/kMu2Var);
      double mu2V3    = max(mu2min, mu2freeze + mu2*kMu2Usr*kMu2Var);
      double alphaSV2 = min(alphaSmax, alphaSptr->alphaS(mu2V2));
      double alphaSV3 = min(alphaSmax, alphaSptr->alphaS(mu2V3));
      // Reweight by alphaVar / alphaTrial
      for (int iVar = 1; iVar < nVariationsNow; ++iVar) {
             if (iVar == 1) Paccept[iVar] *= alphaSdef/alphaTrial;
        else if (iVar == 2) Paccept[iVar] *= alphaSV2/alphaTrial;
        else if (iVar == 3) Paccept[iVar] *= alphaSV3/alphaTrial;
        else                Paccept[iVar] *= alphaSusr/alphaTrial;
      }
    }
    
    //**********************************************************************
    // Matrix Element Corrections
    //**********************************************************************
      
    // Compute MEC and apply to accept probability
    if (doMEC) {
      PME = MEcorrectionsPtr->MEcorrection(partsNow);
      // Modify accept probabilities
      for (int iVar=0; iVar < nVariationsNow; ++iVar) 
        Paccept[iVar] *= 1.0 + DeltaDamp * (PME[iVar] - 1.0);
      // Fill diagnostics histos?
      if (verbose >= 3 || (verbose >= 2 && doMEC)) {
        string state = num2str(2*nQQnow,1) + "q" + num2str(nGnow,1) + "g";   
        if (trialPtr->new2.colType() == 2) state += "Emit";
        else if (trialPtr->new2.colType() == -1) state += "SplitA";
        else if (trialPtr->new2.colType() == 1) state += "SplitB";
        string HPacc = "Log10(ME/AntTrial):" + state;
        if (vinciaHistos.find(HPacc) != vinciaHistos.end())
          vinciaHistos[HPacc].fill(log10(max(TINY,Paccept[0])));
        string HqTrial = "Ln(q2trial/sSystem):" + state; 
        if (vinciaHistos.find(HqTrial) != vinciaHistos.end())
          vinciaHistos[HqTrial].fill(log(max(TINY,pow2(qNew/mSystem[iSysWin]))));
      }
    }
    
    //**********************************************************************
    // NLO K-factor variation (if not doing explicit NLO matching)
    // (largely degenerate with scale variation at LO*LL, but could in 
    // principle behave differently at (N)LO*(N)LL)
    //**********************************************************************
    
    bool doMatchNLO = false;
    if (uncertaintyBandsNow && !doMatchNLO) {
      double mu2nlo    = max( max(mu2min, mu2freeze), sAnt);
      double alphaSnlo = min(alphaSmaxDef, alphaSdefPtr->alphaS(mu2nlo));
      Paccept[6] *= (1.+alphaSnlo);
      Paccept[7] /= (1.+alphaSnlo);
    }
    
    //**********************************************************************
    // Print MC violations if no reweighting is allowed
    //**********************************************************************
    
    if (verbose >= 8)
      printOut("TimeAntShower::acceptTrial"," Trial Paccept = "
               +num2str(Paccept[0]));
      
    bool violation  = (Paccept[0] > 1.0 + TINY);
    bool negPaccept = (Paccept[0] < 0.0);
    
    if (violation) {
      // Statistics
      if (meCorrections && nQCDnow <= matchingLO && hardSys[iSysWin]) {
        if (doMEC) {
          nMCviolDidMEC++;
          if (Paccept[0] > maxMCviolDidMEC) maxMCviolDidMEC = Paccept[0];
        } else {
          nMCviolPlannedMEC++;
          if (Paccept[0] > maxMCviolPlannedMEC) maxMCviolPlannedMEC = Paccept[0];
        }
      } else {
        nMCviolShower++;
        if (Paccept[0] > maxMCviolShower) maxMCviolShower = Paccept[0];
      }
      // Automatically increase headroom factor when violations encountered
      setHeadRoom( (iAntPhys == iQGsplit || iAntPhys == iGGsplit) ? 1 : 0, -1, 
                   nG[iSysWin], nQQ[iSysWin], iSysWin, qNew, min(1.2,Paccept[0]) );
    } else if (negPaccept) {
      // Statistics
      if (meCorrections && nQCDnow <= matchingLO && hardSys[iSysWin]) {
        if (doMEC) {
          nNegPaDidMEC++;
          if (Paccept[0] < maxNegPaDidMEC) maxNegPaDidMEC = Paccept[0];
        } else {
          nNegPaPlannedMEC++;
          if (Paccept[0] < maxNegPaPlannedMEC) maxNegPaPlannedMEC = Paccept[0];
        }
      } else {
        nNegPaShower++;
        if (Paccept[0] < maxNegPaShower) maxNegPaShower = Paccept[0];
      }
    }
    
    if (violation || negPaccept) {
      bool report = false;
      bool isLast = false;
      
      // Report violation with MECs with verbose >= 3
      // only for first nReportMCviol times
      if (doMEC && verbose >= 3) {
        if (violation) {
          if (nMCviolDidMEC <= nReportMCviol) {
            report = true;
            if (nMCviolDidMEC == nReportMCviol) isLast = true;
          }
        }
        if (negPaccept) {
          if (nNegPaDidMEC <= nReportMCviol) {
            report = true;
            if (nNegPaDidMEC == nReportMCviol) isLast = true;
          }
        }
      }
      
      // Report violation pure shower with verbose >= 3 or
      //                                   verbose >= 1 && fussy
      // only for first nReportMCviol times
      else if (!doMEC && ( verbose >= 3 || (verbose >= 1 && 
                                            Paccept[0] > pAcceptFuss && 
                                            qNew > pAcceptPT0) )) {
        if (violation) {
          if (nMCviolPlannedMEC+nMCviolShower <= nReportMCviol) {
            report = true;
            if (nMCviolPlannedMEC+nMCviolShower == nReportMCviol) 
              isLast = true;
          }
        }
        if (negPaccept) {
          if (nNegPaPlannedMEC+nNegPaShower <= nReportMCviol) {
            report = true;
            if (nNegPaPlannedMEC+nNegPaShower == nReportMCviol) 
              isLast = true;
          }
        }
      }
      
      // Always report in debugging mode
      if (report || verbose >= 4) {
        printOut("TimeAntShower::acceptTrial",
                 (doMEC ? " Paccept(shower and ME, order" : " Paccept(shower, order")
                 +num2str(nQCDnow)+") = "+num2str(Paccept[0])+" at q = "
                 +num2str(qNew)+" for gluon "+((iAntPhys==iQGsplit || iAntPhys==iGGsplit) ? 
                 "splitting" : "emission"));
        if (verbose >= 4) {
          if (doMEC)
            cout <<" bare PME = "<<PME[0]<<" qDamp = "<<sqrt(q2damp)<<" qDamp/qMatch = 1/"
                 <<sqrt(q2match/q2damp)<<" DeltaDamp = "<<DeltaDamp<<endl;
          cout <<" IK -> ijk = "<<trialPtr->old1.id()<<" "<<trialPtr->old2.id()<<" -> "
               <<trialPtr->new1.id()<<" "<<trialPtr->new2.id()<<" "<<trialPtr->new3.id()
               <<" mij = "<<sqrt(m2(trialPtr->new1.p()+trialPtr->new2.p()))
               <<" mjk = "<<sqrt(m2(trialPtr->new2.p()+trialPtr->new3.p()))
               <<" mijk = "<<sqrt(sAnt)<<endl;
        }
        if (isLast && verbose < 4) cout <<" Further warnings suppressed!"<<endl;
      
        // Print more info for bad cases
        if (verbose >= 5) trialPtr->list();
        if ((Paccept[0] > 2.0 && verbose >= 4) || (Paccept[0] > 10.0 && verbose >= 3)) {
          cout << " Momentum Listing after the branching: " << endl;
          for (int ip=0; ip<(int)partsNow.size(); ip++)
            cout << " id" << ip << (ip > 9 ? " = " : "  = ") << num2str(partsNow[ip].id(),7)
                 << " p" << " = " << partsNow[ip].p();
          cout << " Momentum Listing before the branching: " << endl;
          for (int ip=0; ip<(int)particles[iSysWin].size(); ip++)
            cout << " id" << ip << (ip > 9 ? " = " : "  = ") << num2str(particles[iSysWin][ip].id(),7)
                 << " p" << " = " << particles[iSysWin][ip].p();
        }
      }
    }
    
    //**********************************************************************
    // Perform reweighting (if allowed)
    //**********************************************************************
    
    double acceptWeight = 1.0;
    double rejectWeight = 1.0;
    if (allowReweighting) {
      // f(t) = physical, g(t) = overestimate, h(t) = auxilary
      double physVSover = Paccept[0]; // = f(t)/g(t)
      double physVSaux  = physVSover; // = f(t)/h(t)
      double aFactor    = 1.0;        // = h(t)/g(t)
      if (physVSover >= 1.0) { // overestimate failed, choose h(t) = 1.1 f(t)
        physVSaux = 1.0/1.1;          // = f(t)/h(t) = 1/1.1 < 1
        aFactor   = 1.1*physVSover;   // = h(t)/g(t) = 1.1 f(t)/g(t)
      }
      else if (physVSover < 0.0) { // negative physical point, choose h(t) = - 1.1 f(t)
        physVSaux = 1.0/1.1;          // = f(t)/h(t) = 1/1.1 < 1
        aFactor   = 1.1*physVSover;   // = h(t)/g(t) = - 1.1 f(t)/g(t)
      }
      // Set accept/reject weights
      acceptWeight = aFactor; // = h(t)/g(t) [= 1 if overestimate ok]
      rejectWeight = (1.0-physVSover)/(1.0-physVSaux); // = (1-f(t)/g(t))/(1-f(t)/h(t)) [= 1 if overestimate ok]
      if (verbose >= 6) {
        printOut("TimeAntShower::acceptTrial","Reweighting event with:");
        cout << "  accept weight = " << acceptWeight << endl;
        cout << "  reject weight = " << rejectWeight << endl;
      }
      // Update accept probability (now < 1)
      Paccept[0] = physVSaux;
      if (uncertaintyBandsNow) {
        for (int iVar=1; iVar<nVariationsNow; iVar++)
          Paccept[iVar] /= aFactor; // = f'(t)/g(t) * 1/(h(t)/g(t)) = f'(t)/h(t)
      }
    }
    
    // Limit uncertainty to factor specified by the user for first branching
    // to capture possibly large finite term differences, then level off at 
    // 20% asymptotically, to capture multiple-emission uncertainties.
    if (uncertaintyBandsNow) {
      double limit = uncertaintyBandsMaxDev - 
                     min(0.8+uncertaintyBandsMaxDev-2.0,
                         0.05*pow2(uncertaintyBandsMaxDev/2.0)*(nQCDnow-1));
      for (int iVar=1; iVar<nVariationsNow; iVar++) {
        if (Paccept[iVar] > limit*Paccept[0]) Paccept[iVar] = limit * Paccept[0];
        if (Paccept[iVar] < Paccept[0]/limit) Paccept[iVar] = Paccept[0] / limit;
      }      
    }
    
    // Check merging functions
    // 1) Print old list of partons
    /*cout << endl << "Time Shower, before branching:" << endl;
    for (int i=0; i<(int)particles[iSysWin].size(); i++) {
      cout << "  " << setw(3) << particleNrs[iSysWin][i];
      cout << " " << setw(2) << particles[iSysWin][i].id();
      cout << " " << setw(3) << particles[iSysWin][i].col();
      cout << " " << setw(3) << particles[iSysWin][i].acol();
      cout << " " << particles[iSysWin][i].p();
    }
    // 2) Print branching probability
    cout << " alphaS = " << alphaSusr << endl; 
    cout << " col*ant = " << antPhys[0]/sAnt << endl; 
    cout << " Branching probability = " << antPhys[0]/sAnt*alphaSusr << endl;*/
    
    //**********************************************************************
    // Accept / Reject step
    //**********************************************************************
    
    if (rndmPtr->flat() > Paccept[0]) {
      if (verbose >= 5)
        printOut("TimeAntShower::acceptTrial","=== Branching Vetoed. wPhys/wTrial = "
                 +num2str(Paccept[0]));
        
      // In case of reweighting, set the reject weight
      if (allowReweighting) weightsPtr->scaleWeightAll(rejectWeight);
      
      // Check if branching would have happened if we had been using 
      // another radiation function
      if (uncertaintyBandsNow) {
        double Pacc = Paccept[0];
        for (int iVar=1; iVar<nVariationsNow; iVar++) {
          double Pvar = Paccept[iVar];
          double rw = (1.0 - Pvar)/(1.0 - Pacc);
          weightsPtr->scaleWeight(rw,iVar);
        }
      }
      
      // For each failed trial, attempt to (slowly) optimize headroomfactors
      if (meCorrections && nQCDnow <= matchingLO && hardSys[iSysWin] && 
          trialPtr->nVeto >= 1)
        setHeadRoom( (iAntPhys == iQGsplit || iAntPhys == iGGsplit) ? 1 : 0, 
                     -1, nG[iSysWin], nQQ[iSysWin], iSysWin, qNew, 0.999 );
      
      // Count up number of vetoed branchings
      trialPtr->nVeto++;
      ++nTrialsVeto;
      return false;
    }
    
    // In case of reweighting, set the accept weight
    if (allowReweighting) weightsPtr->scaleWeightAll(acceptWeight);
    
    // For accepted branching, rescale uncertainty event weights
    if (uncertaintyBandsNow) {
      for (int iVar=1; iVar<nVariationsNow; iVar++) {
        weightsPtr->scaleWeight(Paccept[iVar]/Paccept[0],iVar);
      }
    }
    
    // Fill diagnostics histos?
    if (verbose >= 3 || (verbose >= 2 && doMEC)) {
      string state = num2str(2*nQQnow,1) + "q" + num2str(nGnow,1) + "g";
      if (trialPtr->new2.colType() == 2) state += "Emit";
      else if (trialPtr->new2.colType() == -1) state += "SplitA";
      else if (trialPtr->new2.colType() == 1) state += "SplitB";
      string HqPhys = "Ln(q2/sSystem):" + state;
      if (vinciaHistos.find(HqPhys) != vinciaHistos.end()) {
        vinciaHistos[HqPhys].fill(log(max(TINY,pow2(qNew/mSystem[iSysWin]))));
      }
      string HalphaS = "alphaS:" + state;
      if (vinciaHistos.find(HalphaS) != vinciaHistos.end()) {
        vinciaHistos[HalphaS].fill(alphaSdef);
      }      
      string HalphaSratio = "alphaSratio:" + state;
      if (vinciaHistos.find(HalphaSratio) != vinciaHistos.end()) {
        vinciaHistos[HalphaSratio].fill(alphaSdef/alphaTrial);
      }      
      if (Pari != 1.0) {
        string Hari = "pAri:" + state;
        if (vinciaHistos.find(Hari) != vinciaHistos.end()) {
          vinciaHistos[Hari].fill(Pari);
        }
      }
      if (doMEC) {
        string HPacc = "Log10(ME/AntPhys):" + state;
        if (vinciaHistos.find(HPacc) != vinciaHistos.end()) {
          vinciaHistos[HPacc].fill(log10(max(TINY,PME[0])));
        }
      }
    }
    
    // Verbose output
    if (verbose >= 6) 
      printOut("TimeAntShower::acceptTrial","finished --------------");

    // All clear, yey
    return true;
  }
  
  // ******
  // 
  // assignColourFlow():
  // 
  // Paint a trialPtr with colour-flow information,
  // according to the type of branching.
  // Nearest-neighbours are never assigned the same "index" (for CR).

  bool TimeAntShower::assignColourFlow(Event& event, BranchelementalFSR* trialPtr) {
    bool usedColTag = false;
    // Set colour flow, deleting branched dipoles and adding new ones.
    trialPtr->new1.acol(trialPtr->old1.acol());
    trialPtr->new1.col(0);
    trialPtr->new2.acol(0);
    trialPtr->new2.col(0);
    trialPtr->new3.acol(0);
    trialPtr->new3.col(trialPtr->old2.col());
    int colOld = trialPtr->old1.col();
    // Gluon emission:
    if (trialPtr->new2.id() == 21) {
      double s12 = dot4(trialPtr->new1.p(),trialPtr->new2.p());
      double s23 = dot4(trialPtr->new2.p(),trialPtr->new3.p());
      // Who inherits the colour?
      bool inh12 = reconnectionsPtr->inherit12(s12,s23);
      // NOTE: Use lastColTag here, if branching accepted we tell
      // Pythia that we used one
      int nTag   = event.lastColTag()+1;
      usedColTag = true;
      // Check other colourtag for gluons
      int colL   = ( (trialPtr->iL != 0) ? trialPtr->old1.acol() : 0 );
      int colR   = ( (trialPtr->iR != 0) ? trialPtr->old2.col()  : 0 );
      int nextTag = 10*int(nTag/10) + 10;
      int colNew  = nextTag + int(colOld%10 + rndmPtr->flat()*8)%9 + 1;
      if (inh12) { // new1 inherits colour
        while (colNew%10 == colR%10) 
          colNew = nextTag + int(colOld%10 + rndmPtr->flat()*8)%9 + 1;   
        trialPtr->new1.col(colOld);
        trialPtr->new2.acol(colOld);
        trialPtr->new2.col(colNew);
        trialPtr->new3.acol(colNew);
      } else { // new3 inherits colour
        while (colNew%10 == colL%10) 
          colNew = nextTag + int(colOld%10 + rndmPtr->flat()*8)%9 + 1;
        trialPtr->new1.col(colNew);
        trialPtr->new2.acol(colNew);
        trialPtr->new2.col(colOld);
        trialPtr->new3.acol(colOld);
      }
    }
    // Xg -> X qbar q
    else if (trialPtr->new2.colType() == -1) {
      trialPtr->new1.col(colOld);
      trialPtr->new2.acol(colOld);
    } 
    // gX -> qbar q X
    else if (trialPtr->new2.colType() == 1) {
      trialPtr->new2.col(colOld);
      trialPtr->new3.acol(colOld);
    }
    return usedColTag;
  }
  
  //*********
    
  /* Check whether radiation falls into sector
     Inputs: 
     event : event record before branching
     trial : proposed trial branching
     Output:
     sectorAccept : false if branching should be vetoed
  */

  bool TimeAntShower::sectorAccept(Event& event,BranchelementalFSR* trialPtr) {

    // Only do check if evolution mode is not global
    if ( ! sectorShower ) return true;

    if (verbose >= 7)
      printOut("TimeAntShower::sectorAccept","begin --------------");
  
    // Check if this sector has smallest sector measure
    int sectorMeasureType = 10;
    if (trialPtr->new2.id() < 0) sectorMeasureType = 8;  
    else if (trialPtr->new2.id() != 21) sectorMeasureType = 9;  
    double QS2 = resolutionPtr->Q2E(trialPtr->new1,trialPtr->new2,trialPtr->new3,
                                    sectorMeasureType);
    // Now compute those of alternative clusterings (= competing sectors)
    double QS2Min = pow2(mSystem[iSysSel]);
    for (int iAnt = 0; iAnt < int(branchelementals[iSysWin].size()); ++iAnt) {
      int ia = branchelementals[iSysWin][iAnt].iOld1;
      int ir = branchelementals[iSysWin][iAnt].iOld2;
      int ib = branchelementals[iSysWin][iAnt].iR;
      // skip if no right-hand neighbor
      if (ib == 0) continue;
      // Compute sector resolution function QS2 for this triplet
      Vec4 pa = event[ia].p();
      Vec4 pr = event[ir].p();
      Vec4 pb = event[ib].p();
      //int idpa = event[ia].id();
      int idpr = event[ir].id();
      //int idpb = event[ib].id();
      
      // use new particles if neigboring pts affected by recoils
      if (ir == trialPtr->iOld1) {
        pr = trialPtr->new1.p();
        pb = trialPtr->new2.p();
        idpr = trialPtr->new1.id();
        //idpb = trialPtr->new2.id();      
      } else if (ib == trialPtr->iOld1) {
        pb = trialPtr->new1.p();
        //idpb = trialPtr->new1.id();
      } //it is not "else if" to account for cyclic topologies
      if (ia == trialPtr->iOld2) {
        pa = trialPtr->new3.p();
        //idpa = trialPtr->new3.id();
      } else if (ir == trialPtr->iOld2) {
        pa = trialPtr->new2.p();
        pr = trialPtr->new3.p();
        //idpa = trialPtr->new2.id();
        idpr = trialPtr->new3.id();
      }
      
      // we use David's prescription for quarks. We rely on color ordering:
      // antiquarks of splitted pairs are always before the corresponding
      // quark
      // TODO: make sure types 8 and 9 also make sense for massive quarks
      sectorMeasureType = 10;
      if (idpr < 0) sectorMeasureType = 8;
      else if (idpr != 21) sectorMeasureType = 9;        
      QS2Min = min(QS2Min,resolutionPtr->Q2E(pa,pr,pb,sectorMeasureType));
      
      if (verbose >= 8) {
        printOut("sectorAccept","testing sector no "+num2str(iAnt)+" QS2 = "
                 +num2str(QS2)+" vs QS2MinNow = "+num2str(QS2Min));
      }
      if (QS2 > QS2Min) return false;
    }
  
    // Return 
    return true;
      
  }
  
  
  //********* 
  
  // Print header information
  void TimeAntShower::header() {
    
    // Avoid printing header several times
    headerIsPrinted = true;

    cout<<setprecision(3);
    cout.setf(ios::left);
    cout << "\n";
    cout << " *-------  VINCIA "<<fixed<<setw(6)<<version
         <<" Global Initialization  ------------------------------------------*\n";
    cout << " | \n";
    cout << " | Vincia                               = "
         <<bool2str(settingsPtr->flag("Vincia"),9) <<"\n |\n";
    
    if (!settingsPtr->flag("Vincia")) {
      cout << " |\n *-------  End VINCIA FSR Global Initialization  "
           <<"-----------------------------------------*\n\n";
      cout.setf(ios::right);
      return;
    }
    
    // Print header information about Matching 
    cout << " | Matrix Element Corrections           = ";
    if (meCorrections && matchingLO >= 1) {
      cout << bool2str(true,9)<<"\n"; 
      cout <<" |                 matchingFullColor    = ";
      cout << bool2str(matchingFullColor,9)<<"\n";
      cout <<" |                 LO QCD               = ";
      cout << num2str(matchingLO,9)<<"  (coupling powers requested)\n";
      cout <<" |                 regOrder             = "
           << num2str(matchingRegOrder,9)<<endl;
      if (matchingScaleIsAbs) 
        cout <<" |                 regScale             = "
             << num2str(matchingScale,9)<<endl;
      else 
        cout <<" |                 regScaleRatio        = "
             << num2str(matchingScaleRatio,9)<<endl;
      if (verbose >= 2) {
        cout <<" |                 regShape             = "
             << num2str(matchingRegShape,9)<<endl;
      }
      cout <<" |                 MGInterface          = "
           << bool2str(isMGOn,9)<<"\n";
    }    
    else {
      cout<<bool2str(false,9)<<"\n";
    }
    
    // Print header information about shower
    cout << " |\n";      
    cout << " | Shower:         evolutionType (FSR)  = "
         << num2str(settingsPtr->mode("Vincia:evolutionType"),9)
         <<"  (1:pT, 2:mD)\n";
    cout << " |                 evolveAllInPT (ISR)  = "
         << bool2str(settingsPtr->flag("Vincia:evolveAllInPT"),9)<<"\n";
    cout << " |                 orderingMode         = "
         << num2str(settingsPtr->mode("Vincia:orderingMode"),9)
         <<"  (1:strong, 2:smooth, 3:smooth MECs + strong)\n";
    cout << " |                 orderingModeMPI      = "
         << num2str(settingsPtr->mode("Vincia:orderingModeMPI"),9)
         <<"  (1:strong, 2:smooth)\n";
    //    cout << " |                 helicityShower       = "
    //     << bool2str(settingsPtr->flag("Vincia:helicityShower"),9)<<"\n";
    //cout << " |                 sectorShower         = "
    //     << bool2str( settingsPtr->flag("Vincia:sectorShower"),9)<<"\n";
    cout << " |                 nGluonToQuarkF (FSR) = "
         << num2str(nGluonToQuarkF,9)
         <<"  (massCorrectionSplitF = "<<massCorrectionSplit<<")"
         <<"\n";
    cout << " |                 nGluonToQuarkI (ISR) = "
         << num2str(settingsPtr->mode("Vincia:nGluonToQuarkI"),9)<<"\n";
    
    // Print header information about alphaS
    cout << " |\n | Alpha_s:        alphaS(mZ)|MSbar     = "
         << num2str(alphaSvalue,9)<<"\n"
         << " |                 order                = "
         << num2str(alphaSorder,9)<<"\n";
    if (alphaSorder >= 1) {
      cout << " |                 LambdaQCD[nF]        = "
           << num2str(alphaSptr->Lambda3(),9)<<"[3] "
           << num2str(alphaSptr->Lambda4(),7)<<"[4] "
           << num2str(alphaSptr->Lambda5(),7)<<"[5] "
           << num2str(alphaSptr->Lambda6(),7)<<"[6]\n";
      cout << " |                 LambdaQCD[nF] (CMW)  = "
           << num2str(alphaSCMWptr->Lambda3(),9)<<"[3] "
           << num2str(alphaSCMWptr->Lambda4(),7)<<"[4] "
           << num2str(alphaSCMWptr->Lambda5(),7)<<"[5] "
           << num2str(alphaSCMWptr->Lambda6(),7)<<"[6]\n";
      cout << " |                 CMW rescaling (FF)   = "
           << num2str(settingsPtr->mode("Vincia:CMWtypeFF"),9)<<"\n";
      cout << " |                 CMW rescaling (II)   = "
           << num2str(settingsPtr->mode("Vincia:CMWtypeII"),9)<<"\n";
      cout << " |                 CMW rescaling (IF)   = "
           << num2str(settingsPtr->mode("Vincia:CMWtypeIF"),9)<<"\n";
      cout << " |                 kMuF                 = "
           << num2str(settingsPtr->parm("Vincia:alphaSkMuF"),9)
           <<" (FSR emissions)\n";
      cout << " |                 kMuSplitF            = "
           << num2str(settingsPtr->parm("Vincia:alphaSkMuSplitF"),9)
           <<" (FSR splittings)\n";
      cout << " |                 kMuI                 = "
           << num2str(settingsPtr->parm("Vincia:alphaSkMuI"),9)
           <<" (ISR emissions)\n";
      cout << " |                 kMuSplitI            = "
           << num2str(settingsPtr->parm("Vincia:alphaSkMuSplitI"),9)
           <<" (ISR splittings)\n";
      cout << " |                 kMuConv              = "
           << num2str(settingsPtr->parm("Vincia:alphaSkMuConv"),9)
           <<" (ISR conversions)\n";

      cout << " |                 alphaSmuFreeze       = "
           << num2str(alphaSmuFreeze,9)<<"\n";
      cout << " |                 alphaSmax            = "
           << num2str(alphaSmax,9)<<"\n";
    }
    
    // Print header information about IR regularization
    cout << " |\n"
         << " | IR Reg.:        cutoffType           = "
         << num2str(settingsPtr->mode("Vincia:cutoffTypeF"),9)<<"\n" 
         << " |                 cutoffScaleFF        = "
         << num2str(cutoffScale,9)<<"\n"
         << " |                 cutoffScaleII        = "
         << num2str(settingsPtr->parm("Vincia:cutoffScaleII"),9)<<"\n"
         << " |                 cutoffScaleIF        = "
         << num2str(settingsPtr->parm("Vincia:cutoffScaleIF"),9)<<"\n"
         << " |                 matchingIRcutoff     = "
         << num2str(matchingIRcutoff,9)<<"\n";
         
    // Print header information from Colour Reconnections 
    reconnectionsPtr->header();
    cout<<" |\n";
      
    // Print header information about Antennae
    if (verbose >= 2) {
      for (int iAnt=0;iAnt<=4;++iAnt) {
        cout.setf(ios::left);
        cout<<setprecision(2);
        cout<<" | *------------------------------------------------------------------------\n";
        cout<<" | * "<<getAnt(iAnt)->vinciaName()<<" ["
            <<getAnt(iAnt)->humanName()<<"]"<<endl;
        if (getAnt(iAnt)->isOn()) { 
          // Print color/charge factor
          double chargeFactor = getAnt(iAnt)->getChargeFactor();
          cout<<" |   chargeFactor       = "<<fixed
              <<setw(6)<<chargeFactor<<"\n";
          int kineMapType = getAnt(iAnt)->getKineMapType();
          cout<<" |   kineMapType        = "
              <<setw(6)<<kineMapType<<"\n";
          cout<<" |   useCollinearTerms  = "<<setw(6)<<
            ( settingsPtr->flag("Vincia:useCollinearTerms") ? "on":"off")
              <<"\n"; 
          cout<<" |   useFiniteTerms     = "<<setw(6)<<
            ( settingsPtr->flag("Vincia:useFiniteTerms") ? "on":"off")<<"\n";
          cout << " | \n";
        } else { // Switched off
          cout<<" |   OFF \n | \n";
        }
      }
    }
    // Print references
    cout <<     " |-------------------------------------------"
         << "------------------------------------------\n |\n";
    cout << " | References :"<<endl;
    cout << " |    VINCIA     : Giele, Kosower, Skands, PRD78(2008)014026, arXiv:0707.3652"<<endl;
    cout << " |    PYTHIA 8   : Sjostrand et al., CPC191(2015)159, arXiv:1410.3012"<<endl;
    if (meCorrections && matchingLO >= 1 && isMGOn) {
      cout << " | The MADGRAPH interface relies on:"<<endl;
      cout << " |    MADGRAPH 4 : Alwall et al., JHEP09(2007)028, arXiv:0706.2334"<<endl;
      cout << " |    HELAS      : Murayama et al., KEK-91-11 (1992)"<<endl;
    }
  
    cout << " |\n *-------  End VINCIA FSR Global Initialization  "
         << "-----------------------------------------*\n\n";
    cout.setf(ios::right);  
  }  
  
  // Print the list of dipoles.
  void TimeAntShower::list(ostream& os) const {
    // Loop over dipole list and print it.
    for (int iSyst = 0; iSyst < (int)branchelementals.size(); iSyst++) {
      for (int iAnt = 0; iAnt < (int)branchelementals[iSyst].size(); ++iAnt) {
        if (branchelementals[iSyst].size() == 1) branchelementals[iSyst][iAnt].list(os,true,true);
        else if ( iAnt == 0 ) branchelementals[iSyst][iAnt].list(os,true,false);
        else if ( iAnt == int(branchelementals[iSyst].size())-1 ) branchelementals[iSyst][iAnt].list(os,false,true);
        else branchelementals[iSyst][iAnt].list(os);
      }
    }
  }

  void TimeAntShower::printInfo(ostream& os, bool pluginCall) {
    if (!isInit) return;
    
    // Weight info
    int nTotWeightsInt         = weightsPtr->nTotWeights;
    int nNonunityInitialWeight = weightsPtr->nNonunityInitialWeight;
    int nNegativeInitialWeight = weightsPtr->nNegativeInitialWeight;
    int nNonunityWeight        = weightsPtr->nNonunityShowerWeight;
    int nNegativeWeight        = weightsPtr->nNegativeShowerWeight;
    double nTotWeights         = max(1.0,((double)(nTotWeightsInt)));
    vector<double> weightSum   = weightsPtr->weightSum;
    vector<double> weightMax   = weightsPtr->weightMaxInitMC;
    
    os << "\n";
    if (pluginCall)
      os << " *--------  VINCIA FSR and ISR Statistics  ------------------------------------------------------------------------*\n";
    else
      os << " *--------  VINCIA FSR Statistics  --------------------------------------------------------------------------------*\n";
    os << " |                                                                                                                 |\n"; 
    os << " |                                                                        "<<setw(40)<<" "<<" |\n";
    os << " | Total Number of (accepted) events                           = "
       <<num2str(nTotWeightsInt,9)<<setw(40)<<" "<<" |\n";
    os << " | Number of events with nonunity initial weight               = "
       <<( (nNonunityInitialWeight <= 0) ? "     none                                " 
           : num2str(nNonunityInitialWeight,9)+" <-- (INITIALLY) WEIGHTED EVENTS" )
       <<setw(8)<<" "<<" |\n";
    os << " | Number of events with negative initial weight               = "
       <<( (nNegativeInitialWeight <= 0) ? "     none" 
           : num2str(nNegativeWeight,9))
       <<setw(40)<<" "<<" |\n";
    os << " | Number of events with nonunity reweight                     = "
       <<( (nNonunityWeight <= 0) ? "     none                      " 
           : num2str(nNonunityWeight,9)+" <-- REWEIGHTED EVENTS" )
       <<setw(18)<<" "<<" |\n";
    os << " | Number of events with negative reweight                     = "
       <<( (nNegativeWeight <= 0) ? "     none" 
           : num2str(min(nNegativeWeight,nNonunityWeight),9))
       <<setw(40)<<" "<<" |\n";
    os << " |                                                                        "<<setw(40)<<" "<<" |\n";   
    os << " |                                   weight(i)          Avg Wt   Avg Dev  rms(dev)      kUnwt     Expected effUnw  |\n";
    os << " | This run                            i =     IsUnw       <w>     <w-1>                1/<w>   Max Wt  <w>/MaxWt  |\n";
    os << " |"<<setw(4)<<" "<<"User settings                         0 "; 
    os <<( (abs(1.0-weightSum[0]/nTotWeights) < TINY) ? "   yes " : "    no " );
    os << num2str(weightSum[0]/nTotWeights,9)<<" ";
    os << num2str(weightSum[0]/nTotWeights-1.0,9)<<" ";
    os << setw(9)<<"-"<<"  ";
    os << ((weightSum[0] != 0.0) ? num2str(nTotWeights/weightSum[0],9): num2str(0.0,9));
    os << num2str(weightMax[0],9)<<" ";
    os << ((weightMax[0] != 0.0) ? num2str(weightSum[0]/nTotWeights/weightMax[0],9): num2str(0.0,9))<<" ";
    os <<"  |\n";
    if (uncertaintyBands) {
      for (int iVar=1; iVar<nWeights; iVar++) {
        if (iVar== 1) os<<" |"<<setw(4)<<" "<<"Var : Default settings              ";
        if (iVar== 2) os<<" |"<<setw(4)<<" "<<"Var : AlphaS:Hi                     ";
        if (iVar== 3) os<<" |"<<setw(4)<<" "<<"Var : AlphaS:Lo                     ";
        if (iVar== 4) os<<" |"<<setw(4)<<" "<<"Var : Antennae:Hi                   ";
        if (iVar== 5) os<<" |"<<setw(4)<<" "<<"Var : Antennae:Lo                   ";
        if (iVar== 6) os<<" |"<<setw(4)<<" "<<"Var : NLO:Hi                        ";
        if (iVar== 7) os<<" |"<<setw(4)<<" "<<"Var : NLO:Lo                        ";
        if (iVar== 8) os<<" |"<<setw(4)<<" "<<"Var : NLC:Hi                        ";
        if (iVar== 9) os<<" |"<<setw(4)<<" "<<"Var : NLC:Lo                        ";
        if (iVar==10) os<<" |"<<setw(4)<<" "<<"Var : ORD:PIMP                      ";
        os << num2str(iVar,3);
        os <<( (abs(1.0-weightSum[iVar]/nTotWeights) < TINY) ? "    yes " : "     no " );
        os << num2str(weightSum[iVar]/nTotWeights,9)<<" ";
        os << num2str(weightSum[iVar]/nTotWeights-1.0,9)<<" ";
        os << setw(9)<<"-"<<"  ";
        os << ((weightSum[iVar] != 0.0) ? num2str(nTotWeights/weightSum[iVar],9): num2str(0.0,9));
        os << num2str(weightMax[iVar],9)<<" ";
        os << ((weightMax[iVar] != 0.0) ? num2str(weightSum[iVar]/nTotWeights/weightMax[iVar],9): num2str(0.0,9))<<" ";
        os <<"  |\n";
      }
    }
    os << " |                                                                        "<<setw(40)<<" "<<" |\n";
    if (pluginCall) {
      os << " | - - - -  FSR only Statistics  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - |\n";
      os << " |                                                                        "<<setw(40)<<" "<<" |\n";
    }
    if (verbose >= 2 && nTrialsSum > 0) {
      os << " | Trial Veto Rates:"
         <<"   Hull = "<<nTrialsHull*1.0/nTrialsSum
         <<"   Massive = "<<nTrialsHullMassive*1.0/nTrialsSum
         <<"   Had M+Q = "<<nTrialsCutM*1.0/nTrialsSum<<" + "<<nTrialsCutQ*1.0/nTrialsSum
         <<"   HQ PS = "<<nClosePSforHQ*1.0/nTrialsSum
         <<"   Veto = "<<nTrialsVeto*1.0/nTrialsSum
         <<"     |"<<endl;
      os << " |                                                                        "<<setw(40)<<" "<<" |\n";
    }
    os << " | Encountered "<<setw(6)<<nMCviolShower+nMCviolDidMEC+nMCviolPlannedMEC<<" MC violations,                         "
       <<"with a maximum violation of "<<setw(15)<<max(max(maxMCviolShower,maxMCviolDidMEC),maxMCviolPlannedMEC)<<setw(10)<<" "<<" |\n";
    os << " |             "<<setw(6)<<nMCviolShower<<" pure shower,                           with a maximum violation of "
       <<setw(15)<<maxMCviolShower<<setw(10)<<" "<<" |\n";
    os << " |             "<<setw(6)<<nMCviolDidMEC<<" with MECs,                             with a maximum violation of "
       <<setw(15)<<maxMCviolDidMEC<<setw(10)<<" "<<" |\n";
    os << " |             "<<setw(6)<<nMCviolPlannedMEC<<" with planned but not carried out MECs, with a maximum violation of "
       <<setw(15)<<maxMCviolPlannedMEC<<setw(10)<<" "<<" |\n";
    os << " |                                                                        "<<setw(40)<<" "<<" |\n";
    os << " | Encountered "<<setw(6)<<nNegPaShower+nNegPaDidMEC+nNegPaPlannedMEC<<" negative accept probabilities,         "
       <<"with a minimum value of "<<setw(15)<<max(max(maxNegPaShower,maxNegPaDidMEC),maxNegPaPlannedMEC)<<setw(14)<<" "<<" |\n";
    os << " |             "<<setw(6)<<nNegPaShower<<" pure shower,                           with a minimum value of "
       <<setw(15)<<maxNegPaShower<<setw(14)<<" "<<" |\n";
    os << " |             "<<setw(6)<<nNegPaDidMEC<<" with MECs,                             with a minimum value of "
       <<setw(15)<<maxNegPaDidMEC<<setw(14)<<" "<<" |\n";
    os << " |             "<<setw(6)<<nNegPaPlannedMEC<<" with planned but not carried out MECs, with a minimum value of "
       <<setw(15)<<maxNegPaPlannedMEC<<setw(14)<<" "<<" |\n";
    os << " |                                                                        "<<setw(40)<<" "<<" |\n";
    if (!pluginCall)
      os << " *--------  End VINCIA FSR Statistics -----------------------------------------------------------------------------*\n\n";
  }

} // end namespace VINCIA
